/*
 * Event Log Manager
 *   
 *  Copyright (C) 2015 Francesco Santeramo
 *
 *  This file is part of Event Log Manager.
 *
 *  Event Log Manager is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Event Log Manager is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *	
 *  Author:  Francesco Santeramo 
 *  Website: www.francescosanteramo.com
 *  Contact: eventlogmanager@francescosanteramo.com    
 */

package unipi.bpm.elm.santeramo.controller;

import java.awt.Dimension;
import java.awt.Insets;
import java.awt.event.ActionListener;
import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import javax.swing.JInternalFrame;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.xml.XMLConstants;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.factory.XFactoryBufferedImpl;
import org.deckfour.xes.id.XID;
import org.deckfour.xes.in.XesXmlParser;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeBoolean;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XAttributeID;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XAttributeTimestamp;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.out.XesXmlSerializer;
import org.deckfour.xes.util.XsDateTimeConversionJava7;
import org.xml.sax.Attributes;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

import com.google.common.collect.AbstractIterator;
import com.google.common.collect.HashMultiset;
import com.google.common.collect.Multiset;

import unipi.bpm.elm.santeramo.model.EventLogManagerModel;
import unipi.bpm.elm.santeramo.view.EventLogEditor;
import unipi.bpm.elm.santeramo.view.FootprintFilter;
import unipi.bpm.elm.santeramo.view.EventLogManagerView;

/**
 * This class defines the behaviour of all the <code>MenuAreaButton</code> buttons in the
 * application tab menus that are connected to analysis procedures and management.
 * 
 * <p>It is used by the Controller that delegates to this class the implementation of the actions
 * connected to these <code>MenuAreaButton</code> buttons.
 * 
 * <p>The Controller receives the user'choice by the relative listener and call the right method
 * of this class.
 * 
 * @author Francesco Santeramo (eventlogmanager@francescosanteramo.com)
 * 
 * @see unipi.bpm.elm.santeramo.view.MenuAreaButton
 * @see EventLogManagerController
 */
public class AnalysisManager {
	
	/*
	 * THE CONTROLLER, THE VIEW AND THE MODEL REFERENCES
	 */
	
	/**
	 * The <code>EventLogManagerController</code> element 
	 * that represents the Controller of the application
	 * and that is the one that uses this analysis manager.
	 * 
	 * @see EventLogManagerController
	 */
	private EventLogManagerController elmController;	
	
	/**
	 * The <code>EventLogManagerView</code> element 
	 * that represents the View of the application managed
	 * by the Controller. 
	 * 
	 * <p>We set this reference to the View managed by the Controller from the
	 * {@link unipi.bpm.elm.santeramo.controller.EventLogManagerController#getView() getView()}
	 * method of the <code>EventLogManagerController</code>.
	 * 
	 * @see EventLogManagerView
	 * @see EventLogManagerController
	 */
	private EventLogManagerView elmView;		
	
	/**
	 * The <code>EventLogManagerModel</code> element 
	 * that represents the Model of the application managed
	 * by the Controller. 
	 * 
	 * <p>We set this reference to the Model managed by the Controller from the
	 * {@link unipi.bpm.elm.santeramo.controller.EventLogManagerController#getModel() getModel()}
	 * method of the <code>EventLogManagerController</code>.
	 * 
	 * @see EventLogManagerModel
	 */
	private EventLogManagerModel elmModel;
	
	
	
	/*
	 * CONSTRUCTOR
	 */
	
	/**
	 * Constructs a new <code>AnalysisManager</code>, with the <code>EventLogManagerController</code> 
	 * passed as parameter associated to it.
	 * 
	 * <p>The passed Controller is the one that uses this <code>AnalysisManager</code> class. This class
	 * uses this reference to the Controller to communicate with it, calling some of its methods.
	 * 
	 * <p>This class is delegated by the Controller to implement some operations, for this reason it is
	 * able to call the View and the Model directly on behalf of the Controller to handle this operation
	 * correctly.
	 * 
	 * @param elmController the <code>EventLogManagerController</code> object that represents the Controller and uses this <code>AnalysisManager</code> 
	 * 
	 * @see EventLogManagerController
	 * @see EventLogManagerView
	 * @see EventLogManagerModel
	 */
	public AnalysisManager(EventLogManagerController elmController) {	
		this.elmController = elmController;
		// the view to use
		elmView = elmController.getView();
		// the model to use
		elmModel = elmController.getModel();
	}
	
	
	
	/*
	 * METHODS
	 */
	
    /**
     * Validates the content of the selected <code>EventLogEditor</code> editor, that is inside
     * the {@link unipi.bpm.elm.santeramo.view.EventLogManagerView#contentArea contentArea}),
     * against the XES XSD Schema to check if it is a valid XES document.
     * 
     * <p>The method first checks if the editor is saved and the file associated to the editor
     * exists (otherwise it asks the user to save the editor content before continuing), then
     * start the validation process using the XES XSD Schema available online at 
     * <code>http://www.xes-standard.org/</code>, or a local copy if the Internet connection is
     * not available at that specific moment.
     * 
     * <p>During the validation, the <code>SAXParser</code> utilises the 
     * {@link AnalysisManager.ParserContentHandler ParserContentHandler} handler to log all the
     * validation warning or errors that may occur; at the end, the method will shows this log 
     * to the user inside a dialog box.
     * 
     * <p>The method is made for the <code>EventLogEditor</code> frame. In case of using it with other
     * frame types, the method manages this situation with specifics information messages delivered
     * to the user.
     * 
     * <p>The validation uses as default the XES XSD schema 2.2 version. To change this setting and set
     * where to find another version of the schema file to use (from offline and/or online sources), 
	 * modify the respective values inside the {@link unipi.bpm.elm.santeramo.model.EventLogManagerModel EventLogManagerModel}.
	 * The user can modify these values at runtime from the application, using the Settings button.
     * 
     * @return <code>true</code> if we successfully validate the file, <code>false</code> otherwise
     * 
     * @see EventLogEditor
     * @see SAXParser
     * @see ParserContentHandler
     */
    public boolean validateEventLog() {
    	// take the selected internal frame to validate
		JInternalFrame selectedInternalFrame = elmView.getContentArea().getSelectedFrame();
		boolean docValidated = false;
		// if an internal frame is selected
		if (selectedInternalFrame != null) {	
			// if the internal frame is an event log editor
			if (selectedInternalFrame instanceof EventLogEditor) {
				// get the event log editor
				EventLogEditor eventLogEditor = (EventLogEditor)selectedInternalFrame;	
				// check if the editor has been modified or not since the last save
				boolean docSaved;				
				if (eventLogEditor.isContentModified()) { // if the editor content has been modified 
					// retrieve the user's choice		    			
	    			String message = "The document needs to be saved before continuing. \n"
	    							+ "Do you want to save it and proceed with the validation?";
	    			String title = "Save";
	    			Object[] options = {"Yes", "   No   "};
	    	    	int choice = elmView.showOptionDialog(message, title, JOptionPane.WARNING_MESSAGE, options, options[0]);
	    	    	if (choice == 0) { // if the user select Yes
	    	    		// try to save the document
						docSaved = elmController.getEditorManager().saveEventLog();		    	    		
	    			} else {
	    				// the user decides not to save the document, so we cannot continue with the validation
	    				docSaved = false;
	    			}
				} else { // if the content is not modified
					// check if this unmodified editor is a new empty editor without an associated file (and so not to validate)
					String absolutePath = eventLogEditor.getAbsolutePath();
					// if this editor does not have an associated file
					if ((absolutePath == null)||(absolutePath.equals(""))) {
						// empty document, nothing to validate
						docSaved = false;
						elmView.showWarningMessage("You first need to write some code in this editor before validating it.");
					} else { // check if this editor still has an associated file to validate
						// create a Path object using the String path
			    		Path targetFile = Paths.get(absolutePath);		    		
			    		if (Files.exists(targetFile)) { // if the file exists
			    			// the editor as a valid associated file and therefore can be validated
							docSaved = true;
			    		} else if (Files.notExists(targetFile)) { // if the target file does not longer exist	
							elmView.showWarningMessage("The file associated with this editor does not longer exist on your file system. \n" 
														+ "Please save this editor again before continuing (otherwise the validation will stop).");							
							// try to save again the document
							docSaved = elmController.getEditorManager().saveAsEventLog(targetFile.getFileName().toString());
			    		} else { //if the process cannot determine if the file exists or not, because it does not have access to the file
			    			elmView.showErrorMessage("The application cannot access the file associated with this editor on your file system and validate it. \n"
													+ "We advise to change the file permissions or to save the document in a new file, then try again to validate it.");
			    			// do not validate
			    			docSaved = false;
			    		}
					}
				}				
				// if we have a file to validate
				if (docSaved) {	
					// start the busy state
					elmView.setBusyState(true);
					// here we get the xes schema file source
					Source xesSchemaFile = null;
					// check if we have to load the file locally or not;
					boolean xesLocal = false;
					// get the xes schema URL online from the Model
					String xesUrl = elmModel.getXesSchemaUrl();
					// if there is a URL to use
					if (!xesUrl.isEmpty()) {
						// check the Internet connection and the URL where to get the Schema
						if (elmController.checkInternetConnection(xesUrl, 15000)) {
							// if we have connection
							try {
								// load the XES Schema definition
			                    URI xesUri = URI.create(xesUrl);
								xesSchemaFile = new StreamSource(new BufferedInputStream(xesUri.toURL().openStream()));
							} catch (IOException e) {
								// if we cannot open a stream to the resource
								elmView.showWarningMessage("Unable to read the XES Schema file at \""+ xesUrl + "\". \n" 
														+ "The local version of this file will be used.");
								// get the Schema locally
								xesLocal = true;
							}						
						} else { // if we do not have Internet connection
							elmView.showWarningMessage("Unable to connect the XES Schema file at \""+ xesUrl + "\". \n" 
													+ "The local version of this file will be used.");
							// get the Schema locally
							xesLocal = true;
						}
					} else {
						elmView.showWarningMessage("There is not a value for the XES Schema URL. \n"
													+ "The local version of this file will be used.");
						// get the Schema locally
						xesLocal = true;
					}
					// if an error occurred, get the Schema locally
					if (xesLocal) {
						// get the Schema locally, asking the Model for the local path
						String xesPath = elmModel.getXesSchemaLocal();
						// check if there is a path to use
						if (!xesPath.isEmpty()) {
							xesSchemaFile = new StreamSource(getClass().getResourceAsStream(xesPath));
						} else {
							elmView.showErrorMessage("There is not a value for the XES Schema local path. \n"
													+ "The system is unable to validate the document due to the absence of a valid XES Schema.");
							return docValidated; // false
						}						
					}
					
					// here we get the xesext schema file source
					Source xesExtSchemaFile = null;
					// check if we have to load the file locally or not;
					boolean xesExtLocal = false;
					// get the xesext schema URL online from the Model
					String xesExtUrl = elmModel.getXesExtSchemaUrl();
					// if there is a URL to use
					if (!xesExtUrl.isEmpty()) {
						// check the Internet connection and the URL where to get the Schema
						if (elmController.checkInternetConnection(xesExtUrl, 15000)) {
							// if we have connection
							try {
								// load the XESEXT Schema definition
			                    URI xesExtUri = URI.create(xesExtUrl);
								xesExtSchemaFile = new StreamSource(new BufferedInputStream(xesExtUri.toURL().openStream()));
							} catch (IOException e) {
								// if we cannot open a stream to the resource
								elmView.showWarningMessage("Unable to read the XESEXT Schema file at \""+ xesExtUrl + "\". \n" 
														+ "The local version of this file will be used.");
								// get the Schema locally
								xesExtLocal = true;
							}						
						} else { // if we do not have Internet connection
							elmView.showWarningMessage("Unable to connect the XESEXT Schema file at \""+ xesExtUrl + "\". \n" 
													+ "The local version of this file will be used.");
							// get the Schema locally
							xesExtLocal = true;
						}
					} else {
						elmView.showWarningMessage("There is not a value for the XESEXT Schema URL. \n "
													+ "The local version of this file will be used.");
						// get the Schema locally
						xesExtLocal = true;
					}
					
					if (xesExtLocal) {
						// get the Schema locally, asking the Model for the local path
						String xesExtPath = elmModel.getXesExtSchemaLocal();
						// check if there is a path to use
						if (!xesExtPath.isEmpty()) {
							xesExtSchemaFile = new StreamSource(getClass().getResourceAsStream(xesExtPath));
						} else {
							elmView.showErrorMessage("There is not a value for the XESEXT Schema local path. \n"
													+ "The system is unable to validate the document due to the absence of a valid XESEXT Schema.");
							return docValidated; // false
						}						
					}					
					elmView.setBusyState(false);
					
					// create a SchemaFactory capable of understanding WXS schemas
			        SchemaFactory schemaFactory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
			        // create the error handler for the factory to manage SAX exceptions
			        SchemaFactoryErrorHandler schemaFactoryErrorHandler = new SchemaFactoryErrorHandler();
			        // set the error handler to log schema errors and manage them
			        schemaFactory.setErrorHandler(schemaFactoryErrorHandler);
			        // create the Schema
			        Schema schema;
					try {			
						elmView.setBusyState(true);
						schema = schemaFactory.newSchema(new Source[] {xesSchemaFile, xesExtSchemaFile});						
						elmView.setBusyState(false);
					} catch (SAXException e) {	
						/* all the SAXException exceptions are managed by the error handler, this if needed
						 * will throw a SAXException to activate this catch and stop the validation */
						elmView.setBusyState(false);
						return docValidated; // false
					}
			        // creating a SAXParserFactory to generate the parser
			        SAXParserFactory parserFactory = SAXParserFactory.newInstance();
			        // set namespace aware to receive element names
			        parserFactory.setNamespaceAware(true);
			        // set the Schema for validation
			        parserFactory.setSchema(schema);
			        // create the SAXParser
			        SAXParser parser;
					try {
						parser = parserFactory.newSAXParser();
					} catch (ParserConfigurationException e) {
						elmView.showErrorMessage("Unable to start the XES Parser. Please try again.");
						return docValidated; // false
					} catch (SAXException e) {
						elmView.showErrorMessage("Unable to start the XES Parser. Please try again. \n" 
												+ "Error: " + e.getMessage());
						return docValidated; // false
					}
					
			        // create an instance of the parser content handler
			        ParserContentHandler parserContentHandler = new ParserContentHandler();
			        // parse the file using the handler to manage the SAXException exceptions
			        try {	
			        	// set the application in a busy state in the meanwhile	
			        	elmView.setBusyState(true);
						parser.parse(new File(eventLogEditor.getAbsolutePath()), parserContentHandler);	
						elmView.setBusyState(false);
					} catch (SAXException e) {
						/* all the SAXException exceptions are managed by the content handler, this 
						 * if needed will throw a SAXException to activate this catch and stop the 
						 * validation*/
						elmView.setBusyState(false);
						return docValidated; // false
					} catch (IOException e) {
						elmView.setBusyState(false);
						elmView.showErrorMessage("Unable to read the XES log. Please try again.");
						return docValidated; // false
					}
			        // if the parsing process did not throw a fatal error, get the result of the parsing
					docValidated = parserContentHandler.isDocumentValid();
					// if the document has been validated, set the content as validated
					if (docValidated) {
						eventLogEditor.setContentValidated(true);
					}
					// now check if there is something to show to the user (warnings, errors, or just a confirmation)
					StringBuffer parserLog = parserContentHandler.getParserLog();
					String parserLogString = parserLog.toString();
					// check if the log is empty or not
					if (parserLogString.equalsIgnoreCase("")) {
						// the log is empty (no warning or error messages)
						elmView.showInfoMessage("Validation Successful! \n"
												+ "The document is valid against the XSD schema of XES.");
					} else { // the log has some warning or error messages to display						
						Path targetFile = Paths.get(eventLogEditor.getAbsolutePath());
	    	    		String frameTitle = "Validation Result of " + targetFile.getFileName().toString();
	    	    		frameTitle = frameTitle.substring(0,frameTitle.lastIndexOf("."));
	    	    		String frameContent = frameTitle + ": \n" + parserLogString;
						// set a text area with the message in it
						JTextArea dialogTextArea = new JTextArea(frameContent);
						// create a scrollable panel that contains the text area
						JScrollPane dialogScrollPane = new JScrollPane(dialogTextArea);  
						// set the text area
						dialogTextArea.setLineWrap(false);  
						dialogTextArea.setWrapStyleWord(true); 
						dialogTextArea.setEditable(false);
						dialogTextArea.setMargin(new Insets(2,4,2,4));
						// set the panel dimension
						dialogScrollPane.setPreferredSize(new Dimension(500, 300));
						// set title and options
						String dialogTitle = "Event Log Manager XES Validation Process";
						int messageType;
						if (docValidated) {
							// if true, only warnings have been generated by the parser
							messageType = JOptionPane.WARNING_MESSAGE;
						} else {
							// if false, at least one error has been generated by the parser
							messageType = JOptionPane.ERROR_MESSAGE;								
						}
		    			Object[] options = {"Ok", "Show on Editor"};
		    			// ask the View to show the option dialog
		    	    	int choice = elmView.showOptionDialog(dialogScrollPane, dialogTitle, messageType, options, options[0]);
		    	    	if (choice == 1) { // if the user select Show on Editor		    	    					    	    		
		    	    		// try-with-resources Statement
		    	    		try (InputStream stringStream = new ByteArrayInputStream(frameContent.getBytes(StandardCharsets.UTF_8));
		    	    			BufferedReader stringReader = new BufferedReader(new InputStreamReader(stringStream))) {
    		    					elmController.newEventLogEditor(frameTitle, stringReader, null, true);  
		    	    		} catch (IOException e) {
	            			    elmView.showErrorMessage("Unable to display the Validation Report. Please try again.");
	        				} 		    	    		  	    		
		    			}
					}
				}				 
    		} else {
    			elmView.showWarningMessage("The selected document is not an Event Log Editor and therefore it cannot be validated.");
    		}
		} else { // if no internal frames is selected 
			elmView.showInfoMessage("No documents to validate. You first need to create a new Event Log or open one.");
		}
    	return docValidated; // true if the document is valid, false otherwise or if some errors occurred during the process
    }
    
    /**
     * This method analyses the XES log contained inside the selected <code>EventLogEditor</code> (or the passed one as parameter),
     * using the <code>EventLogEditor</code> filter (or the one passed as parameter), creating at the end of the process a HTML report
     * with a footprint matrix in it representing the casual dependencies between event traces. 
     * 
     * <p>The report will be eventually showed to the user using a <code>FootprintMatrixInfo</code>, if asked by using the relative parameter.
     * 
     * <p>If the editor filter is not complete while analysing the document, the process, instead of using that filter (because not complete),
     * records all the needed information and updates the filter to make it complete for the next analysis.
     * 
     * <p>The method first checks if the editor content is validated (and therefore saved) and the associated file still exists. If not, it 
     * proceeds asking to validate the content (demanding the control to the {@link #validateEventLog()}) or to save it again (in case of the
     * file does not exists anymore, using the {@link EditorManager#saveAsEventLog(String)} method of the <code>EditorManager</code>).
     * 
     * <p>If the editor content instead is saved and validated, the method checks if there is already a generated analysis report saved in the editor,
     * and in that case it shows it instead of starting a new analysis (because if a report exists, is updated, otherwise it would not exists).
     * 
     * <p>If an already generated analysis report does not exist, the method proceeds with the analysis as described above.
     * 
     * @param printMatrix <code>true</code> to open a <code>FootprintMatrixInfo</code> with the result of the analysis, <code>false</code> to make the analysis and save the result, but without showing them to the user
     * @param editor the <code>EventLogEditor</code> containing the XES log to analyse, <code>null</code> to use the selected editor inside the application
     * @param filter the <code>FootprintFilter</code> to use for the analysis, nothing (this is a optional parameter) to use the editor filter.
     * 
     * @return <code>true</code> if the process successfully generates a footprint matrix for the log, <code>false</code> otherwise.
     * 
     * @see EventLogEditor
     * @see FootprintFilter
     * @see EditorManager
     * @see unipi.bpm.elm.santeramo.view.FootprintMatrixInfo
     */
    public boolean generateEventLogFootprint(boolean printMatrix, EventLogEditor editor, FootprintFilter... filter) {    	
    	// register if we successfully generate the footprint matrix from the editor content
    	boolean isGenerated = false;    	
    	// take the selected internal frame to analyse 
    	JInternalFrame selectedInternalFrame;
    	// if there is a passed editor, use that editor content instead of the selected one inside the application
    	if (editor != null) {
    		selectedInternalFrame = editor;
    	} else {
    		selectedInternalFrame = elmView.getContentArea().getSelectedFrame();
    	}
		// if an internal frame is selected
		if (selectedInternalFrame != null) {	
			// if the internal frame is an event log editor
			if (selectedInternalFrame instanceof EventLogEditor) {
				// get the event log editor
				EventLogEditor eventLogEditor = (EventLogEditor)selectedInternalFrame;
				// check if there are already one or more footprint matrices associated with this editor to display
				if (eventLogEditor.isFootprintMatrixGenerated() && printMatrix && filter.length == 0) {
					// if true, we just show again the already calculated matrices for this editor
					Iterator<String> footprintMatricesIterator = eventLogEditor.getFootprintMatrices();
					String frameTitle = "Footprint Matrix of " + eventLogEditor.getTitle();
					frameTitle = frameTitle.substring(0,frameTitle.lastIndexOf(".")) + " - Log ";
					int index = 0;
					// for every footprint matrix associated with the editor
					while (footprintMatricesIterator.hasNext()) {
						index++;
						// get the footprint matrix
						String footprintMatrix = footprintMatricesIterator.next();				
						/* open a FootprintMatrixInfo frame with the passed title, content set with the
						 * passed one, associated file as null and exported as false */
						elmController.newFootprintMatrixInfo(frameTitle + index, footprintMatrix, null, false); 								
					}
					// set the footprint matrix as generated
					isGenerated = true;
				} else { // if there was not any footprint matrix associated with the editor
					// if there is not a selected filter and the editor filters are not complete
					if (filter.length == 0 && !eventLogEditor.getFootprintFilter().isComplete()) {
						// if the editor has a filters panel opened associated to it, close that panel
		            	if (eventLogEditor.hasFilterDialog()) {
		            		eventLogEditor.getFilterDialog().dispose();
		            		eventLogEditor.setFilterDialog(null);
		            	}
					}
					/* if false, it means that we have never generated a matrix for this editor before or that a change occurred inside of
					 * it after the last generation and therefore the matrices have been cancelled because there were not valid anymore */
					// check if the document has been validated
					if (!eventLogEditor.isContentValidated()) {
						// retrieve the user's choice		    			
		    			String message = "The document needs to be validated before continuing. \n"
		    							+ "Do you want to validate it and proceed with the analysis?";
		    			String title = "Validate";
		    			Object[] options = {"Yes", "   No   "};
		    	    	int choice = elmView.showOptionDialog(message, title, JOptionPane.WARNING_MESSAGE, options, options[0]);
		    	    	if (choice == 0) { // if the user select Yes
		    	    		// ask to the validation process to check the document
		    	    		validateEventLog();	
		    				// if the content has not been validated, return an error and exit
		    				if (!eventLogEditor.isContentValidated()) {
		    					elmView.showErrorMessage("Unable to analyse the document due to validation errors. \n" 
		    											+ "Please first fix these errors and then try again.");
		    					return isGenerated; // false
		    				}
		    			} else {
		    				// the user decides not to validate the document, so we cannot continue with the analysis
		    				return isGenerated; // false
		    			}	    	    	
					}	
					// the document is validated, and therefore is also saved, so we can proceed				
					boolean fileExists;
					// create a Path object using the editor absolute path
					Path targetFile = Paths.get(eventLogEditor.getAbsolutePath());	   		
					// check if the associated file still exists on the file system
					if (Files.exists(targetFile)) { // if the file exists
						// the editor as a valid associated file and therefore can be analysed
						fileExists = true;
					} else if (Files.notExists(targetFile)) { // if the target file does not longer exist	
						elmView.showWarningMessage("The file associated with this editor does not longer exist on your file system. \n" 
														+ "Please save this editor again before continuing (otherwise the analysis will stop).");							
						// try to save again the document
						fileExists = elmController.getEditorManager().saveAsEventLog(targetFile.getFileName().toString());
					} else { //if the process cannot determine if the file exists or not, because it does not have access to the file
						elmView.showErrorMessage("The application cannot access the file associated with this editor on your file system to analyse it. \n"
								+ "We advise to change the file permissions or to save the document in a new file, then try again to analyse it.");
						// do not analyse it
						fileExists = false;
					}		
					// if the target file exists, we can analyse it
					if (fileExists) {
						elmView.setBusyState(true);
						// update targetFile path in case this changed in the above process (save as)
						targetFile = Paths.get(eventLogEditor.getAbsolutePath());
						// get the file associated with the editor
						File editorFile = new File(eventLogEditor.getAbsolutePath());
						// create the XES XML Parser
						XesXmlParser parser = new XesXmlParser(new XFactoryBufferedImpl());
						// create the destination of the parsing process
						Collection<XLog> logs = null;
						// if the file is a .xes file (the XES parser can handle only .xes files)
						if (parser.canParse(editorFile)) {						
							try {					
								// parse the XML document into a XLog
								logs = parser.parse(editorFile);
							} catch (Exception e) {
								elmView.setBusyState(false);
								elmView.showErrorMessage("Unable to analyse the document due to XES parsing errors. Please try again.");
								return isGenerated;
							}
						} else { // the file is not a .xes
							elmView.setBusyState(false);
							elmView.showWarningMessage("The XES parser can only handle XES files (.xes extension). \n" 
														+ "Please save the document in a XES file before continuing (otherwise the analysis will stop).");
							// ask the user to save the document in a new file with .xes extension
							boolean docSaved = elmController.getEditorManager().saveAsEventLog(targetFile.getFileName().toString());						
							// if the user correctly saved the document
							if (docSaved) {
								// call again the method to check the document with the new associated file
								if (filter.length == 0) {
									return generateEventLogFootprint(printMatrix, eventLogEditor);									
								} else {
									return generateEventLogFootprint(printMatrix, eventLogEditor, filter);
								}
							}
							// stop this method if the user did not save the document correctly
							return isGenerated;
						}
						// set the generated XLog associated with the editor
						eventLogEditor.setXesLogs(logs);
						// now that we have a Java representation of the XES log, we start creating the footprint matrix
						// define the title of the frame, in case of one or more XLog
						String frameTitle = "Footprint Matrix of " + targetFile.getFileName().toString();
						frameTitle = frameTitle.substring(0,frameTitle.lastIndexOf("."));
						if (logs.size() > 1) {												
							frameTitle = frameTitle + " - Log ";
						}						
						int logIndex = 0;
						boolean editorFilter;
						FootprintFilter footprintFilter;
						// if there is an external filter to use
						if (filter.length > 0) {
							// get the external filter
							footprintFilter = filter[0];
							// set that we do not use the editor filter, but an external one
							editorFilter = false;
						} else { // if there is not, it means we want to use the editor one
							// get the editor filter
							footprintFilter = eventLogEditor.getFootprintFilter();
							// set that we will use the editor filter
							editorFilter = true;
						}
						// get if the selected filter is complete or we have to complete it during this process
						boolean filterComplete = footprintFilter.isComplete();
						HashMap<String, Boolean> startEvents = null;
						HashMap<String, Boolean> endEvents = null;
						// if the filter is complete, we get the selected start and end events to use as valid during the analysis
						if (filterComplete) {
							startEvents = footprintFilter.getStartEvents();
							endEvents = footprintFilter.getEndEvents();
						} else { // else we set two new maps for recording new events
							startEvents = new HashMap<String, Boolean>();
							endEvents = new HashMap<String, Boolean>();
						}
						// here we record the classifiers set in the log, for completing the filter
						Collection<XEventClassifier> classifierList = null;
						// here the selected classifier to use
						XEventClassifier eventClassifier;
						// the attributes over which the classifier is defined
						String[] eventClassifierAttributes;
						// if the filter is complete
						if (filterComplete) {
							// get the selected classifier
							eventClassifier = footprintFilter.getSelectedClassifier();	
							// get the classifier attributes
							eventClassifierAttributes = eventClassifier.getDefiningAttributeKeys();
						} else {								
							eventClassifier = null;
							eventClassifierAttributes = null;
						}
						// for each XLog of the editor
						for (XLog log : logs) {
							logIndex++;							
							// if the filter is not complete, get the defined classifiers and add them to the list 
							if (editorFilter && !filterComplete) {
								if (logIndex == 1) {
									classifierList = log.getClassifiers();
								} else {
									classifierList.addAll(log.getClassifiers());
								}								
							}	
							XEventClasses eventClasses;													
							// if the filter is complete, get the event classes based on the selected classifier
							if (filterComplete) {
								eventClasses = XEventClasses.deriveEventClasses(eventClassifier, log);	
							} else { // else get the event classes based on the event name classifier (a default one)
								XLogInfo info = XLogInfoFactory.createLogInfo(log);
								// get the event classes based on the event name classifier
								eventClasses = info.getNameClasses();
								// get the event name classifier 
								eventClassifier = eventClasses.getClassifier();
								// get the event name classifier attributes, to complete the classifier definition
								eventClassifierAttributes = eventClassifier.getDefiningAttributeKeys();
							}
							// if we get the event classes from the classifier
							if (eventClasses != null) {
								// get the number of classes
								int eventClassesNumber = eventClasses.size();
								// if there is at least one class
								if (eventClassesNumber > 0) {	
									/* here we record a set of strings, one for each trace
									 * This string is composed by the values read from the selected event attributes
									 * The events are divided inside the string by a comma
									 * The values of the attribute events are divided inside the string by a plus */
									HashMultiset<String> logTraceSet = HashMultiset.create();								 	
									// for every trace of the log
									for (XTrace trace : log) {	
										/* if we are using the editor filter or an external complete filter, we have a list 
										 * of selected start events and end events, then we have to check if this trace
										 * starts or ends with an selected event or not and discard it in this second case*/
										if (editorFilter || (!editorFilter && filterComplete)) {
											// get the first event of the trace											
											XEvent borderEvent = trace.get(0);
											String attributeValue = "Undefined";
											// get the name of the event
											if (borderEvent.hasAttributes()) {
												XAttributeMap attributeMap = borderEvent.getAttributes();
												// the name of the event is in the concept:name attribute
												if (attributeMap.containsKey("concept:name")) {			
													// get the value of this attribute
													XAttributeLiteral attribute = (XAttributeLiteral)attributeMap.get("concept:name");
													attributeValue = attribute.getValue();												
												}
											}
											// if the filter is complete
											if (filterComplete) {
												// we check if this start event is selected or not
												Boolean isSelected = startEvents.get(attributeValue);
												// if it is not selected, skip this trace
												if (isSelected != null) {
													if (isSelected == false) {
														continue;
													}		
												} else {
													continue;
												}
											} else { // if the filter is not complete
												// add the event to the start event map, for completing the filter
												startEvents.put(attributeValue, true);
											}
											
											// now we get the last event of the trace
											borderEvent = trace.get(trace.size()-1);
											attributeValue = "Undefined";
											// get the name of this event
											if (borderEvent.hasAttributes()) {
												XAttributeMap attributeMap = borderEvent.getAttributes();
												// its name is in the concept:name attribute
												if (attributeMap.containsKey("concept:name")) {		
													// get the value of this attribute
													XAttributeLiteral attribute = (XAttributeLiteral)attributeMap.get("concept:name");
													attributeValue = attribute.getValue();												
												}
											}
											// if filter is complete
											if (filterComplete) {
												// we check if this end event is selected or not
												Boolean isSelected = endEvents.get(attributeValue);
												// if it is not selected, skip this trace
												if (isSelected != null) {
													if (isSelected == false) {
														continue;
													}		
												} else {
													continue;
												}
											} else {// if the filter is not complete
												// add the event to the end event map, for completing the filter
												endEvents.put(attributeValue, true);
											}
										}
										// now that we have discharged traces that had an event or end event not selected
										// here we record the event attribute values, divided between values with a +, and between events with a comma
										StringBuffer traceEventList = new StringBuffer(); 
										boolean eventComma = false;
										// for each event of the trace
										for (XEvent event : trace) {
											// if the event has attributes, it is valuable and we can analyse it
											if (event.hasAttributes()) {
												// get event attribute map
												XAttributeMap attributeMap = event.getAttributes();
												// here we save the list of values read from this event depending on the classifier attribute list
												StringBuffer eventAttributes = new StringBuffer();
												boolean eventPlus = false;
												int attributesFound = 0;
												// for every attribute of the classifier
												for (String eventClassifierAttribute : eventClassifierAttributes) {
													// we check if this event has that attribute
													if (attributeMap.containsKey(eventClassifierAttribute)) {	
														// if it does, get tat attribute
														XAttribute attribute = attributeMap.get(eventClassifierAttribute);	
														String value = null;	
														// read the value of the attribute, depending on its type
														if (attribute instanceof XAttributeLiteral || attribute instanceof XAttributeDiscrete ||
															attribute instanceof XAttributeContinuous || attribute instanceof XAttributeBoolean ||
															attribute instanceof XAttributeID) {															
															value = attribute.toString();
														} else if (attribute instanceof XAttributeTimestamp) {
															Date timestamp = ((XAttributeTimestamp)attribute).getValue();
															XsDateTimeConversionJava7 xsDateTimeConversion = new XsDateTimeConversionJava7();
															value = xsDateTimeConversion.format(timestamp);
														} else { // if an error occurred
															break;
														}
														// if we read a value, append the value to the list
														if (value != null) {
															if (!eventPlus) {
																// the first insert in the list
																eventAttributes.append(value);
																eventPlus = true;
															} else {
																// from the second, we use a + to separate the values
																eventAttributes.append("+").append(value);
															}	
															// set that we read an attribute value
															attributesFound++;	
														} else { // if we did not read a value
															break;
														}
													} else { // if the event does not have the attribute set by the classifier
														// stop analysing the event, because it cannot be classified
														break;
													}
												}
												// at this point we check if this event had all the attributes defined by the classifier
												if (attributesFound == eventClassifierAttributes.length) {
													// if it has them, append the list of attribute values read before to the trace list of event values
													if (!eventComma) {
														// the first insert in the list
														traceEventList.append(eventAttributes);
														eventComma = true;
													} else {
														// from the second, we use a comma to separate the events
														traceEventList.append(",").append(eventAttributes);
													}													
												}
											}
										}
										// if this trace contained at least one valid event
										if (traceEventList.length() > 0) {
											// add the trace and the recorded values of its events to the log list
											logTraceSet.add(traceEventList.toString());
										}								
									}
									// at this point we have a list of traces, events, and values to analyse
									// create a map combining every event class with an ID
									LinkedHashMap<String, Integer> eventClassIdentifiers = new LinkedHashMap<String, Integer>();
									int mapIndex = 0;
									for (XEventClass eventClass : eventClasses.getClasses()) {
										eventClassIdentifiers.put(eventClass.getId(), mapIndex);
										mapIndex++;
									}								
									// get the number of event classes
									int eventClassNumbers = eventClassIdentifiers.size();
									// create a matrix NxN, where N is equal to the event class number
									int[][] relationMatrix = new int[eventClassNumbers][eventClassNumbers];
									// create a set of alias for the event classes, using the alphabet sequence generator
									String[] eventClassAliases = new String[eventClassNumbers];								
									AlphabetSequenceGenerator sequence = new AlphabetSequenceGenerator();
									for (int i=0;i<eventClassNumbers;i++){
										eventClassAliases[i] = sequence.next();
									}
									// start to create the HTML report of the analysis
									// here we save the list of traces and events
									StringBuilder traceList = new StringBuilder();	
									boolean traceComma = false;
									// for every trace in the log set recorded before
									for (Multiset.Entry<String> trace : logTraceSet.entrySet()) {
										// if this trace occurs in the log fewer than the defined limit in the filter, do not consider it
										if (trace.getCount() < footprintFilter.getMinTraceOccurrences()) {
											continue;
										}									
										// add a comma separator between traces in the list
										if (traceComma) {
											traceList.append(", ");
										}	
										boolean error = false;
										// HTML code for the < 
										traceList.append("<span class=\"tracebracket\">&#8249;</span> ");
										// get the trace events. 
										// event = the set of attributes values read from an event, that defines an event class
										String[] events = trace.getElement().split(",");
										// if the set of values has something to use
										if (events.length > 0) {
											/* here we not only add the event to the trace list, using the alias 
											 * of its respecting class, but we also upgrade the matrix with the
											 * event dependencies for being able to create the footprint matrix */
											Integer row;
											// we already fix the first element to use with the id of the class of the first trace event
											Integer col = eventClassIdentifiers.get(events[0].trim());
											if (col != null) {
												// add to the trace list the first event, using the alias of its class name
												traceList.append(eventClassAliases[col]);
												// for every other event in the trace
												for (int i = 1; i < (events.length); i++) {
													// we move the column value to row, to add now the relation between this event and the next one
													row = col;
													// get the id of the class of the next event
													col = eventClassIdentifiers.get(events[i].trim());
													// now we have a row id that is the class of an event, and a column id that is the class of the following one
													if (row != null && col != null) {
														/* record this relation between these two events in the matrix, 
														 * adding in the relative position the number of occurrences of this relation */
														relationMatrix[row][col] = relationMatrix[row][col] + trace.getCount();		
														// add the second event to the trace list
														traceList.append(", " + eventClassAliases[col]);
													} else { // if an error occurred, show it
														error = true;
														// stop the process
														break;
													}
												}
											} else { // if an error occurred, show it
												error = true;
											}
										}
										// the HTML code for >
										traceList.append(" <span class=\"tracebracket\">&#8250;</span><sup>").append(trace.getCount()).append("</sup>");
										traceComma = true;
										// here we check if an error occurred
										if (error) {
											// if it happened before, stop the generation of the footprint matrix for this log
											traceList = new StringBuilder("Error: the Event attributes do not correspond to the Classifier attributes set.");
											// reset the matrix
											relationMatrix = new int[eventClassNumbers][eventClassNumbers];
											break;
										}
									}
									// after writing the trace list and calculating all the relation, recording them in the matrix
									// set the HTML document title
									String htmlTitle = frameTitle;
									if (logs.size() > 1) {												
										htmlTitle = htmlTitle + logIndex;
									}
									// get the line separator specific of the machine system
			    					String newline = System.getProperty("line.separator");
			    					// set the font for the matrix symbols (this is used only for the export in PDF)
									String font = this.getClass().getResource("/fonts/DejaVuSans.ttf").toString();
									// set the HTML value for the matrix symbols
									String casuality = "&#8594;"; // -> &#x2192;
									String inverseCasuality = "&#8592;"; // <- &#x2190;
									String parallel = "||"; // ||
									String choice = "#"; // #							
									int barrier = 1;		
									// write in a string all the HTML code that defines the structure of the report document
									StringBuilder xhtml = new StringBuilder();							
									xhtml.append("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">").append(newline);
									xhtml.append("<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">").append(newline);
									xhtml.append("\t").append("<head>").append(newline);
									xhtml.append("\t\t").append("<title>").append(htmlTitle).append("</title>").append(newline);
									xhtml.append("\t\t").append("<style type=\"text/css\">").append(newline);
									xhtml.append("\t\t\t").append("@font-face{font-family:\"DejaVu Sans\"; src:url(\"").append(font).append("\"); -fs-pdf-font-embed:embed; -fs-pdf-font-encoding: Identity-H;}").append(newline);
									xhtml.append("\t\t\t").append("body{font-family: \"Times New Roman\", Times, serif;}").append(newline);
									xhtml.append("\t\t\t").append("h1{font-size:16px; margin:13px 8px 8px;}").append(newline);
									xhtml.append("\t\t\t").append("h2{font-size:13px; margin:0px 0px 8px;}").append(newline);
									xhtml.append("\t\t\t").append("p{font-size:12px; margin:12px 13px 7px;}").append(newline);									
									xhtml.append("\t\t\t").append("sup{font-size:10px;}").append(newline);
									xhtml.append("\t\t\t").append("table{border-collapse:collapse; border-spacing:0; margin:16px 10px 16px 10px; border-top:1px solid black; border-bottom:1px solid black; padding-bottom:4px;}").append(newline);							
									xhtml.append("\t\t\t").append("th{font-size:12px; text-align:center; vertical-align:middle; padding:6px 15px; border-width:0px;}").append(newline);														
									xhtml.append("\t\t\t").append("td{font-family: \"DejaVu Sans\"; font-size:10px; text-align:center; vertical-align:middle; padding:6px 15px; border-width:0px;}").append(newline);
									xhtml.append("\t\t\t").append("th.rowheader{border-bottom:2px solid black; margin-bottom:4px; padding:6px 15px 7px;}").append(newline);
									xhtml.append("\t\t\t").append("#filters{margin:3px 13px 15px}").append(newline);							
									xhtml.append("\t\t\t").append("#legend{margin:3px 13px 10px}").append(newline);								
									xhtml.append("\t\t\t").append(".list{font-size:12px; margin:3px 7px}").append(newline);
									xhtml.append("\t\t\t").append(".tracebracket{font-size:15px;}").append(newline);
									xhtml.append("\t\t\t").append(".symbol{font-family: \"DejaVu Sans\"; font-size:10px}").append(newline);
									xhtml.append("\t\t").append("</style>").append(newline);
									xhtml.append("\t").append("</head>").append(newline);
									xhtml.append("\t").append("<body>").append(newline);
									xhtml.append("\t\t").append("<h1>").append(htmlTitle).append("</h1>").append(newline);
									
									// add the trace list to the report
									xhtml.append("\t\t").append("<p>").append(newline);							
									xhtml.append("\t\t\t").append("<b>L</b> = [").append(traceList).append("]").append(newline);							
									xhtml.append("\t\t").append("</p>").append(newline);	
									
									// calculate how many columns will be the matrix and prepare a string with LaTeX code (used in case of export in LaTeX)
									StringBuilder latexcols = new StringBuilder();
									for(int i=0; i<eventClassNumbers; i++) {
										latexcols.append("c ");
									}
									latexcols.append("c");
									// add the matrix table to the report, with the LaTeX attribute
									xhtml.append("\t\t").append("<table latexcols='" + latexcols.toString() + "'>").append(newline);
									
									// add the column line to the table, with the first element blank
									xhtml.append("\t\t\t").append("<thead>").append(newline);
									xhtml.append("\t\t\t\t").append("<tr>").append(newline); 
									xhtml.append("\t\t\t\t\t").append("<th class=\"rowheader\">&nbsp;</th>").append(newline);							
									// add all the event class aliases to the column line
									for (int i=0; i<eventClassNumbers; i++) {
										xhtml.append("\t\t\t\t\t").append("<th class=\"rowheader\">").append(eventClassAliases[i]).append("</th>").append(newline);
									}									
									xhtml.append("\t\t\t\t").append("</tr>").append(newline);
									xhtml.append("\t\t\t").append("</thead>").append(newline);
									
									xhtml.append("\t\t\t").append("<tbody>").append(newline);
									// for each event in the matrix row line
									for (int i=0; i<eventClassNumbers; i++) {	
										// new table line
										xhtml.append("\t\t\t\t").append("<tr>").append(newline);
										// the first element of the line is the event class alias
										xhtml.append("\t\t\t\t\t").append("<th>").append(eventClassAliases[i]).append("</th>").append(newline);
										// now calculate the relation between two events based on the values of the relation matrix
										// for every event class in the matrix column line
										for (int j=0; j<eventClassNumbers; j++)
									    {	
											String relation;
											// get the value of i->j
									    	int valueA = relationMatrix[i][j];
									    	// if i and j are the same class (the relation matrix is NxN)
									    	if (i == j) {
									    		// if the value is lower than the barrier (that is 1)
												if(valueA < barrier) {								
													// does not exist a relation between these two events
													relation = choice;									   		
												} else {
													// the relation is parallel
													relation = parallel;
												}
										    } else {// if they are not the same class
										    	// we have to check the value of j->i
										    	int valueB = relationMatrix[j][i];
										    	// if the value A is higher than the barrier (that is 1)
											    if (valueA >= barrier) {
											    	// if the value B is higher than the barrier (that is 1)
											    	if(valueB >= barrier) {
											    		// their relation is parallel
											    		relation = parallel;	
											    	} else {
											    		// there is a relation i->j
											    		relation = casuality;
											    	}
											    } else { // if instead the value A is lower than the barrier (that is 1)
											    	// if the value B is higher than the barrier (that is 1)
											    	if(valueB >= barrier) {
											    		// there is a relation j->i
											    		relation = inverseCasuality;
											    	} else {
											    		// does not exist a relation between these two events
											    		relation = choice;
											    	} 
											    }
										    }
									    	// add the calculated relation to the table, using the relative symbol
									    	xhtml.append("\t\t\t\t\t").append("<td>").append(relation).append("</td>").append(newline);
									    	// go to the next event class in the column line, keeping the event class in the row line
									    }
										// we finish to check the relations in this line
										xhtml.append("\t\t\t\t").append("</tr>").append(newline);
										// go to the next event class in the row line
									}
									xhtml.append("\t\t\t").append("</tbody>").append(newline);
									xhtml.append("\t\t").append("</table>").append(newline);
									
									// add the filters used for the analysis to the report
									xhtml.append("\t\t").append("<div id=\"filters\">").append(newline);
									xhtml.append("\t\t\t").append("<h2>Filters</h2>").append(newline);							
									xhtml.append("\t\t\t").append("<div class=\"list\"><i>Selected Classifier:</i> ").append(eventClassifier.name()).append("</div>").append(newline);
									xhtml.append("\t\t\t").append("<div class=\"list\"><i>Min Trace Occurences:</i> ").append(footprintFilter.getMinTraceOccurrences()).append("</div>").append(newline);
									
									// only if the filter is complete we have this filter option
									if (filterComplete) {
										// calculate the selected start events
										StringBuilder selectedStartEvents = new StringBuilder();
										boolean eventComma = false;
										// for every start event of the log
										for (Map.Entry<String, Boolean> startEvent : startEvents.entrySet()) {
								
											// add it to the filter list if it was selected
											if (startEvent.getValue() == true) {
												// use a comma to separate the list of start events
												if (!eventComma) {
													selectedStartEvents.append(startEvent.getKey());
													eventComma = true;
												} else {
													selectedStartEvents.append(", ").append(startEvent.getKey());
												}											
											}
										}
										// add to the report the list of selected start events
										xhtml.append("\t\t\t").append("<div class=\"list\"><i>Start Events (concept:name):</i> ").append(selectedStartEvents).append("</div>").append(newline);
													
										// calculate the selected end events
										StringBuilder selectedEndEvents = new StringBuilder();
										eventComma = false;
										// for every end event of the log
										for (Map.Entry<String, Boolean> endEvent : endEvents.entrySet()) {
											// add it to the filter list if it was selected
											if (endEvent.getValue() == true) {
												// use a comma to separate the list of end events
												if (!eventComma) {
													selectedEndEvents.append(endEvent.getKey());
													eventComma = true;
												} else {
													selectedEndEvents.append(", ").append(endEvent.getKey());
												}											
											}
										}
										// add to the report the list of selected end events
										xhtml.append("\t\t\t").append("<div class=\"list\"><i>End Events (concept:name):</i> ").append(selectedEndEvents).append("</div>").append(newline);
									}
									xhtml.append("\t\t").append("</div>").append(newline);
									// add to the report a legend containing the used aliases and the associated event class name
									xhtml.append("\t\t").append("<div id=\"legend\">").append(newline);
									StringBuilder classifierAttributes = new StringBuilder();
									boolean attributeComma = false;
									// make a list of the event classifier attributes used for the analysis, separated by a plus character
									for (String attribute : eventClassifierAttributes) {
										if (!attributeComma) {
											classifierAttributes.append(attribute);
											attributeComma = true;
										} else {
											classifierAttributes.append("+").append(attribute);
										}										
									}
									// add legend title and legend column title, using the attribute list just made
									xhtml.append("\t\t\t").append("<h2>Matrix Legend</h2>").append(newline);							
									String leftArrow = "<span class=\"symbol\"> &#8594; </span>";
									xhtml.append("\t\t\t").append("<div class=\"list\"><b>alias</b>").append(leftArrow).append(classifierAttributes).append("</div>").append(newline);
									// add a number of rows to the legend, one for each alias, containing the alias name and the associated event class
									for (Map.Entry<String, Integer> eventClass : eventClassIdentifiers.entrySet()) {								    
									    xhtml.append("\t\t\t").append("<div class=\"list\"><b>").append(eventClassAliases[eventClass.getValue()]).append("</b>").append(leftArrow).append(eventClass.getKey()).append("</div>").append(newline);					    
									}
									xhtml.append("\t\t").append("</div>").append(newline);
									
									// close the HTML document
									xhtml.append("\t").append("</body>").append(newline);
									xhtml.append("</html>");
									
									// get the string representing this HTML document
									String footprintMatrix = xhtml.toString();
									// if we used the editor filter
									if (filter.length == 0) {
										// associate the generated HTML document with the editor
										eventLogEditor.addFootprintMatrix(footprintMatrix);																			
									}
									// if the option was set as "print" the report after the analysis, we show it to the user
									if (printMatrix) {
										/* open a FootprintMatrixInfo with the selected title and the content set
										 * with the just generated HTML code, the absolute path of the associated file
										 * as null and the content as not exported. */
					    	    		elmController.newFootprintMatrixInfo(htmlTitle, footprintMatrix, null, false); 		    	    			
									}																					
								} else { // if we did not find any event classes using the selected classifier
									// if we used the editor filter and it was not complete
									if (editorFilter && !filterComplete) {
										// add values to start/end event maps
										startEvents.put("Undefined", true);
										endEvents.put("Undefined", true);
									}
									// show the situation to the user
									elmView.showErrorMessage("The selected Classifier did not generate any Event Class to use for the analysis. \n"
															+"Please select another Classifier from the Filters Panel to analyse the document correctly.");
								}
							} else { // if the selected classifier is not associated with the analysed log (i.e. not defined inside the log)
								// if we used the editor filter and it was not complete
								if (editorFilter && !filterComplete) {
									// add values to start/end event maps
									startEvents.put("Undefined", true);
									endEvents.put("Undefined", true);
								}
								// show the situation to the user
								elmView.showErrorMessage("The Classifier selected for the analysis is not associated with the document. \n"
														+ "Please select another Classifier from the Filters Panel to analyse the document correctly.");
							}
						}
						// if we used the editor filter and it was not complete 
						if (editorFilter && !filterComplete) {
							// we update the editor filter with the calculated values during the analysis
							eventLogEditor.updateFootprintFilter(footprintFilter.getMinTraceOccurrences(), classifierList, eventClassifier, startEvents, endEvents);
						}
						elmView.setBusyState(false);
						// set the process as completed
						isGenerated = true;
					}
				}	
			} else { // if the frame is not an event log editor
			elmView.showWarningMessage("The selected document is not an Event Log Editor and therefore it cannot be analysed.");
			}
		} else { // if no internal frames is selected 
			elmView.showInfoMessage("No documents to analyse. You first need to create a new Event Log or open one.");
		}
		// return true if the footprint matrix has been generated successfully, false otherwise
		return isGenerated;
	}

    /**
     * This method gets the selected internal frame and, if it is an <code>EventLogEditor</code>,
     * it checks its <code>FootprintFilter</code> to ensure it is complete (and if it is not, the 
     * method asks the user if it wants or not to update the filter). 
     * 
     * <p>Then it asks the View to show a filters dialog associated with the editor. The dialog
     * content will be calculated based on the editor <code>FootprintFilter</code>, to show to
     * the user the possible selections.
     * 
     * <p>An {@link ActionListener} will be associated to the dialog buttons
     * to implement the panel functions.
     * 
     * @param listener the {@link ActionListener} listener that will be associated with the dialog buttons
     * 
     * @see EventLogEditor
     * @see FootprintFilter
     * @see ActionListener
     */
    public void setFootprintFilters(ActionListener listener) {
    	// take the selected internal frame to analyse 
		JInternalFrame selectedInternalFrame = elmView.getContentArea().getSelectedFrame();
		// if an internal frame is selected
		if (selectedInternalFrame != null) {	
			// if the internal frame is an event log editor
			if (selectedInternalFrame instanceof EventLogEditor) {
				// get the event log editor
				EventLogEditor eventLogEditor = (EventLogEditor)selectedInternalFrame;
				// get the footprint filter of the editor
				FootprintFilter footprintFilter = eventLogEditor.getFootprintFilter();
				// if the filter is not complete
				boolean showFilter = true;
				if (!footprintFilter.isComplete()) {
					// retrieve the user's choice		    			
	    			String message = "The application will generate a complete list of filters after the first analysis of the document. \n"
	    							+ "Do you want to analyse the document now before continuing to get a complete list of filters, \n"
	    							+ "or continue without analysis and get a subset of them?";
	    			String title = "Incomplete Filters";
	    			Object[] options = {"Show All Filters", "Show Filters Subset"};
	    	    	int choice = elmView.showOptionDialog(message, title, JOptionPane.WARNING_MESSAGE, options, options[0]);	    	    	
	    	    	if (choice == 0) { // if the user select Show all
	    	    		// call the algorithm to calculate a complete filter over the selected editor content, without printing any matrix
	    	    		showFilter = generateEventLogFootprint(false, eventLogEditor);
	    			} else if (choice == -1) { // if the user closes the dialog
	    				// stop the process
	    				return;
	    			}
				}
				if(showFilter) {					
					// get the title for the dialog
			    	String dialogTitle = "Set Filters for " + eventLogEditor.getTitle();
					dialogTitle = dialogTitle.substring(0,dialogTitle.lastIndexOf("."));
					// ask the View to create a new filters dialog with the passed title and associated to the editor
					elmView.newFiltersDialog(dialogTitle, eventLogEditor, listener);		
				} else { // if an error occurred while analysing the document
					elmView.showErrorMessage("Error while analysing the document. Please try again.");
				}
			} else { // if the frame is not an event log editor
				elmView.showWarningMessage("The selected document is not an Event Log Editor and therefore there are no filters available.");
			}
		} else { // if no internal frames is selected 
			elmView.showInfoMessage("No documents selected. You first need to create a new Event Log or open one.");
		}
    }
    
    /**
     * This method generates a XES log from a CSV document that is inside the selected 
     * <code>EventLogEditor</code>.
     * 
     * <p>This method first checks if the selected frame is an <code>EventLogEditor</code>,
     * then it calls the method {@link #parseCSV(String, String)} to implement the process.
     *
     * <p>At the end of the process, the system will show to the user the generated XES log 
     * inside a new editor.
     * 
     * @return <code>true</code> if the process successfully generates a XES log from the CSV Document, <code>false</code> otherwise. 
     */
    public boolean generateXesFromCsv() {
    	// get the result of the generation
    	boolean logGenerated = false;
    	// take the selected internal frame to analyse 
		JInternalFrame selectedInternalFrame = elmView.getContentArea().getSelectedFrame();
		// if an internal frame is selected
		if (selectedInternalFrame != null) {	
			// if the internal frame is an event log editor
			if (selectedInternalFrame instanceof EventLogEditor) {
				// get the event log editor
				EventLogEditor eventLogEditor = (EventLogEditor)selectedInternalFrame;				
				Path csvTempFile = null;
				try {
					// set the busy state
					elmView.setBusyState(true); 
					// create a temp file in the system default temp folder
					csvTempFile = Files.createTempFile("EventLogManager_CSV-XES_", ".txt");
					// delete the file when the virtual machine terminates
					csvTempFile.toFile().deleteOnExit();
					elmView.setBusyState(false); 									
				} catch (IOException e) {
					// error while creating the temp file
					elmView.setBusyState(false);	
					elmView.showErrorMessage("Unable to generate the XES log. \n" 
											+ "A problem with the system temporary-file directory occurred. \n" 
											+ "Please try again.");
					return logGenerated; // false
				}
				// if we successfully created the temp file, open an output stream over it, using try-with-resource statement 
				try (OutputStream fileStream = new FileOutputStream(csvTempFile.toString());
	    				BufferedWriter fileWriter = new BufferedWriter(new OutputStreamWriter(fileStream))) {
	    			// set a busy state	    			        
			        elmView.setBusyState(true);
			        // write the editor content in the temp file
			        eventLogEditor.getTextAreaContent(fileWriter);
			        elmView.setBusyState(false);
	            } catch (IOException e1) {
	            	// error while writing the file
	            	elmView.setBusyState(false);
	            	// at this point we do not need anymore the temp file, so we delete it
		            if (csvTempFile != null) {		            	
						try {
							Files.delete(csvTempFile);
						} catch (IOException e2) {
							/* we tried to delete the temp file that we created, but we are not able to.
							 * No worry about that, if we cannot, we leave the file and the virtual machine will
							 * delete it on exit. */
						}
	            	}
		            // show the error to the user
	            	elmView.showErrorMessage("Unable to generate the XES log. \n" 
							+ "A problem with the system temporary-file directory occurred. \n" 
							+ "Please try again.");
					return logGenerated; // false
	            }
				// here we have a temp file with the editor content in it
				// ask the parseCSV method to parse the CSV file and generate the XES log, managing possible errors
				logGenerated = parseCSV(csvTempFile.toString(), eventLogEditor.getTitle());
				
				// at this point we do not need anymore the temp file, so we delete it
	            if (csvTempFile != null) {		            	
					try {
						Files.delete(csvTempFile);
					} catch (IOException e) {
						/* we tried to delete the temp file that we created, but we are not able to.
						 * No worry about that, if we cannot, we leave the file and the virtual machine will
						 * delete it on exit. */
					}
            	}
			} else { // if the frame is not an event log editor
				elmView.showWarningMessage("The selected document is not an Event Log Editor and therefore there is no text to analyse.");
			}
		} else { // if no internal frames is selected 
			elmView.showInfoMessage("No documents selected. You first need to create a new CSV document or open one.");
		}	
		// return true if we successfully generated the log from the csv document, false otherwise
		return logGenerated;
    }
    
    /**
     * This method manages the parsing process of the target file, from its
     * CSV content to the XES version, then from XES to the XML format to show
     * the generated content inside an <code>EventLogEditor</code>.
     * 
     * <p>If some warnings are recorded during the parsing, the method will show
     * a dialog with them to the user.
     * 
     * @param filePath the path of the file to parse
     * @param frameTitle the title for the frame, usually it is the name of the file or the name of the editor inside the application where the content resides.
     * 
     * @return <code>true</code> if we successfully parse the file to XES and open an editor with its content in XML format, <code>false</code> otherwise.
     * 
     * @see EventLogEditor
     */
    protected boolean parseCSV(String filePath, String frameTitle) {
    	// check if at the end the file has been parsed or not 
    	boolean fileParsed = false;
    	// get the csv to xes parser
    	CsvXesParser csvParser = new CsvXesParser();
    	// if the selected parser can parse the target file
		if (csvParser.canParse(new File(filePath))) {   					
			// try to read from the file all its content  
        	// try-with-resources Statement (close the resource automatically at the end in any case)
			try (InputStream fileStream = new FileInputStream(filePath)) { 
				// start a busy state
				elmView.setBusyState(true);
				// read all the logs from the csv file
				List<XLog> logList;
				try {
					logList = csvParser.parse(fileStream);  						        	
				} catch (Exception e) {
					// if an error occurred while parsing the document
					elmView.setBusyState(false);
					elmView.showErrorMessage("Unable to parse the document. The document is no a valid CSV. \n"
										   + "Please correct the document and try again.");
					return fileParsed; // false 
				}
				elmView.setBusyState(false);
		        // if no logs have been read
		        if (logList.isEmpty()) {
		        	// show the error from the parser
		        	elmView.showErrorMessage(csvParser.getErrors());
		        } else {	
		        	int logIndex = 0;
		        	boolean allGenerated = true;
		        	// get the xes to xml serializer
		        	XesXmlSerializer xesSerializer = new XesXmlSerializer(); 
		        	// for each log in the csv file
		        	for (XLog log : logList) {	
		        		logIndex++;
		        		String logName = "Log";
		        		// create the name for the frame based on the log number
		        		if (logList.size() > 1) {												
		        			logName = logName + "-" + logIndex;
						}
		        		// get an output stream
		        		// try-with-resources Statement (close the resource automatically at the end in any case)
		    			try (ByteArrayOutputStream outputStream = new ByteArrayOutputStream()) {					    				               			        	
		    				// start a busy state
		    				elmView.setBusyState(true);
		    				// serialise the xes log in xml format on the output stream
		    				xesSerializer.serialize(log, outputStream);
                        	elmView.setBusyState(false);	
                        	// open an input stream over the generated output
				        	// try-with-resources Statement 
				        	try (InputStream inputStream = new ByteArrayInputStream(outputStream.toByteArray());
				        			BufferedReader inputReader = new BufferedReader(new InputStreamReader(inputStream))) { 
				        		// create the frame title
	        					String editorTitle = logName + " generated from " + frameTitle;
	        					if (editorTitle.endsWith("*")) {
	        						editorTitle = editorTitle.substring(0,editorTitle.length()-1);
	        					}
	        					/* create a new event log editor set with the target file name, the target file content,
	                             * the absolute path as null and its content as modified */
	                        	elmController.newEventLogEditor(editorTitle, inputReader, null, true);	                        	                  	
			    			} catch (IOException e) {
			    				allGenerated = false;
	            			    elmView.showErrorMessage("Unable to display the "+logName+" from the CSV document.");
	        				}
		    			} catch (IOException e) {
	    					elmView.setBusyState(false);
	    					allGenerated = false;
	        			    elmView.showErrorMessage("Unable to serialize the "+logName+" in a XML format from the CSV document.");
	    				}
		        	}
		        	// at the end of the serialisation process of all the logs into xml (written at the end in a dedicated editor)
		        	// if everything went right, we set the parsing result to true, false otherwise
		        	fileParsed = allGenerated;
		        	// if during the parsing some errors occurred, we show them to the user in a dialog
		        	if (csvParser.getWarnings().length() > 0) {
                		String dialogTitle = "Warning Report for " + frameTitle;
	    	    		String dialogContent = "These lines contain more values than defined in their respective column line: \n\n" + csvParser.getWarnings();
						// set a text area with the message in it
						JTextArea dialogTextArea = new JTextArea(dialogContent);
						// create a scrollable panel that contains the text area
						JScrollPane dialogScrollPane = new JScrollPane(dialogTextArea);  
						// set the text area
						dialogTextArea.setLineWrap(false);  
						dialogTextArea.setWrapStyleWord(true); 
						dialogTextArea.setEditable(false);
						dialogTextArea.setMargin(new Insets(2,4,2,4));
						// set the panel dimension
						dialogScrollPane.setPreferredSize(new Dimension(550, 250));
						// set title and options
		    			Object[] options = {"Ok", "Show on Editor"};
		    			// ask the View to show the option dialog
		    	    	int choice = elmView.showOptionDialog(dialogScrollPane, dialogTitle, JOptionPane.WARNING_MESSAGE, options, options[0]);
		    	    	if (choice == 1) { // if the user select Show on Editor	
		    	    		// try-with-resources Statement
		    	    		try (InputStream stringStream = new ByteArrayInputStream(dialogContent.getBytes(StandardCharsets.UTF_8));
		    	    				BufferedReader stringReader = new BufferedReader(new InputStreamReader(stringStream))) {
    		    				// show on a editor the warning list
		    	    			elmController.newEventLogEditor(dialogTitle, stringReader, null, true); 
		    	    		} catch (IOException e) {
	            			    elmView.showErrorMessage("Unable to display the Warning Report.");
	        				}                            		    	    			    		
		    			}
                	} 		        	   			        	    			        	
		        }                                	   			        	        			        	      			                	        
			} catch(Exception e) {
				elmView.setBusyState(false);
			    elmView.showErrorMessage("Unable to parse the CSV document. Please try again.");
			}
		} else {
			elmView.showErrorMessage("Unable to parse the CSV document. Please try again.");
		}
		// return the result of the parsing, false if one or more logs have been not imported successfully
		return fileParsed;
    }
    
    /**
	 * This method first generates a set of strings representing all the possible combination
	 * of a set of elements, passed using an infix expression that defines the possible 
	 * structure of the generated strings. Then shows to the user the generated set and asks
	 * if he wants to generate from them a XES log, a CSV document or discard them. After this
	 * selection, the method implements the respective operation for the generation of the
	 * requested document.
	 *  
	 * <p>The expression is composed by characters and strings for the definition of
	 * elements, the operator + to define the sum of two set of elements,
	 * the operator | to define the interleaving between two set of elements and
	 * spaces between all of these to define elements composed by multiple characters.
	 * 
	 * <p>Here an example of a possible expression: a b c + ( ef g | 123 4 5). The
	 * element "ef" and the element "123" are a single block, while for example "4" and "5" are
	 * two different element of the same set, because they are divided by a space. Note that
	 * operators define also the division between different sets, as they are operation between
	 * them.
     * 
     * @param exp the expression in infix notation that defines the possible combination its elements, using the standard the standard defined above.
     */
    public void generateEventLog(String exp) {
    	// get the event log generator that will manage the process
    	EventLogGenerator eventLog = new EventLogGenerator();
    	// get the array for the set of generated traces
    	String[] eventTraces;
    	// try to generate the traces from the passed expression
    	try {
    		// set a busy state
    		elmView.setBusyState(true);
    		// generate the traces from the expression
    		eventTraces = eventLog.generateTraces(exp);
    		elmView.setBusyState(false);	
		} catch (Exception e) {
			elmView.setBusyState(false);
			// the generator found an error in the passed expression, then show the error to the user
			elmView.showErrorMessage(e.getMessage());
			return;
		} 
    	String dialogTitle = "Traces generated from expression: "+ exp;
    	// generate a list of strings, each for a generated trace
    	StringBuilder eventTraceList = new StringBuilder();
    	for (String eventTrace : eventTraces) {
    		eventTraceList.append(eventTrace).append("\n");
    	}
		String dialogContent = eventTraceList.toString();
		// set a text area with the message in it
		JTextArea dialogTextArea = new JTextArea(dialogContent);
		// create a scrollable panel that contains the text area
		JScrollPane dialogScrollPane = new JScrollPane(dialogTextArea);  
		// set the text area
		dialogTextArea.setLineWrap(false);  
		dialogTextArea.setWrapStyleWord(true); 
		dialogTextArea.setEditable(false);
		dialogTextArea.setMargin(new Insets(2,4,2,4));
		// set the panel dimension
		dialogScrollPane.setPreferredSize(new Dimension(550, 250));
		// set title and options
		Object[] options = {"Generate XES log", "Generate CSV file", "Cancel"};
		// ask the View to show the option dialog
    	int choice = elmView.showOptionDialog(dialogScrollPane, dialogTitle, JOptionPane.INFORMATION_MESSAGE, options, options[0]);
    	if (choice == 0) { // if the user select Generate XES log
    		// set a busy state
    		elmView.setBusyState(true);
    		// get the factory for generating XES log elements
    		XFactoryBufferedImpl xesFactory = new XFactoryBufferedImpl();    		
    		// create the log
    		XLog log = xesFactory.createLog();
    		// list to take trace of all generated XID for unique values
    		HashSet<XID> xidList = new HashSet<XID>();
    		// get the current timestamp
    		long timestamp = System.currentTimeMillis();
    		// set the time difference between events
    		int seconds = elmModel.getEventTimestampDistance();
    		// for each generated trace
    		for (String eventTrace : eventTraces) {      
    			// create an attribute map for the trace
        		XAttributeMap traceMap = xesFactory.createAttributeMap();
        		// get an unique id for the trace inside the log        		
        		XID xid = new XID(UUID.randomUUID());
        		// if this XID has already been generated for a trace in this log
        		while (xidList.contains(xid)) {
        			// get another XID and try to check again
        			xid = new XID(UUID.randomUUID());
        		}
        		// add the unique XID to the xidList
        		xidList.add(xid);
        		// add the id to the attribute map
				XAttributeID traceXIDAttribute = xesFactory.createAttributeID("identity:id", xid, null);				
				traceMap.put("identity:id", traceXIDAttribute);
				// create a trace with the attribute map just set
				XTrace trace = xesFactory.createTrace(traceMap);
				// add the trace to the log
				log.add(trace);
				// get the event names from the generated trace
        		String[] events = eventTrace.split("\\s+");
        		// for every event name in the generated trace
        		for (String eventName : events) {
        			// create an attribute map
        			XAttributeMap eventMap = xesFactory.createAttributeMap(); 		
        			// add the attribute name to the map
            		XAttributeLiteral eventNameAttribute = xesFactory.createAttributeLiteral("concept:name", eventName, XConceptExtension.instance());
            		eventMap.put(eventNameAttribute.getKey(), eventNameAttribute);
            		// add the attribute timestamp to the map, adding seconds to its execution
            		timestamp = (timestamp + (seconds * 1000));
            		XAttributeTimestamp eventTimestampAttribute = xesFactory.createAttributeTimestamp("time:timestamp", timestamp, XTimeExtension.instance());
                    eventMap.put(eventTimestampAttribute.getKey(), eventTimestampAttribute);
            		// create the event with the attribute map just set
            		XEvent event = xesFactory.createEvent(eventMap);
            		// add the event to the trace
    	        	trace.add(event);
        		}
        	}   
    		elmView.setBusyState(false);
    		// get the serializer for displaying the log in a XML format
    		XesXmlSerializer xesSerializer = new XesXmlSerializer(); 
    		// get an output for the serialisation
    		// try-with-resources Statement (close the resource automatically at the end in any case)
			try (ByteArrayOutputStream outputStream = new ByteArrayOutputStream()) {					    				               			        	
				// set the busy state
				elmView.setBusyState(true);	
				// serialise the log in the output stream
				xesSerializer.serialize(log, outputStream);
            	elmView.setBusyState(false);
            	// open an input stream over the output
	        	// try-with-resources Statement (close the resource automatically at the end in any case)
	        	try (InputStream inputStream = new ByteArrayInputStream(outputStream.toByteArray());
	        			BufferedReader inputReader = new BufferedReader(new InputStreamReader(inputStream))) { 
					/* create a new event log editor set with the target file name, the target file content,
                     * the absolute path as null and its content as modified */
                	elmController.newEventLogEditor("Generated XES Log", inputReader, null, true);	                        	                  	
    			} catch (IOException e) {
    			    elmView.showErrorMessage("Unable to display the generated XES Log.");
				}
			} catch (IOException e) {
				elmView.setBusyState(false);
			    elmView.showErrorMessage("Unable to display the generated XES Log.");
			}    		                            		    	    			    		
		} else if (choice == 1) { // if the user select Generate CSV file
			// set a busy state
    		elmView.setBusyState(true);
			int traceIndex = 0;
			// start transforming the list of traces in a document with CSV format definition
			StringBuilder csv = new StringBuilder();
			// add the column line
			csv.append("trace:id,concept:name,time:timestamp").append("\n");
    		// get the current timestamp
    		long timestamp = System.currentTimeMillis();
    		// set the time difference between events
    		int seconds = elmModel.getEventTimestampDistance();
    		// get the xsDateTimeConverter to convert timestamp to XES timestamp
    		XsDateTimeConversionJava7 xsDateTimeConverter = new XsDateTimeConversionJava7();
			// for every trace in the generated set of traces
        	for (String eventTrace : eventTraces) {
        		traceIndex++;
        		// get the event names
        		String[] events = eventTrace.split("\\s+");
        		// for each event name
        		for (String event : events) {
            		// add seconds to this event execution
            		timestamp = (timestamp + (seconds * 1000));
            		// calculate the xsDateTime to use as value for the XES timestamp         		
            		String xsDateTime = xsDateTimeConverter.format(new Date(timestamp));
        			// add to the document the id of the trace and the value of the event attributes
        			csv.append(traceIndex).append(",").append(event).append(",").append(xsDateTime).append("\n");
        		}
        	}
        	elmView.setBusyState(false);
        	// open an input stream over the generated document
        	// try-with-resources Statement
    		try (InputStream stringStream = new ByteArrayInputStream(csv.toString().getBytes(StandardCharsets.UTF_8));
    				BufferedReader stringReader = new BufferedReader(new InputStreamReader(stringStream))) {
				/* create a new event log editor set with the target file name, the target file content,
                 * the absolute path as null and its content as modified */
    			elmController.newEventLogEditor("Generated CSV", stringReader, null, true); 
    		} catch (IOException e) {
			    elmView.showErrorMessage("Unable to display the generated CSV.");
			}        	
		}    	
    }
    
    
       
    /*
     * INNER CLASSES
     */     
    
    /**
     * This inner class implements <code>ErrorHandler</code> for creating 
     * a specific error handler for the <code>SchemaFactory</code> object.
     * 
     * <p>It will be used by the factory to manage <code>SAXParseException</code>
     * exceptions that may occur during the creation of the <code>Schema</code> object
     * ({@link EventLogManagerController#validateEventLog() validateEventLog()} method).
     * 
     * <p>This class will report to the user warnings and errors that may occur during the
     * creation of the <code>Schema</code> object and stop the validation process if an error
     * occurs, throwing a <code>SAXException</code>.
     * 
     * @author Francesco Santeramo (eventlogmanager@francescosanteramo.com)
     * 
     * @see ErrorHandler
     * @see SchemaFactory
     * @see Schema
     * @see SAXException
     */
    class SchemaFactoryErrorHandler implements ErrorHandler {

    	/**
    	 * Receive notification of a warning.
    	 * 
    	 * <p>The method asks to the View to show a Warning dialog with the warning information in it. 
    	 * 
    	 * @param exception the warning information encapsulated in a SAX parse exception
    	 * 
    	 * @see SAXParseException
    	 * @see SAXException
    	 */
    	@Override
        public void warning(SAXParseException exception) {
    		StringBuffer warningMessage = new StringBuffer();
    		// add the text
    		warningMessage.append("Warning information (the process will normally continue after this message):");
    		// add the warning info (line, column, message) 
    		addInfo(warningMessage, exception);   
    		// show the warning message to the user, then continue the process
    		elmView.showWarningMessage(warningMessage.toString());
        }

    	/**
    	 * Receive notification of a recoverable error. 
    	 * 
    	 * <p>The method asks to the View to show an Error dialog with the error information in it
    	 * and then throws a <code>SAXParseException</code> to tell to the caller method that it has to stop.
    	 * 
    	 * @param exception the error information encapsulated in a SAX parse exception.
    	 * 
    	 * @throws SAXException this method takes and manages the exception received and then throws it to the caller
    	 * 
    	 * @see SAXParseException
    	 * @see SAXException
    	 */
    	@Override
        public void error(SAXParseException exception) throws SAXException {
    		StringBuffer errorMessage = new StringBuffer();  
    		// add the text
    		errorMessage.append("Error found in the XES Schema content. Please fix the problem and try again.");
    		// add the error info (line, column, message) 
    		addInfo(errorMessage, exception);
    		// show the error message to the user, then stop the process by throwing an exception
    		elmView.showErrorMessage(errorMessage.toString());
        	throw exception;
    	}

    	/**
    	 * Receive notification of a non-recoverable error. 
    	 * 
    	 * <p>The method asks to the View to show an Error dialog with the error information in it
    	 * and then throws a <code>SAXParseException</code> to tell to the caller method that it has to stop.
    	 * 
    	 * @param exception the error information encapsulated in a SAX parse exception.
    	 * 
    	 * @throws SAXException this method takes and manages the exception received and then throws it to the caller
    	 * 
    	 * @see SAXParseException
    	 * @see SAXException
    	 */
    	@Override
        public void fatalError(SAXParseException exception) throws SAXException {
    		StringBuffer fatalErrorMessage = new StringBuffer(); 
    		// add the text
    		fatalErrorMessage.append("Error found in the XES Schema content. Please fix the problem and try again.");
    		// add the fatal error info (line, column, message) 
    		addInfo(fatalErrorMessage, exception);    		
    		// show the error message to the user, then stop the process by throwing an exception
    		elmView.showErrorMessage(fatalErrorMessage.toString());
        	throw exception;
        }
    	
    	/**
    	 * This method reads the line number, column number and message of the passed 
    	 * <code>SAXParseException</code> and then append this information to the passed
    	 * <code>StringBuffer</code>.
    	 * 
    	 * @param message the <code>StringBuffer</code> where to append the information
    	 * @param exception the <code>SAXParseException</code> where to read the information
    	 * 
    	 * @see StringBuffer
    	 * @see SAXParseException
    	 */
    	private void addInfo(StringBuffer message, SAXParseException exception) {
    		String newline = System.getProperty("line.separator");
    		message.append(newline);
    		message.append("  Line number: " + exception.getLineNumber());
    		message.append(newline);
    		message.append("  Column number: " + exception.getColumnNumber());
    		message.append(newline);
    		message.append("  Message: " + exception.getMessage());
        }

    }
    
    /**
     * This inner class extends <code>DefaultHandler</code> for creating 
     * a specific content handler for the <code>SAXParser</code> object.
     * 
     * <p>It will be used by the parser to manage <code>SAXParseException</code>
     * exceptions that may occur during the parsing of the <code>EventLogEditor</code> content 
     * ({@link EventLogManagerController#validateEventLog() validateEventLog()} method).
     * 
     * <p>This class will log into a <code>StringBuffer</code> object warnings and errors
     * that may occur during the parsing and stop the entire parsing process if a fatal error
     * occurs, throwing a <code>SAXException</code>. 
     * 
     * <p>This class is also used by the {@link EditorManager} in the 
     * {@link EditorManager#importDocument(String) importDocument(String)} to report warnings and errors
     * that may occur during the parsing of an XHTML document to import. 
     * 
     * <p>The {@link EditorManager} will use this class of the Analysis Manager in the moments it needs
     * to check a Wellformed-ness of a document before importing it. 
     * 
     * @author Francesco Santeramo (eventlogmanager@francescosanteramo.com)
     * 
     * @see DefaultHandler
     * @see SAXParser
     * @see SAXParseException
     * @see SAXException
     * @see EventLogEditor
     * @see StringBuffer
     */
    class ParserContentHandler extends DefaultHandler {

    	/*
    	 * VARIABLES
    	 */
    	
    	/**
    	 * This variable records the element name of the block that
    	 * the parser is checking at a certain moment of the process.
    	 * 
    	 * <p>If a <code>SAXParseException</code> occurs, this is the 
    	 * element name of the block that generates the exception.
    	 * 
    	 * @see SAXParseException
    	 */
    	private String element; 
    	
    	/**
    	 * The system line separator
    	 */
		private String newline;
		
		/**
		 * This variable records warnings and errors from the parsing process.
		 * 
		 * @see StringBuffer
		 */
		private StringBuffer parserLog;	
		
		/**
		 * This variable records if an error occurred during the parsing or not.
		 * 
		 * <p>If <code>true</code>, nothing happened or at least one or more warnings have
		 * been generated; if <code>false</code>, at least one error or a fatal error have
		 * been generated. 
		 */
		private boolean isValid;
		
		
		/*
		 * CONSTRUCTOR
		 */
		
        /**
         * Creates a new <code>ParserContentHandler</code>, calling the 
         * superclass constructor and then initialising the class variables.
         */
        public ParserContentHandler() {
        	// call the superclass constructor
        	super();
        	// initialise class variables
        	element = "";       	
        	newline = System.getProperty("line.separator");
        	parserLog = new StringBuffer();
        	isValid = true;
        } 
        
        
    	/*
    	 * METHODS
    	 */
        
    	/**
    	 * This method returns the <code>StringBuffer</code> {@link #parserLog}
    	 * that contains the log of the parsing process.
    	 * 
    	 * @return the <code>parserLog</code> component.
    	 */
        public StringBuffer getParserLog() {
            return parserLog;
        }
        
    	/**
    	 * This method returns the value of the {@link #isValid} variable.
    	 * that records if an error occurred during the parsing or not.
    	 * 
    	 * @return the value of the <code>isValid</code> variable.
    	 */
        public boolean isDocumentValid() {
            return isValid;
        }

        /**
         * Receive notification of the start of an element. 
         * 
         * <p>When the parser start to process a new element, this method sets the
         * the {@link #element} variable with the value of the parameter <code>localName</code>
         * if this is not null or empty, with the value of the parameter <code>qName</code> otherwise.
         * 
         * <p>On this way, the {@link #element} variable will always contain the name of the element
         * that the parser is processing in a certain moment.
         * 
         * @param uri the Namespace URI, or the empty string if the element has no Namespace URI or if Namespace processing is not being performed
		 * @param localName the local name (without prefix), or the empty string if Namespace processing is not being performed
		 * @param qName the qualified name (with prefix), or the empty string if qualified names are not available
		 * @param attributes the attributes attached to the element. If there are no attributes, it shall be an empty Attributes object
		 *  
    	 * @see SAXException
         */
        @Override
        public void startElement(String uri, String localName, String qName, Attributes attributes) {
        	// use localName or qName to set the element name that the parser is processing
            if ((localName != null) && (!localName.isEmpty())) {
            	element = localName;
            } else {
            	element = qName;
            }                
        }

        /**
    	 * Receive notification of a parser warning. 
    	 * 
    	 * <p>The method appends to the {@link #parserLog} the warning information,
    	 * with details like the element that was processed when the exception occurred,
    	 * the line number, the column number and the warning message. 
    	 * 
    	 * @param exception the warning information encoded as an exception
    	 * 
    	 * @see SAXParseException
    	 * @see SAXException
    	 */
        @Override
        public void warning(SAXParseException exception) {
        	parserLog.append(newline);
        	if (!element.isEmpty()) {
        		parserLog.append("Warning on \"" + element + "\"");
        	} else {
        		parserLog.append("Warning");
        	}        	
        	addInfo(parserLog, exception);  
        	parserLog.append(newline);
        }

        /**
    	 * Receive notification of a recoverable parser error. 
    	 * 
    	 * <p>The method appends to the {@link #parserLog} the error information,
    	 * with details like the element that was processed when the exception occurred,
    	 * the line number, the column number and the error message. 
    	 * 
    	 * @param exception the error information encoded as an exception
    	 * 
    	 * @see SAXParseException
    	 * @see SAXException
    	 */
        @Override
        public void error(SAXParseException exception) {
        	// set the validation as false
        	isValid = false;
        	// append the error to the result
        	parserLog.append(newline);
        	if (!element.isEmpty()) {
        		parserLog.append("Error on \"" + element + "\"");
        	} else {
        		parserLog.append("Error");
        	}        	
        	addInfo(parserLog, exception);
        	parserLog.append(newline);        	
        }

        /**
    	 * Report a fatal XML parsing error. 
    	 * 
    	 * <p>The method asks to the View to show an Error dialog with the error information in it
    	 * and then throws a <code>SAXParseException</code> to tell to the caller method that it has to stop.
    	 * 
    	 * @param exception the error information encoded as an exception
    	 * 
    	 * @throws SAXException this method takes and manages the exception received and then throws it to the caller
     	 * 
    	 * @see SAXParseException
    	 * @see SAXException
    	 */
        @Override
        public void fatalError(SAXParseException exception) throws SAXException { 
        	// set the validation as false
        	isValid = false;
        	// set the error
    		StringBuffer fatalErrorMessage = new StringBuffer();    		
    		// add the text
        	if (!element.isEmpty()) {
        		fatalErrorMessage.append("Fatal Error on \"" + element + "\". Unable to validate the document.");
        	} else {
        		fatalErrorMessage.append("Fatal Error. Unable to validate the document.");
        	}    		
    		// add the fatal error info (line, column, message) 
    		addInfo(fatalErrorMessage, exception); 
    		// show the error message to the user, then stop the process by throwing an exception
    		elmView.showErrorMessage(fatalErrorMessage.toString());
            throw exception;
        }
    	
    	/**
    	 * This method reads the line number, column number and message of the passed 
    	 * <code>SAXParseException</code> and then append this information to the passed
    	 * <code>StringBuffer</code>.
    	 * 
    	 * @param message the <code>StringBuffer</code> where to append the information
    	 * @param exception the <code>SAXParseException</code> where to read the information
    	 * 
    	 * @see StringBuffer
    	 * @see SAXParseException
    	 */
    	private void addInfo(StringBuffer message, SAXParseException exception) {
    		String newline = System.getProperty("line.separator");
    		message.append(newline);
    		message.append("  Line number: " + exception.getLineNumber());
    		message.append(newline);
    		message.append("  Column number: " + exception.getColumnNumber());
    		message.append(newline);
    		message.append("  Message: " + exception.getMessage());
        }
    }
    
    /**
     * This class implements an alphabet sequence generator, extending 
     * <code>AbstractIterator</code>.
     * 
     * <p>For every call to the {@link #computeNext()} method, the object 
     * of this class return the next alphabet element of an infinite list
     * of elements in alphabet order.
     * 
     * <p>The list is defined infinite because every of its elements can be
     * composed by an infinite number of characters, not just one. Therefore 
     * when the generator reaches the last element in alphabet order of a list
     * of one character(i.e "a,b,c,...,z"), it starts again using
     * this time two characters instead of one (i.e "aa,ab,ac,...,az"), moving
     * both the second character as the first one, when the second reaches the
     * end of possible alphabet values, and so on with four, five, etc. 
     * 
     * @author Francesco Santeramo (eventlogmanager@francescosanteramo.com)
     */
    class AlphabetSequenceGenerator extends AbstractIterator<String> {
    	
    	/**
    	 * The position in the list of the last element generated
    	 */
        private int now;
        
        /**
         * The prefix used for the last element generated
         */
        private String prefix;
        
        /**
         * The list of all the possible characters that the generator can use for the sequence
         */
        private char[] vs;
        
        /**
         * Constructs an <code>AlphabetSequenceGenerator</code> ready to be used.
         */
        public AlphabetSequenceGenerator() {
        	now = -1;
        	prefix = "";
            vs = new char['z' - 'a' + 1];
            for(char i='a'; i<='z';i++) vs[i - 'a'] = i;
        }

        /**
         * Generates the prefix basing of the one that we were using for 
         * the preceding element.
         *  
         * @param prefix the prefix used for the last element generated
         * 
         *  @return the prefix for the next element that we will generate
         */
        private String fixPrefix(String prefix) {
            if (prefix.length() == 0) return Character.toString(vs[0]);
            int last = prefix.length() - 1;
            char next = (char) (prefix.charAt(last) + 1);
            String sprefix = prefix.substring(0, last);
            return next - vs[0] == vs.length ? fixPrefix(sprefix) + vs[0] : sprefix + next;
        }

        /**
         * Generates the next element in alphabet order of an infinite
         * list of elements, starting from "a" and using infinite characters for 
         * defining a single element (i.e. elements go from a single character 
         * composition, "a" for example, to an infinite one, "cdf" for example
         * using three characters).
         */
        @Override 
        protected String computeNext() {
        	// calculate if needed the prefix
            if (++now == vs.length) prefix = fixPrefix(prefix);
            now %= vs.length;
            // return the next element in alphabet order
            return new StringBuilder().append(prefix).append(vs[now]).toString();
        }
    }    
}
