/*
 * Event Log Manager
 *   
 *  Copyright (C) 2015 Francesco Santeramo
 *
 *  This file is part of Event Log Manager.
 *
 *  Event Log Manager is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Event Log Manager is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *	
 *  Author:  Francesco Santeramo 
 *  Website: www.francescosanteramo.com
 *  Contact: eventlogmanager@francescosanteramo.com    
 */

package unipi.bpm.elm.santeramo.controller;

import cz.kebrt.html2latex.FatalErrorException;
import cz.kebrt.html2latex.Parser;
import cz.kebrt.html2latex.ParserHandler;

import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.awt.print.PrinterException;
import java.beans.PropertyVetoException;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.StringReader;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.text.MessageFormat;
import java.util.Collection;

import javax.imageio.ImageIO;
import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JInternalFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.JTextPane;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.deckfour.xes.model.XLog;
import org.fife.ui.rsyntaxtextarea.RSyntaxTextArea;
import org.xhtmlrenderer.pdf.ITextRenderer;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import com.lowagie.text.DocumentException;

import unipi.bpm.elm.santeramo.controller.AnalysisManager.ParserContentHandler;
import unipi.bpm.elm.santeramo.model.EventLogManagerModel;
import unipi.bpm.elm.santeramo.view.EventLogEditor;
import unipi.bpm.elm.santeramo.view.EventLogManagerView;
import unipi.bpm.elm.santeramo.view.FootprintMatrixInfo;

/**
 * This class defines the behaviour of all the <code>MenuAreaButton</code> buttons in the
 * application tab menus that are connected to the management of logs, files and editors.
 * 
 * <p>It is used by the Controller that delegates to this class the implementation of the actions
 * connected to these <code>MenuAreaButton</code> buttons.
 * 
 * <p>The Controller receives the user'choice by the relative listener and call the right method
 * of this class.
 * 
 * @author Francesco Santeramo (eventlogmanager@francescosanteramo.com)
 * 
 * @see unipi.bpm.elm.santeramo.view.MenuAreaButton
 * @see EventLogManagerController
 */
public class EditorManager {
	
	/*
	 * THE CONTROLLER, THE VIEW AND THE MODEL REFERENCES
	 */
	
	/**
	 * The <code>EventLogManagerController</code> element 
	 * that represents the Controller of the application
	 * and that is the one that uses this editor manager.
	 * 
	 * @see EventLogManagerController
	 */
	private EventLogManagerController elmController;	
	
	/**
	 * The <code>EventLogManagerView</code> element 
	 * that represents the View of the application managed
	 * by the Controller. 
	 * 
	 * <p>We set this reference to the View managed by the Controller from the
	 * {@link unipi.bpm.elm.santeramo.controller.EventLogManagerController#getView() getView()}
	 * method of the <code>EventLogManagerController</code>.
	 * 
	 * @see EventLogManagerView
	 */
	private EventLogManagerView elmView;	
	
	/**
	 * The <code>EventLogManagerModel</code> element 
	 * that represents the Model of the application managed
	 * by the Controller. 
	 * 
	 * <p>We set this reference to the Model managed by the Controller from the
	 * {@link unipi.bpm.elm.santeramo.controller.EventLogManagerController#getModel() getModel()}
	 * method of the <code>EventLogManagerController</code>.
	 * 
	 * @see EventLogManagerModel
	 */
	private EventLogManagerModel elmModel;

	
	
	/*
	 * CONSTRUCTOR
	 */
	
	/**
	 * Constructs a new <code>EditorManager</code>, with the <code>EventLogManagerController</code> 
	 * passed as parameter associated to it.
	 * 
	 * <p>The passed Controller is the one that uses this <code>EditorManager</code> class. This class
	 * uses this reference to the Controller to communicate with it, calling some of its methods.
	 * 
	 * <p>This class is delegated by the Controller to implement some operations, for this reason 
	 * it is able to call the View and the Model managed by the Controller directly on behalf of the Controller to 
	 * handle this operation correctly.
	 * 
	 * @param elmController the <code>EventLogManagerController</code> object that represents the Controller and uses this <code>EditorManager</code> 
	 * 
	 * @see EventLogManagerView
	 * @see EventLogManagerController
	 * @see EventLogManagerModel
	 */
	public EditorManager(EventLogManagerController elmController) {	
		this.elmController = elmController;
		// the view to use
		elmView = elmController.getView();
		// the model to use
		elmModel = elmController.getModel();
	}
	
	
	
	/*
	 * METHODS
	 */
	
    /**
     * This method asks the View to open a dialog to make the user decide
     * between an empty Event Log or a generation of an Event Log starting
     * from an expression that defines the possible event traces.
     * 
     * <p>If the user selects the first option, the method opens an empty
     * <code>EventLogEditor</code> internal frame; if the user selects the
     * second option, the method first defines the structure of the component
     * that will get the expression from the user, then asks the View to show
     * the created component inside a dialog to get the user's choice. The method
     * also adds the passed <code>ActionListener</code> to the component button
     * to get the user's action when pressed and start processing the expression.
     * 
     * @param listener the {@link ActionListener} listener that will be associated with the the dialog button 
     * 					(the dialog for the generation of event logs from expressions).
     *
     * @see EventLogEditor
     */
    public void newEventLog(ActionListener listener) {    	
    	// retrieve the user's choice
    	String message = "Select how to create a new Event Log:";
    	String title = "Create new Event Log";
    	Object[] options = {"Empty Event Log Editor","Generate new Event Log"};    	
    	// ask the View to show the dialog
    	int choice = elmView.showOptionDialog(message, title, JOptionPane.QUESTION_MESSAGE, options, options[0]);	
    	
    	// if the user select Empty Event Log Editor
    	if (choice == 0) { 
    		// set the editor with default title, no content and no associated file
			elmController.newEventLogEditor(null, null, null, false);
			
		} else if (choice == 1){ // if the user select to generate a new event log				
			// create the panel
			JPanel expPanel = new JPanel();
	    	// set an empty border around the content
	    	expPanel.setBorder(BorderFactory.createEmptyBorder(12, 5, 0, 5));
			// set the panel layout
	    	expPanel.setLayout(new GridBagLayout());
			// create the constraints to use for positioning the elements
			GridBagConstraints constraints = new GridBagConstraints();
			// set the general constraints
			constraints.fill = GridBagConstraints.HORIZONTAL;
			constraints.weightx = 0.5;				
			
			// put the first label in the panel, setting its position in the layout and insets
			constraints.insets = new Insets(0,3,4,3);
			constraints.gridwidth = 2;
			constraints.gridx = 0;
			constraints.gridy = 0;   
			JLabel expLabel1 = new JLabel("Write an expression that defines possible event traces using");
			expLabel1.setFont(new Font("Verdana", Font.PLAIN, 11));
	    	expPanel.add(expLabel1, constraints);
	    	
	    	// put the second label in the panel, setting its position in the layout and insets
	    	constraints.insets = new Insets(0,3,4,3);
	    	// grid width = 2 still hold
			constraints.gridx = 0;
			constraints.gridy = 1;   
			JLabel expLabel2 = new JLabel("[ A-Z a-z 0-9 ] for event names, [ | + ] as operators, spaces");
			expLabel2.setFont(new Font("Verdana", Font.PLAIN, 11));
	    	expPanel.add(expLabel2, constraints);	
	    	
	    	// put the third label in the panel, setting its position in the layout and insets
	    	constraints.insets = new Insets(0,3,12,3);
	    	// grid width = 2 still hold
			constraints.gridx = 0;
			constraints.gridy = 2;   
			JLabel expLabel3 = new JLabel("to divide events of a set and parentheses to divide operations:");
			expLabel3.setFont(new Font("Verdana", Font.PLAIN, 11));
	    	expPanel.add(expLabel3, constraints);
	    	
	    	// put the text field in the panel, setting its position in the layout and margins
	    	constraints.gridwidth = 1; // now we set a different grid width = 1
	    	// insets still hold
			constraints.gridx = 0;
			constraints.gridy = 3;  
			JTextField expField = new JTextField(15);
			expField.setMargin(new Insets(3,3,3,3));
			expField.setFont(new Font("Verdana", Font.PLAIN, 12));
	    	expPanel.add(expField, constraints);
	    	
	    	// put the button in the panel, setting its position in the layout and margins
	    	// grid width and insets still hold
			constraints.gridx = 1;
			constraints.gridy = 3; 
			JButton expButton = new JButton("Generate");
			expButton.setMargin(new Insets(3,3,3,3));			
			expButton.putClientProperty("field", expField); // set the target field		
			expButton.addActionListener(listener); // set the action listener
			expPanel.add(expButton, constraints);    	
	    	
			// set the panel opaque
	    	expPanel.setOpaque(true);
	    	// ask the View to show the panel inside a dialog
	    	elmView.showDialog("Event Log Generator", expPanel, false);
		}
    }
    
    /**
     * Opens an "Open" dialog to ask the user to select which file to open, then the method tries
     * to open it (.XES, .CSV and .TXT extensions available).
     * 
     * <p>The method first looks if there is already an opened <code>EventLogEditor</code> editor
     * associated with that file (and in this case it sets it as selected to show it to the user, 
     * asking him then if he wants to reload or to open in a new editor the file content, in case
     * another application modified it in the meanwhile). 
     * 
     * <p>Either if the process does not find any editor associated with the target file,
     * or the user selected to reload or to open in a new editor the file content, the method tries 
     * to read the content of the selected file and to open it inside an <code>EventLogEditor</code>
     * (setting its title using the file name, its content using the file content and its source
     * file using the path of the selected file) in the first case, to reload it in the second one,
     * and to open it inside an <code>EventLogEditor</code> (setting its title as a default value,
     * its content using the file content and without source file associated) in the third one.
     * 
     * <p>The method implements the common behaviour of the Open button in a standard application,
     * with management of errors that may occur during the process.
     * 
     * @param selectedFileName the file name to set as selected in the dialog, <code>null</code> for no selection
     * @param isAbsolutePath optional value to set as <code>true</code> to indicate that the selectedFileName is not just
     * 							a file name, but is a complete path to a file and the method should try to open it directly 
     * 							without asking any selection to the user.
     * 
     * @see EventLogEditor
     */
    public void openEventLog(String selectedFileName, boolean... isAbsolutePath) {
    	
    	String selectedSourceFile;
    	boolean openDirectly = (isAbsolutePath.length > 0 && isAbsolutePath[0] == true);
    	// if selectedFileName is a specific file path to open 
    	if (openDirectly) {
    		selectedSourceFile = selectedFileName;
     	} else { // if we do not have a specific file path to open directly, go for the open dialog and user selection   		
           	// set the available file extensions
        	FileNameExtensionFilter xesFilter = new FileNameExtensionFilter("eXtensible Event Stream (*.xes)", "xes");
        	FileNameExtensionFilter csvFilter = new FileNameExtensionFilter("Comma-Separated Values (*.csv)", "csv");
        	FileNameExtensionFilter txtFilter = new FileNameExtensionFilter("Text File (*.txt)", "txt");
        	String selectedFile;
        	// check if we have a selected file to show, blank string otherwise
        	if (selectedFileName == null) {
        		selectedFile = "";
        	} else {
        		selectedFile = selectedFileName;
        	}  
        	// open a dialog to make the user select a file, using filters
        	selectedSourceFile = elmView.showFileSelectionDialog("Open", selectedFile, xesFilter, csvFilter, txtFilter);     		
    	}
    	// check if the user have selected a file, otherwise do nothing
    	if (selectedSourceFile != null) {
    		// create a Path object using the String path
    		Path targetFile = Paths.get(selectedSourceFile);
    		// if the target file exist
    		if (Files.exists(targetFile)) {   
				// here we first check if there is already an opened event log editor associated to this file
				// get all opened internal frames
		        JInternalFrame[] internalFrameList = elmView.getContentArea().getAllFrames();
		        // if the file is already opened 
		        //if the user selects to reload the content, here the opened editor to reload 
		        EventLogEditor targetEventLogEditor = null;
		        // here the choice to open a new editor with only the file content and keep the opened editor as it is
		        boolean openContentOnly = false;
		        // for every internal frame
		        for (JInternalFrame internalFrame: internalFrameList) {  
		        	// check if it is an event log editor
		            if (internalFrame instanceof EventLogEditor) {
		            	// get the editor and the absolute path of the associated file
		            	EventLogEditor eventLogEditor = (EventLogEditor)internalFrame;
		            	String absolutePath = eventLogEditor.getAbsolutePath();
		            	// if it has not an associated file, skip to the next event log editor
		            	if (absolutePath != null) {
		            		// check if the associated file is the same of the one that we are trying to open
	    	            	if (absolutePath.compareTo(selectedSourceFile) == 0) { 
	    	            		/* if we find an opened event log editor associated to the target file, we do not
	    	            		 * need to open a new internal frame but just to set as selected the one already
	    	            		 * opened, to show that there is already an existing event log editor for this file */
	    	            		if (internalFrame.isIcon()) { // if the event log editor is minimised
	    	    	            	try {
	    	    	            		// set the event log editor maximised (and selected)
	    	    	            		internalFrame.setIcon(false);
	    	    					} catch (PropertyVetoException exception) {
	    	    						// just get here the exception to not throw an error, the frame won't just be maximised
	    	    					}  
	    	    	            	/* setting the event log editor as maximised will automatically invokes the 
	    	    					 * internalFrameDeiconified(InternalFrameEvent e) method of the InternalFrameStatusAdapter
	    	    					 * adapter of the internal frame. This method will make all the other step */
	    	    	            } else { // if event log editor is maximised
	    	    	            	try {
	    	    	            		// set the event log editor selected
	    	    	            		internalFrame.setSelected(true);
	    	    	        		} catch (PropertyVetoException exception) {
	    	    	        			// just get here the exception to not throw an error, the frame won't just be selected
	    	    	                } 
	    	    	            }	 
	    	            		// inform the user that there is already an opened editor associated with the file
	    	            		String message = "There is already an opened editor associated with the selected file. \n"
	    	            						+ "If in the meanwhile the file has been modified by another application, and you want \n"
	    	            						+ "to reload its content in the opened editor, select \"Reload Content\" (the content \n"
	    	            						+ "inside the editor before the reload will be lost).\n"
	    	            						+ "Otherwise select \"Open New\" to open a new editor with the file content (the already \n"
	    	            						+ "opened editor will keep its content and the association with the file, the new editor \n"
	    	            						+ "will contain only the file content, without any file association). \n"
	    	            						+ "Select instead \"Do Nothing\" to keep the already opened editor as it is without \n"
	    	            						+ "reloading or opening the file.";
	    	            		// retrieve the user's choice       			
	    		    			String title = "File already opened";
	    		    			Object[] options = {" Reload Content ", "Open New", "Do Nothing"};
	    						int choice = elmView.showOptionDialog(message, title, JOptionPane.WARNING_MESSAGE, options, options[2]);
	    						// check user's decision
	    						if (choice == 0) { // if Reload
	    							// set the target editor to reload and stop the loop 
	    							targetEventLogEditor = eventLogEditor;
		    	            		break;
	    						} else if (choice == 1) { // if Open New
	    							// set the target editor to reload and stop the loop 
	    							targetEventLogEditor = eventLogEditor;
	    							// set to open a new editor with just the file content
	    							openContentOnly = true;
	    							break;
	    	            		}else { // otherwise for Do Nothing or if the user closed the dialog
	    							// clear the file selection and exit the process (we already set the editor as selected, so the user will find it on top)
	    				    		elmView.getFileChooser().setSelectedFile(new File(""));
	    				    		return;
	    						}	    	            		
	    	            	}
		            	}
		            }
		        }
		        // if we have not found any opened event log editor or the user selected reload
		        
	        	// try to read from the file all its content  
	        	// try-with-resources Statement (close the resource automatically at the end in any case)
    			try (InputStream fileStream = new FileInputStream(selectedSourceFile);
    					BufferedReader fileReader = new BufferedReader(new InputStreamReader(fileStream))) {   					
					// if there is not a target editor to reload, it means we are opening a new file
					if (targetEventLogEditor == null) {
                        /* create a new event log editor set with the target file name, the target file content,
                         * the absolute path of the target file and its content as not modified */
                	    elmController.newEventLogEditor(targetFile.getFileName().toString(), fileReader, selectedSourceFile, false);
					} else { // if we have a target editor
						if (openContentOnly) { // if we have to open a new editor with just the content in it
	                        /* create a new event log editor set with default name, the target file content,
                             * no absolute path and its content as modified */
                    	    elmController.newEventLogEditor(null, fileReader, null, true);
						} else { // if we have to reload the file instead
							// set the new content
    						targetEventLogEditor.setTextAreaContent(fileReader);
    						// set the editor content as not modified (because now the editor has the same content of the associated file, so it is saved)
    						if (targetEventLogEditor.isContentModified()) {
    							// if not already set as false, set it
    							targetEventLogEditor.setContentModified(false);
    							// remove the asterisk (if the content is modified, the title has an asterisk at the end)
    				    		String savedTitle = targetEventLogEditor.getTitle();
    				    		savedTitle = savedTitle.substring(0, savedTitle.length()-1);
    				    		elmController.modifyJInternalFrameTitles(targetEventLogEditor, savedTitle);
    			    		}  
    						// set the editor content as not validated (because it is a new content)
    						targetEventLogEditor.setContentValidated(false);
    						// the title and the associated file are already set, then no needs to modify them
						}    						
					} 
					// add the path of the opened file to the recent list in the Model					
					elmModel.addRecentLog(selectedSourceFile);
    			} catch(IOException e) {
    			    elmView.showErrorMessage("Unable to read the file. Please try again.");
    			}  		        
    		} else if (Files.notExists(targetFile)) { // if the target does not file exist	    			
    			if (openDirectly) {
    				elmView.showWarningMessage("Unable to open the file \""+targetFile.getFileName().toString()+"\", the file does not exist anymore.");
    			} else {
    				elmView.showWarningMessage("Unable to open the file \""+targetFile.getFileName().toString()+"\", the file does not exist. \n" 
							+ "Check the file name and try again.");	
    				// star a new open process to select another file, with the last selected file shown
    				openEventLog(targetFile.getFileName().toString());
    			}
    		} else { //if the process cannot determine if the file exists or not, because it does not have access to the file
    			if (openDirectly) {
    				elmView.showWarningMessage("Unable to open the file \""+targetFile.getFileName().toString()+"\", the program does not have access to the file.");	
    			} else {
        			elmView.showWarningMessage("Unable to open the file \""+targetFile.getFileName().toString()+"\", the program does not have access to the file. \n" 
							+ "Please select another file or change file permissions.");	
        			// star a new open process to select another file, with the last selected file shown
        			openEventLog(targetFile.getFileName().toString());
    			}
    		}
    		// if we used the open dialog
    		if (!openDirectly) {
    			// at the end of the opening process (also the recursive ones), clear the file selection
        		elmView.getFileChooser().setSelectedFile(new File(""));
    		}   		
    	}		
    }
    
    /**
     * Closes the <code>JInternalFrame</code> passed as parameter, if one is passed,
     * or the selected one inside the 
     * {@link unipi.bpm.elm.santeramo.view.EventLogManagerView#contentArea contentArea},
     * if a frame is set as selected.
     * 
     * <p>This method first checks if the internal frame is an <code>EventLogEditor</code>, in that
     * case it looks if this frame has been modified by the user or not: in the first case, it 
     * asks if the user wants to save or not the document before closing; in the second one, checks if the
     * associated file still exists and decides to close or to ask to save again the document depending on this.
     * 
     * <p>If the internal frame is not an <code>EventLogEditor</code>, the method then checks if is a <code>FootprintMatrixInfo</code>.
     * If it is, the method checks if the document has already been exported or not: in the first case, it checks if the associated file
     * still exists and decides to close or to ask to export again the document depending on this; in the second one, it asks directly 
     * if the user wants to export or not the document before closing.
     * 
     * <p>If the frame to close is not an <code>FootprintMatrixInfo</code> either, the method closes directly the internal frame, because
     * there is not any specific action to do before closing for a generic internal frame.
     * 
     * @param internalFrame the internal frame to close, <code>null</code> to close the selected internal frame inside the
     *  {@link unipi.bpm.elm.santeramo.view.EventLogManagerView#contentArea contentArea}
     *  
     * @return <code>true</code> if the closing operation successfully close the internal frame, <code>false</code> otherwise. 
     * 
     * @see JInternalFrame
     * @see EventLogEditor
     * @see FootprintMatrixInfo
     */
    public boolean closeInternalFrame(JInternalFrame internalFrame) {
    	// take the selected internal frame to close
    	JInternalFrame selectedInternalFrame;  
    	boolean fileClosed = false;
    	if (internalFrame != null) { // try to close the passed internal frame
    		selectedInternalFrame = internalFrame;
    	} else { // or close the selected internal frame inside the contentArea if no frame has been passed
    		selectedInternalFrame = elmView.getContentArea().getSelectedFrame();
    	}
		// if an internal frame is selected
		if (selectedInternalFrame != null) {
			// if the frame is an EventLogEditor
			if (selectedInternalFrame instanceof EventLogEditor) {
				EventLogEditor eventLogEditor = (EventLogEditor)selectedInternalFrame;
				// default settings
				String location = null;  
				String message = "Do you want to save the document before closing?";
				// get the absolute path of the file associated with this editor
				String absolutePath = eventLogEditor.getAbsolutePath();
				// if the editor content has not been modified by the user, check the associated file and then decide
				if (!eventLogEditor.isContentModified()) {
					if ((absolutePath == null)||(absolutePath.equals(""))) {
						// this is a new empty event log, so we can just close it
						eventLogEditor.dispose();
						fileClosed = true;
					} else {
						// create a Path object using the String path
						Path targetFile = Paths.get(absolutePath);	
						// if the file still exists 
						if (Files.exists(targetFile)) {
							// the editor has no modifications to save and its associated file still exists, so we can close it
							eventLogEditor.dispose();
							fileClosed = true;						
						} else if (Files.notExists(targetFile)) { // if the target file does not longer exist	
							// set a specific message for the save procedure and the name for the file
							message = "The file associated with this editor does not longer exist in your file system. \n" 
    								+ "Do you want to save the document in a new file before closing?";
							location = targetFile.getFileName().toString();
						} else { //if the process cannot determine if the file exists or not, because it does not have access to the file
							// show an error message to the user and then set the saving name for the file
							elmView.showErrorMessage("The application cannot access the file associated with this editor in your file system and check it. \n"
													+ "We advise to save the document in a new file before closing it, to not loose the editor content.");
							location = targetFile.getFileName().toString();
						}
					}
				}
				// if the file is still open, it means we need to save it
				if(!fileClosed) {
					// retrieve the user's choice, using the message decided before	    			
	    			String title = "Save";
	    			// if we have to save in a new location
	    			if (location != null) {
	    				title = title + " As";
	    			}
	    			Object[] options = {"Yes", "No", "   Cancel   "};
					int choice = elmView.showOptionDialog(message, title, JOptionPane.QUESTION_MESSAGE, options, options[0]);
					// check user's decision
					if (choice == 0) { // if YES, call the save method
    					boolean fileSaved;
    					// if we have to save in a new location
    					if (location != null) {
    						fileSaved = saveAsEventLog(location);
    					} else { // otherwise just save in the same location
    						fileSaved = saveEventLog();
    					}	
						if (fileSaved) { // if we have successfully saved the file, then set the editor as closable
							eventLogEditor.dispose();
							fileClosed = true;
						}// if we have not saved the file, then do not close the editor because its content has not been saved  									
					} else if (choice == 1) { // if NO, just set the internal frame as closable, without saving it
						eventLogEditor.dispose();
						fileClosed = true;
					}
					// if the choice is Cancel (2) or the user closed the dialog (-1), do nothing										
				}
			} else if (selectedInternalFrame instanceof FootprintMatrixInfo) { // if the frame is a FootprintMatrixInfo			
				FootprintMatrixInfo footprintMatrixInfo = (FootprintMatrixInfo)selectedInternalFrame;
				// default settings
				String location = null;
				String message = "Do you want to export the document before closing? \n" 
								+ "(Export the document in HTML format if you want to open it again in this application)";
				// check if the file has already been exported												
				if (footprintMatrixInfo.isContentExported()) { // if the document has already been exported by the user, check the associated file and then decide
					// create a Path object using the String path
					Path targetFile = Paths.get(footprintMatrixInfo.getAbsolutePath());	   		
					// if the file still exists 
					if (Files.exists(targetFile)) {
						// it has been exported and its associated file still exists, so we can close it
						footprintMatrixInfo.dispose();
						fileClosed = true;	
					} else if (Files.notExists(targetFile)) { // if the target file does not longer exist	
						// set a specific message for the export procedure and the name for the file    			
			    		message = "The HTML file associated with this document does not longer exist in your file system. \n" 
			    							+ "Do you want to export the document in a new HTML file before closing? \n"
			    							+ "(Export the document in HTML format if you want to open it again in this application)";
			    		location = targetFile.getFileName().toString();
					} else { //if the process cannot determine if the file exists or not, because it does not have access to the file
						// show an error message to the user, then set a specific message for the export procedure and the export name for the file
						elmView.showErrorMessage("The application cannot access the HTML file associated with this editor in your file system and check it. \n"
									+ "We advise to export the document in a new HTML file before closing it, to not loose the content.");
						message = "Do you want to export the document before closing?";
						location = targetFile.getFileName().toString();
					}
				}	
				// if the file is still open, it means we need to export it
				if (!fileClosed) {					
					// retrieve the user's choice, using the message decided before	        			
	    			String title = "Export As";
	    			Object[] options = {"Yes", "No", "   Cancel   "};
					int choice = elmView.showOptionDialog(message, title, JOptionPane.QUESTION_MESSAGE, options, options[0]);
					// check user's decision
					if (choice == 0) { // if YES, call the export method
						// if we have to export in a new location or in the same is decided by the value of location (null or String)
						boolean fileExported = exportDocument(location);
						if (fileExported) { // if we have successfully exported the file, then set the editor as closable
							selectedInternalFrame.dispose();
							fileClosed = true;
						}// if we have not exported the file, then do not close the editor because its content has not been saved 					
					} else if (choice == 1) { // if NO, just set the internal frame as closable, without exporting it
						selectedInternalFrame.dispose();
						fileClosed = true;
					}
					// if the choice is Cancel (2) or the user closed the dialog (-1), do nothing
				}
    		} else { // for every other type of frame, close them directly		
    			selectedInternalFrame.dispose();
				fileClosed = true;
    		}
		} else { // if no internal frames is selected 
			elmView.showInfoMessage("No documents to close. You first need to create a new Event Log or open one.");
		} 
		return fileClosed; // true if we close the file, false otherwise
    }
    
    /**
     * Saves the content of the selected <code>EventLogEditor</code> editor (that is inside 
     * the {@link unipi.bpm.elm.santeramo.view.EventLogManagerView#contentArea contentArea})
     * in its own file, if an editor is selected.
     * 
     * <p>This file is the source file of the internal frame, set by the "Open" option or the
     * "Save As" option. If no source file has been set before, the method delegates the 
     * {@link #saveAsEventLog(String)} method to save the document.
     * 
     * <p>The method is made for the <code>EventLogEditor</code> frame. In case of using it with other
     * frame types, the method manages this situation with specifics information messages delivered to
     * the user.
     * 
     * <p>The method implements the common behaviour of the Save button in a standard application, with
     * overwrite checking, selected file existence checking and management of errors that may occur
     * during the process.
     * 
     * <p>At the end of the process, if the saving procedure went right, the method updates the title of the
     * <code>EventLogEditor</code>, removing the asterisk at the end of it (if present).
     * 
     * @return <code>true</code> if we successfully save the file, <code>false</code> otherwise.
     * 
     * @see EventLogEditor
     */
    public boolean saveEventLog() {
    	// take the selected internal frame to save
		JInternalFrame selectedInternalFrame = elmView.getContentArea().getSelectedFrame();
		boolean fileSaved = false;
		// if an internal frame is selected
		if (selectedInternalFrame != null) {	
			// if the frame is an EventLogEditor
			if (selectedInternalFrame instanceof EventLogEditor) {
				EventLogEditor eventLogEditor = (EventLogEditor)selectedInternalFrame;
				// get the absolute path of the file associated with this editor
				String absolutePath = eventLogEditor.getAbsolutePath();
				// if this editor does not have an associated file
				if ((absolutePath == null)||(absolutePath.equals(""))) {
					// ask for the saving the "save as" procedure
					return saveAsEventLog(null);
				} else {
					// create a Path object using the String path
		    		Path targetFile = Paths.get(absolutePath);	
		    		// get the target file name
		    		String targetFileName = targetFile.getFileName().toString();
		    		// if the file still exists 
		    		if (Files.exists(targetFile)) {
						// if the editor content has not been modified by the user
						if (!eventLogEditor.isContentModified()) {
							// the editor as a valid associated file and no content to save
							elmView.showInfoMessage("The selected document has already been saved.");
							fileSaved = true;
							return fileSaved;
						}		    			
		    		} else if (Files.notExists(targetFile)) { // if the target file does not longer exist, we create it again
		    			try {
							Files.createFile(targetFile);
						} catch (IOException e) {
							elmView.showErrorMessage("Unable to save the document: error while creating the file. \n" 
														+ "Please select another file where to save.");
							// ask for the saving the "save as" procedure
							return saveAsEventLog(targetFileName);
						}
		    		} else { //if the process cannot determine if the file exists or not, because it does not have access to the file
		    			elmView.showErrorMessage("Unable to save the document. \n" 
		    										+ "The program does not have access to the file associated with this editor. \n"
		    										+ "Please select another file where to save.");
						// ask for the saving the "save as" procedure
						return saveAsEventLog(targetFileName);
		    		}
		    		// write the content of the event log in the target file
	    			// try-with-resources Statement (close the resource automatically at the end in any case)
		    		try (OutputStream fileStream = new FileOutputStream(targetFile.toString());
		    				BufferedWriter fileWriter = new BufferedWriter(new OutputStreamWriter(fileStream))) {
		    			// start writing the file and set the application in a busy state in the meanwhile	    			        
    			        elmView.setBusyState(true);
    			        eventLogEditor.getTextAreaContent(fileWriter);
    			        elmView.setBusyState(false);
		            } catch (IOException e) {
		            	elmView.setBusyState(false);		            	
		    		    elmView.showErrorMessage("Unable to save the document in the selected file. \n"
		    		    							+ "Please select another file where to save.");
		    		    // ask for the saving the "save as" procedure
						return saveAsEventLog(targetFileName);
		            }
		    		if (eventLogEditor.isContentModified()) {
		    			// set the event log editor as saved and remove the asterisk from its title
			    		eventLogEditor.setContentModified(false);
			    		String savedTitle = eventLogEditor.getTitle();
			    		savedTitle = savedTitle.substring(0, savedTitle.length()-1);
			    		elmController.modifyJInternalFrameTitles(eventLogEditor, savedTitle);
		    		}		    		
		    		// we successfully saved the file
		    		fileSaved = true;
				}
			} else if (selectedInternalFrame instanceof FootprintMatrixInfo) { // if the frame is a FootprintMatrixInfo
				elmView.showInfoMessage("The selected document cannot be saved with the \"Save\" procedure. \n" + 
											"Please use the \"Export as\" button to export this document correctly.");
			} else { // for every other type of frame
				elmView.showInfoMessage("The selected document cannot be saved by this application.");
			}
		} else { // if no internal frames is selected 
			elmView.showInfoMessage("No documents to save. You first need to create a new Event Log or open one.");
		}
		return fileSaved; // true if we have successfully saved the file, false otherwise
    }
    
    /**
     * Opens a "Save As" dialog to ask the user to decide where to save the content of
     * the selected <code>EventLogEditor</code> editor (that is inside the
     * {@link unipi.bpm.elm.santeramo.view.EventLogManagerView#contentArea contentArea}),
     * then it tries to save the content in the selected file (.XES, .CSV or .TXT extensions available).
     * 
     * <p>The method is made for the <code>EventLogEditor</code> frame. In case of using it with other
     * frame types, the method manages this situation with specifics information messages delivered to
     * the user.
     * 
     * <p>The method implements the common behaviour of the Save As button in a standard application, with
     * overwrite checking, selected file existence checking and management of errors that may occur
     * during the process.
     * 
     * <p>At the end of the process, if the saving procedure went right, the method updates the path of the
     * <code>EventLogEditor</code> source file and its title, removing the asterisk at the end of it (if present).
     * 
     * @param selectedFileName the file name to set as selected in the dialog, <code>null</code> for using the <code>EventLogEditor</code> title
     * 
     * @return <code>true</code> if we successfully save the file, <code>false</code> otherwise.
     * 
     * @see EventLogEditor
     */
    public boolean saveAsEventLog(String selectedFileName) {  
    	// take the selected internal frame to save
		JInternalFrame selectedInternalFrame = elmView.getContentArea().getSelectedFrame();
		boolean fileSaved = false;
		// if an internal frame is selected
		if (selectedInternalFrame != null) {
			// if the frame is an EventLogEditor
			if (selectedInternalFrame instanceof EventLogEditor) {
				EventLogEditor eventLogEditor = (EventLogEditor)selectedInternalFrame;
				// show the save dialog to make the user select a file
		    	FileNameExtensionFilter xesFilter = new FileNameExtensionFilter("eXtensible Event Stream (*.xes)", "xes");
		    	FileNameExtensionFilter csvFilter = new FileNameExtensionFilter("Comma-Separated Values (*.csv)", "csv");
		    	FileNameExtensionFilter txtFilter = new FileNameExtensionFilter("Text File (*.txt)", "txt");
		    	String selectedFile;
		    	// check if we have a specific selected file to show, the document name otherwise
		    	if (selectedFileName == null) {
		    		// set the selected file with the document name
		    		selectedFile = eventLogEditor.getTitle();
		    		// if the editor is set as modified, we need to eliminate the asterisk at the end of the title
		    		if (eventLogEditor.isContentModified()) {
		    			// get the selected file name without the asterisk
		    			selectedFile = selectedFile.substring(0, selectedFile.length()-1);
		    		}	    		
		    	} else {
		    		selectedFile = selectedFileName;
		    	} 	
		    	// open a dialog to make the user select a file where to save, using filters
		    	String selectedDestinationFile = elmView.showFileSelectionDialog("Save As", selectedFile, xesFilter, csvFilter, txtFilter);
		    	// check if the user have selected a file, otherwise do nothing
		    	if (selectedDestinationFile != null) {		    				    			
		    		// create a Path object using the String path
		    		Path targetFile = Paths.get(selectedDestinationFile);
		    		// get the target file name
		    		String targetFileName = targetFile.getFileName().toString();		    		
		    		// remember if we create a new file
		    		boolean newFileCreated = false;
		    		// if the file still exists
		    		if (Files.exists(targetFile)) {
		    			// retrieve the user's choice		    			
		    			String message = "The file already exists. Do you want to overwrite it?";
		    			String title = "Save As";
		    			Object[] options = {"Yes", "   No   "};
		    	    	int choice = elmView.showOptionDialog(message, title, JOptionPane.WARNING_MESSAGE, options, options[1]);
		    	    	if (choice != 0) { // if the user select No (choice=1) or close the dialog (choice=-1)
		    	    		return saveAsEventLog(targetFileName);		    	    		
		    			} // otherwise go on to the saving procedure	    			
		    		} else if (Files.notExists(targetFile)) { // if the target file does not exist
		    			try {
		    				// create it
							Files.createFile(targetFile);
							newFileCreated = true;
						} catch (IOException e) {
							elmView.showErrorMessage("Unable to save the document: error while creating the file. \n" 
														+ "Please select another file where to save.");
							return saveAsEventLog(targetFileName);
						}
		    		} else { //if the process cannot determine if the file exists or not, because it does not have access to the file
		    			elmView.showErrorMessage("Unable to save the document. \n" 
		    										+ "The program does not have access to the selected file. \n"
													+ "Please select another file where to save.");
						return saveAsEventLog(targetFileName);
		    		}
		    		// write the content of the event log in the target file
		    		// try-with-resources Statement (close the resource automatically at the end in any case)
		    		try (OutputStream fileStream = new FileOutputStream(targetFile.toString());
		    				BufferedWriter fileWriter = new BufferedWriter(new OutputStreamWriter(fileStream))) {
		    			// start writing the file and set the application in a busy state in the meanwhile	    			        
    			        elmView.setBusyState(true);
    			        eventLogEditor.getTextAreaContent(fileWriter);
    			        elmView.setBusyState(false);
		            } catch (IOException e1) {
		            	elmView.setBusyState(false);
		    		    elmView.showErrorMessage("Unable to save the document in the selected file. \n"
								+ "Please select another file where to save.");
		            	if (newFileCreated) {		            	
							try {
								Files.delete(targetFile);
							} catch (IOException e2) {
								/* we tried to delete the file that we created when we saw that the selected file did not exist,
								 * but we are not able to. No worry about that, if we cannot, we leave the file where the user
								 * decided to save it, because anyway we are going to inform him about the saving error.
								 * We just left the empty file there, as was going to happen without this try */
							}
		            	}
		    		    return saveAsEventLog(targetFileName);
		            }		    		
		    		// set the event log editor as saved
		    		eventLogEditor.setContentModified(false);
		    		/* if the internal frame has a different title than the file name where we just saved it,
		    		 * or it has an asterisk at the end showing that it was not saved, then we need to update it */
		    		if (!eventLogEditor.getTitle().equalsIgnoreCase(targetFileName)) {		    			
		    			elmController.modifyJInternalFrameTitles(eventLogEditor, targetFileName);        		
		    		}		    		
		    		// get the absolute path of the file associated with this editor, if exists
					String eventLogEditorAbsolutePath = eventLogEditor.getAbsolutePath();
					// if the file path where we just saved is different from the last one, we update it
		    		if (eventLogEditorAbsolutePath == null) {
		    			eventLogEditor.setAbsolutePath(selectedDestinationFile);
		    		} else if (!eventLogEditorAbsolutePath.equalsIgnoreCase(selectedDestinationFile)) {
		    			eventLogEditor.setAbsolutePath(selectedDestinationFile);
		    		}
		    		fileSaved = true;
		    	}
	    		// at the end of the saving process (also the recursive ones), clear the file selection
	    		elmView.getFileChooser().setSelectedFile(new File(""));
			} else if (selectedInternalFrame instanceof FootprintMatrixInfo) { // if the frame is a FootprintMatrixInfo
				elmView.showInfoMessage("The selected document cannot be saved with the \"Save As\" procedure. \n" 
										+ "Please use the \"Export as\" button to export this document correctly.");
			} else { // for every other type of frame
				elmView.showInfoMessage("The selected document cannot be saved by this application.");
			}
		} else { // if no internal frames is selected 
			elmView.showInfoMessage("No documents to save. You first need to create a new Event Log or open one.");
		}
		return fileSaved; // true if we have successfully saved the file, false otherwise
    }
    
    /**
     * Opens a "Print" dialog to make the user select the printing settings, then prints
     * the document that is inside the selected <code>EventLogEditor</code> editor (the one 
     * inside the {@link unipi.bpm.elm.santeramo.view.EventLogManagerView#contentArea contentArea}). 
     * 
     * @return true if we successfully print the document, false otherwise 
     * 
     * @see EventLogEditor
     */
    public boolean printDocument() {        	
    	// take the selected internal frame to print
		JInternalFrame selectedInternalFrame = elmView.getContentArea().getSelectedFrame();
    	boolean isPrinted = false;
    	// if an internal frame is selected
		if (selectedInternalFrame != null) {
			// if the frame is an EventLogEditor
			if (selectedInternalFrame instanceof EventLogEditor) {
				//get the editor
				EventLogEditor eventLogEditor = (EventLogEditor)selectedInternalFrame;
				// create an header and a footer for the document
				MessageFormat header = new MessageFormat(eventLogEditor.getTitle());
				MessageFormat footer = new MessageFormat("Page - {0}");				 
				// get the actual font size of the editor
				int fontSize = eventLogEditor.getFontSize();
				// set the font size to a good value for a good printing of the code
				eventLogEditor.setFontSize(8);
				// get the textArea where is the code to print
				RSyntaxTextArea textArea = eventLogEditor.getTextArea();
				// set the line wrap to true so the code can wrap inside the document
				textArea.setLineWrap(true);
				try {			
					// try to open the print settings dialog
					// if the user select OK in the dialog print the document and return true, false otherwise
					isPrinted = textArea.print(header, footer);
				} catch (PrinterException e) {
					elmView.showErrorMessage("Unable to print the selected document. Please try again.");
				}	
				// restore the font size and the no-wrap setting of the textArea
				eventLogEditor.setFontSize(fontSize);
				textArea.setLineWrap(false);
				
			} else if (selectedInternalFrame instanceof FootprintMatrixInfo) { // if the selected frame is a FootprintMatrixInfo
				
				elmView.showInfoMessage("To get a printable version of the selected document, use the \"Export As\" button and select PDF.");									
			}else { // for every other type of frame  
				elmView.showInfoMessage("The selected document cannot be printed by this application.");
			}
		} else { // if no internal frames is selected 
			elmView.showInfoMessage("No documents to print. You first need to create a new Event Log or open one.");
		} 
        return isPrinted;
    }
    
    /**
     * Opens an "Import" dialog to ask the user to select which file to import, then the method tries
     * to import it (.CSV and .HTML extensions available).
     * 
     * <p>If a CSV file is selected, the method delegates the process to the <code>AnalysisManager</code>,
     * using its method {@link AnalysisManager#parseCSV(String, String)}. This method manages the parsing
     * process of the target file, from its CSV content to the XES version, then from XES to the XML format
     * to show the generated content inside an <code>EventLogEditor</code>. If some warnings are recorded
     * during the parsing, the method will show a dialog with them to the user.
     * 
     * <p>If an XHTML file is selected, the method first looks if there is already an opened 
     * <code>FootprintMatrixInfo</code> window associated with that file (and in this case it sets it as selected to show it to the user, 
     * asking him then if he wants to reload or to open in a new editor the file content, in case
     * another application modified it in the meanwhile). 
     * 
     * <p>Either if the process does not find any window associated with the target file,
     * or the user selected to reload or to open in a new window the file content, the method first
     * tries to read the content of the selected file and then to validate it, to ensure that
     * the file content is a well-formed XHTML code. If the document is valid against the specified DTD,
     * the process opens the file content inside an <code>FootprintMatrixInfo</code>
     * (setting its title using the file name, its content using the file content and its source
     * file using the path of the selected file) in the first case, to reload it in the second one,
     * and to open it inside an <code>FootprintMatrixInfo</code> (setting its title as a default value,
     * its content using the file content and without source file associated) in the third one.
     * 
     * <p>The validation of the selected XHTML file against its DTD is needed to ensure that the document
     * is well-formed and therefore it will be displayed correctly inside the application and the export
     * procedure will work perfectly.
     * 
     * <p>This method uses the {@link #validateAgainstDTD(String)} method to validate the selected document
     * against its DTD, check its documentation for detailed information about how it works.
     * 
     * @param selectedFileName the file name to set as selected in the dialog, <code>null</code> for no selection
     * 
     * @return <code>true</code> if we successfully import the file, <code>false</code> otherwise.
     * 
     * @see EventLogEditor
     * @see FootprintMatrixInfo
     */
    public boolean importDocument(String selectedFileName) {
    	boolean fileImported = false;
       	// set the available file extensions
    	FileNameExtensionFilter csvFilter = new FileNameExtensionFilter("Comma-Separated Values (*.csv, *.txt)", "csv", "txt");
		FileNameExtensionFilter htmlFilter = new FileNameExtensionFilter("Hyper Text Markup Language (*.html)", "html");
    	String selectedFile;
    	// check if we have a selected file to show, blank string otherwise
    	if (selectedFileName == null) {
    		selectedFile = "";
    	} else {
    		selectedFile = selectedFileName;
    	}  
    	// open a dialog to make the user select a file, using filters
    	String selectedSourceFile = elmView.showFileSelectionDialog("Import", selectedFile, csvFilter, htmlFilter);
    	// check if the user have selected a file, otherwise do nothing
    	if (selectedSourceFile != null) {
    		// create a Path object using the String path
    		Path targetFile = Paths.get(selectedSourceFile);
    		// if the target file exist
    		if (Files.exists(targetFile)) {   
				// if the user selects a CSV file
    			if (selectedSourceFile.endsWith(".csv") || selectedSourceFile.endsWith(".txt")) {
    				// ask the parseCSV method of the analysis manager to parse the CSV file and generate the XES log, managing possible errors
    				fileImported = elmController.getAnalysisManager().parseCSV(selectedSourceFile, targetFile.getFileName().toString());    				    			
    			} else if (selectedSourceFile.endsWith(".html")) { // if the user selects a HTML file
    				// here we first check if there is already an opened footprint matrix associated to this file  				
    		        //if the file is already opened and the user selects to reload the content, here the opened frame to reload 
    		        FootprintMatrixInfo targetFootprintMatrixInfo = null;
    		        // here the choice to open a new frame with only the file content and keep the opened frame as it is
    		        boolean openContentOnly = false;    				
    				// get all opened internal frames
    		        JInternalFrame[] internalFrameList = elmView.getContentArea().getAllFrames();    		        
    		        // for every internal frame
    		        for (JInternalFrame internalFrame: internalFrameList) {  
    		        	// check if it is a footprint matrix frame
    		            if (internalFrame instanceof FootprintMatrixInfo) {
    		            	// get the footprint matrix frame and the absolute path of the associated file
    		            	FootprintMatrixInfo footprintMatrixInfo = (FootprintMatrixInfo)internalFrame;
    		            	String absolutePath = footprintMatrixInfo.getAbsolutePath();
    		            	// if it has not an associated file, skip to the next frame
    		            	if (absolutePath != null) {
    		            		// check if the associated file is the same of the one that we are trying to import
    	    	            	if (absolutePath.compareTo(selectedSourceFile) == 0) { 
    	    	            		/* if we find an opened footprint matrix frame associated to the target file, we do not
    	    	            		 * need to open a new internal frame but just to set as selected the one already opened,
    	    	            		 * to show that there is already an existing footprint matrix frame for this file */
    	    	            		if (internalFrame.isIcon()) { // if the footprint matrix is minimised
    	    	    	            	try {
    	    	    	            		// set the footprint matrix frame maximised (and selected)
    	    	    	            		internalFrame.setIcon(false);
    	    	    					} catch (PropertyVetoException exception) {
    	    	    						// just get here the exception to not throw an error, the frame won't just be maximised
    	    	    					}  
    	    	    	            	/* setting the footprint matrix frame as maximised will automatically invokes the 
    	    	    					 * internalFrameDeiconified(InternalFrameEvent e) method of the InternalFrameStatusAdapter
    	    	    					 * adapter of the internal frame. This method will make all the other step */
    	    	    	            } else { // if footprint matrix is maximised
    	    	    	            	try {
    	    	    	            		// set the footprint matrix frame selected
    	    	    	            		internalFrame.setSelected(true);
    	    	    	        		} catch (PropertyVetoException exception) {
    	    	    	        			// just get here the exception to not throw an error, the frame won't just be selected
    	    	    	                } 
    	    	    	            }
    	    	            		// inform the user that there is already an opened frame associated with the file
    	    	            		String message = "There is already an opened document associated with the selected file. \n"
    	            						+ "If in the meanwhile the file has been modified by another application, and you want to\n"
    	            						+ "reload its content in the opened document, select \"Reload Content\" (the content \n"
    	            						+ "inside the document before the reload will be lost).\n"
    	            						+ "Otherwise select \"Open New\" to open a new document with the file content (the already \n"
    	            						+ "opened document will keep its content and the association with the file, the new \n"
    	            						+ "document will contain only the file content, without any file association). \n"
    	            						+ "Select instead \"Do Nothing\" to keep the already opened document as it is without \n"
    	            						+ "reloading or opening the file.";
    	    	            		// retrieve the user's choice       			
    	    	            		String title = "File already opened";
    	    	            		Object[] options = {" Reload Content ", "Open New", "Do Nothing"};
    	    						int choice = elmView.showOptionDialog(message, title, JOptionPane.QUESTION_MESSAGE, options, options[2]);
    	    						// check user's decision
    	    						if (choice == 0) { // if Reload
    	    							// set the target footprint matrix to reload and stop the loop 
    	    							targetFootprintMatrixInfo = footprintMatrixInfo;
    		    	            		break;
    	    						} else if (choice == 1) { // if Open New
    	    							// set the target footprint matrix to reload and stop the loop 
    	    							targetFootprintMatrixInfo = footprintMatrixInfo;
    	    							// set to open a new document with the content only
    	    							openContentOnly = true;
    	    							break;
    	    	            		} else { // otherwise for Continue or if the user closed the dialog
    	    							// clear the file selection and exit the process (we already set the frame as selected, so the user will find it on top)
    	    				    		elmView.getFileChooser().setSelectedFile(new File(""));
    	    				    		return fileImported; // false
    	    						}
    	    	            	}
    		            	}
    		            }
    		        }
    		        // if we have not found any opened frame associated with the target file or the user selected reload 		        		        	  		        			        	
		        	// try-with-resources Statement (close the resource automatically at the end in any case)
	    			try (InputStream fileStream = new FileInputStream(selectedSourceFile);
	    					BufferedReader fileReader = new BufferedReader(new InputStreamReader(fileStream))) {   	    					 
    					// ask the user if he wants to validate the file, if NO stop the process	    			
    	    			String message = "The application needs to check the well-formedness of the document before continuing. \n" 
    	    							+ "The validation process could also take few minutes to complete. \n"    	    							
    	    							+ "You can also import the document without validating it, but be aware of possible problems\n"
    	    							+ "with the document visualization.\n"
    	    							+ "Do you want to validate the document and proceed with the import?";
    	    			String title = "Validate";
    	    			Object[] options = {"Validate and Import", "Do not Import", "   Do not Validate and Import   "};
    	    	    	int choice = elmView.showOptionDialog(message, title, JOptionPane.WARNING_MESSAGE, options, options[0]);
    	    	    	if (choice == 0) { // if the user select Yes    	    	    	    		  	    	    	    		
	    					StringBuffer fileContent = new StringBuffer();
	    					String line = null;
	    					// get the line separator specific of the machine system
	    					String newline = System.getProperty("line.separator");
	    					// read the content of the file and set the application in a busy state in the meanwhile	    			        
	    			        elmView.setBusyState(true);  
	    					while ((line = fileReader.readLine()) != null) {
	            	        	fileContent.append(line);
	            	        	fileContent.append(newline);
	            	        }
	            	        elmView.setBusyState(false);
	            	        String xhtml = fileContent.toString();
	            	        // ask to the validation process to check the document
	            	        if (validateAgainstDTD(xhtml)) { // if valid	            	        	
	            	        	// if there is not a target footprint matrix to reload, it means we are importing a new file
	        					if (targetFootprintMatrixInfo == null) {
	        						/* create a new footprint matrix set with the target file name, the target file content,
        	                         * the absolute path of the target file and its content as exported */
        	            	        String frameTitle = targetFile.getFileName().toString();
        							frameTitle = frameTitle.substring(0,frameTitle.lastIndexOf("."));
        	                	    elmController.newFootprintMatrixInfo(frameTitle, xhtml, selectedSourceFile, true);     	        					
	        					} else { // if we have a target footprint matrix
	        						if (openContentOnly) { // if we have to open a new frame with just the content in it
	        	                        /* create a new footprint matrix set with default name, the target file content,
	                                     * no absolute path and its content as not exported */
	        							elmController.newFootprintMatrixInfo(null, xhtml, null, false);
	        						} else { // if we have to reload the file instead
    	        						// set the new xhtml content
    	        						targetFootprintMatrixInfo.setTextAreaContent(xhtml); 
    	        						// the title, the associated file and the exported=true values are already set, then no needs to modify them
	        						}
	        					}    	            	        	      	        	
    	                	    fileImported = true;
	            	        } else { // if not valid
    	    					elmView.showErrorMessage("Unable to import the document due to validation errors. \n" 
														+ "Please first fix these errors and then try again.");
	            	        }        	            	        
    	    			} else if (choice==2) { // import without validation
    	    				StringBuffer fileContent = new StringBuffer();
	    					String line = null;
	    					// get the line separator specific of the machine system
	    					String newline = System.getProperty("line.separator");
	    					// read the content of the file and set the application in a busy state in the meanwhile	    			        
	    			        elmView.setBusyState(true);  
	    					while ((line = fileReader.readLine()) != null) {
	            	        	fileContent.append(line);
	            	        	fileContent.append(newline);
	            	        }
	            	        elmView.setBusyState(false);
	            	        String xhtml = fileContent.toString();
	            	        // if there is not a target footprint matrix to reload, it means we are importing a new file
        					if (targetFootprintMatrixInfo == null) {
        						/* create a new footprint matrix set with the target file name, the target file content,
    	                         * the absolute path of the target file and its content as exported */
    	            	        String frameTitle = targetFile.getFileName().toString();
    							frameTitle = frameTitle.substring(0,frameTitle.lastIndexOf("."));
    	                	    elmController.newFootprintMatrixInfo(frameTitle, xhtml, selectedSourceFile, true);     	        					
        					} else { // if we have a target footprint matrix
        						if (openContentOnly) { // if we have to open a new frame with just the content in it
        	                        /* create a new footprint matrix set with default name, the target file content,
                                     * no absolute path and its content as not exported */
        							elmController.newFootprintMatrixInfo(null, xhtml, null, false);
        						} else { // if we have to reload the file instead
	        						// set the new xhtml content
	        						targetFootprintMatrixInfo.setTextAreaContent(xhtml); 
	        						// the title, the associated file and the exported=true values are already set, then no needs to modify them
        						}
        					}
    	    			}
	    			} catch(IOException e) {
	    				elmView.setBusyState(false);
	    			    elmView.showErrorMessage("Unable to read the XHTML file. Please try again.");
	    			}      		        
    			}    			
    		} else if (Files.notExists(targetFile)) { // if the target does not file exist		
				elmView.showWarningMessage("Unable to import the file \""+targetFile.getFileName().toString()+"\", the file does not exist. \n" 
											+ "Check the file name and try again.");	
				// star a new import process to select another file, with the last selected file shown
				fileImported = importDocument(targetFile.getFileName().toString());
    		} else { //if the process cannot determine if the file exists or not, because it does not have access to the file
    			elmView.showWarningMessage("Unable to import the file \""+targetFile.getFileName().toString()+"\", the program does not have access to the file. \n"
    										+ "Please select another file or change file permissions.");	
				// star a new import process to select another file, with the last selected file shown
    			fileImported = importDocument(targetFile.getFileName().toString());
    		}
    		// at the end of the import process (also the recursive ones), clear the file selection
    		elmView.getFileChooser().setSelectedFile(new File(""));
    	}    	
    	return fileImported; // true if we import the file, false otherwise
    }
    
    /**
     * Validates the given XHTML code against the DTD specified inside the document <code>DOCYPE</code>.
     * 
     * <p>The method needs Internet connection to retrieve the correct DTD specified by the URL
     * inside the document <code>DOCTYPE</code>. If the connection is present and the URL is reachable,
     * it validates the document and shows to the user a detailed result; if the connection is not present
     * or an error with the URL occurs, it advises the user that it cannot validate the document, therefore
     * it is going to take it as it is if the user wants to  proceed anyway, warning him about possible error
     * while displaying or exporting the document.
     * 
     * <p>This method uses the 
     * {@link EventLogManagerController#checkInternetConnection(String, int) checkInternetConnection(String, int)}
     * method of the Controller to check if the Internet connection is present and if the W3C web-site 
     * (where to download the specified DTD) is reachable. Check its documentation for detailed information 
     * about how it works.
     * 
     * <p>During the validation, the <code>SAXParser</code> utilises the 
     * {@link AnalysisManager.ParserContentHandler ParserContentHandler} handler to log all the
     * validation warning or errors that may occur; at the end, the method will shows this log 
     * to the user inside a dialog box.
     * 
     * @param xhtml the XHTML code to validate against its DTD
     * 
     * @return <code>true</code> if we successfully validate the XHTML code, <code>false</code> otherwise
     * 
     * @see SAXParser
     * @see ParserContentHandler
     */   
	private boolean validateAgainstDTD(String xhtml) {		
		boolean docValidated = false;
		// check the Internet connection and the URL where to get the DTDs
		if (elmController.checkInternetConnection("http://www.w3.org/", 15000)) {			
			// creating a SAXParserFactory to generate the parser
	        SAXParserFactory parserFactory = SAXParserFactory.newInstance();
	        // set to check only the well-formedness of the document against its DTD, this is what we just need for our purposes
	        parserFactory.setValidating(false);
	        // set namespace aware to receive element names
	        parserFactory.setNamespaceAware(true);
	        // create the SAXParser
	        SAXParser parser;
			try {
				parser = parserFactory.newSAXParser();
			} catch (ParserConfigurationException e) {
				elmView.showErrorMessage("Unable to start the validation. Please try again.");
				return docValidated; // false
			} catch (SAXException e) {
				elmView.showErrorMessage("Unable to start the validation. Please try again. \n" 
										+ "Error: " + e.getMessage());
				return docValidated; // false
			}			
			// create an instance of the parser content handler
	        ParserContentHandler parserContentHandler = elmController.getAnalysisManager().new ParserContentHandler();
	        // parse the file using the handler to manage the SAXException exceptions
	        try {	
	        	// set the application in a busy state in the meanwhile	   
	        	elmView.setBusyState(true); 	
	        	parser.parse(new InputSource(new StringReader(xhtml)), parserContentHandler);
	        	elmView.setBusyState(false); 	        								            
			} catch (SAXException e) {
				/* all the SAXException exceptions are managed by the content handler, this 
				 * if needed will throw a SAXException to activate this catch and stop the 
				 * validation*/
				elmView.setBusyState(false);
				return docValidated; // false
			} catch (IOException e) {
				elmView.setBusyState(false);
				elmView.showErrorMessage("Unable to validate the document. Please try again.");
				return docValidated; // false
			}
	        // if the parsing process did not throw a fatal error, get the result of the parsing
			docValidated = parserContentHandler.isDocumentValid();
			// now check if there is something to show to the user (warnings, errors, or just a confirmation)
			StringBuffer parserLog = parserContentHandler.getParserLog();
			String parserLogString = parserLog.toString();
			// check if the log is empty or not
			if (parserLogString.equalsIgnoreCase("")) {
				// the log is empty (no warning or error messages)
				elmView.showInfoMessage("Validation Successful! \n"
										+ "The document is well-formed.");
			} else { // the log has some warning or error messages to display						
	    		String frameTitle = "HTML/XHTML Validation Result";
	    		String frameContent = frameTitle + ": \n" + parserLogString;
				// set a text area with the message in it
				JTextArea dialogTextArea = new JTextArea(frameContent);
				// create a scrollable panel that contains the text area
				JScrollPane dialogScrollPane = new JScrollPane(dialogTextArea);  
				// set the text area
				dialogTextArea.setLineWrap(false);  
				dialogTextArea.setWrapStyleWord(true); 
				dialogTextArea.setEditable(false);
				dialogTextArea.setMargin(new Insets(2,4,2,4));
				// set the panel dimension
				dialogScrollPane.setPreferredSize(new Dimension(500, 300));
				// set title, type and options
				String dialogTitle = "Event Log Manager HTML/XHTML Validation Process";
				int messageType;
				if (docValidated) {
					// if true, only warnings have been generated by the parser
					messageType = JOptionPane.WARNING_MESSAGE;
				} else {
					// if false, at least one error has been generated by the parser
					messageType = JOptionPane.ERROR_MESSAGE;								
				}
    			Object[] options = {"Ok", "Show on Editor"};
    			// ask the View to show the option dialog
    	    	int choice = elmView.showOptionDialog(dialogScrollPane, dialogTitle, messageType, options, options[0]);
    	    	if (choice == 1) { // if the user select Show on Editor		    	    					    	    		
    	    		// try-with-resources Statement
    	    		try (InputStream stringStream = new ByteArrayInputStream(frameContent.getBytes(StandardCharsets.UTF_8));
    	    				BufferedReader stringReader = new BufferedReader(new InputStreamReader(stringStream))) {
	    				// open an editor with the passed title and the reader over the content to display, without an associated file and set as modified
	    				elmController.newEventLogEditor(frameTitle, stringReader, null, true); 
    	    		} catch (IOException e) {
        			    elmView.showErrorMessage("Unable to display the HTML/XHTML Validation Result. Please try again.");
    				}  	    		
    			}
			}
		} else { // if there is not Internet connection or the URL is unreachable, we cannot validate
			/* tell to the user that we cannot validate the document, therefore we are going to take it as it is
			 * if he wants to proceed anyway, warning him about possible error while displaying it or exporting it */
			String message =   "Failed to connect to the W3C website to get the specified DTD: \n"
							 + "Internet connection or the specified resource may be unavailable. \n\n"
							 + "The application therefore is not able to validate the document against its DTD \n" 
							 + "and ensure its well-formedness. \n"
							 + "The document will be imported without any guarantee that the application is able \n" 
							 + "to display it correctly and export it without errors. \n\n"
							 + "The application however always supports the importing of well-formed XHTML files. \n"
							 + "Importing one of them (for example an XHTML document previously exported from \n"
							 + "the application itself) will result in a perfect display and an export without errors. \n";
			String title = "Internet connection unavailable";
			Object[] options = {"Proceed", "Stop Import"};
	    	int choice = elmView.showOptionDialog(message, title, JOptionPane.WARNING_MESSAGE, options, options[0]);
	    	// if the user select Proceed, return true to continue the import, otherwise the method will return false
	    	if (choice == 0) { 
	    		docValidated = true;
	    	}
		}
        return docValidated; // true if the document is valid, false otherwise or if some errors occurred during the process
	}
    
    /**
     * Opens an "Export" dialog to ask the user to decide where to export (and in which format) 
     * the content either of the selected <code>EventLogEditor</code> editor or of the selected
     * <code>FootPrintMatrixInfo</code> document (that are both inside the
     * {@link unipi.bpm.elm.santeramo.view.EventLogManagerView#contentArea contentArea}). 
     * The method then tries to export the content of the selected frame in the selected file 
     * (CSV format available for the <code>EventLogEditor</code>; HTML, LaTeX, PDF and PNG formats
     * available for the <code>FootPrintMatrixInfo</code>). 
     * 
     * <p>If CSV format is selected, the method reads the content of the <code>EventLogEditor</code> editor
     * (that is a XES log) and converts it in a CSV format, using the <code>XesCsvSerializer</code>. At the 
     * end of the process it writes the generated CSV document in the selected file.
     * 
     * <p>If HTML format is selected, the method reads the content of the <code>FootPrintMatrixInfo</code>
     * document (that is in XHTML format) and write it directly to the selected file. At the end of the process,
     * if the export procedure went right, the method sets the associated file of the <code>FootPrintMatrixInfo</code>
     * with the target file and its exported status as <code>true</code>. This is the only case when the exported
     * status of the <code>FootPrintMatrixInfo</code> is set to <code>true</code>, and it is because the HTML format
     * is the "save" format of this type of frame: if the user in fact saves the content of a <code>FootPrintMatrixInfo</code>
     * in HTML and then closes that frame, it is possible for him to open it again inside the application, using the 
     * Import function. Also, the Close function checks this value before closing a <code>FootPrintMatrixInfo</code>,
     * to ask the user if he wants to close the frame without saving it in HTML, and ensure that he is advised
     * about this special saving procedure.
     * 
     * <p>If LaTeX format is selected, the method reads the content of the <code>FootPrintMatrixInfo</code>
     * document (that is in XHTML format) and uses the "htmltolatex" library to convert the XHTML document 
     * into a LaTex one and to write it in the selected file.
     * 
     * <p>If PDF format is selected, the method reads the content of the <code>FootPrintMatrixInfo</code>
     * document (that is in XHTML format) and uses the "flyingsaucer" library to create a PDF in the selected
     * destination that represents the XHTML document. This library, first than creating the PDF file, it 
     * validates the document to ensure that is a well-formed XHTML document, otherwise the library cannot
     * create a PDF that is a correct representation of the given document. This is why the application passes
     * to the export procedure only internal XHTML documents (the ones created by the Analysis Manager) or
     * imported XHTML documents that have been first validated against their DTDs, to ensure that the library
     * will only work with well-formed documents and therefore the library validation process will always 
     * have success.
     * 
     * <p>If PNG format is selected, the method first creates an image that is the representation of the
     * <code>FootPrintMatrixInfo</code> text area (with the same dimension and content), then it writes the
     * image in the selected file. This means that the user can change the frame dimension (and therefore
     * also the content position in it, because this changes based on the frame dimension) to get the right
     * content disposition before exporting it in a PNG image format.
     * 
     * <p>The method implements a behaviour similar to the Save As button, with overwrite checking,
     * selected file existence checking and management of errors that may occur during the process.
     * It simply at the end exports the content in the specified format instead save it as it is
     * in a text file, implementing on this way other saving options for the selected content.
     * 
     * <p>The method is made for the <code>EventLogEditor</code> and <code>FootPrintMatrixInfo</code> frames.
     * In case of using it with other frame types, the method manages this situation with specifics information
     * messages delivered to the user.
     * 
     * @param selectedFileName the file name to set as selected in the dialog, <code>null</code> for using the <code>EventLogEditor</code> title
     * 
     * @return <code>true</code> if we successfully export the file, <code>false</code> otherwise.
     * 
     * @see EventLogEditor
     * @see FootprintMatrixInfo
     * @see XesCsvSerializer
     * @see AnalysisManager
     */
    public boolean exportDocument(String selectedFileName) {
    	// take the selected internal frame to export
		JInternalFrame selectedInternalFrame = elmView.getContentArea().getSelectedFrame();
		boolean fileExported = false;
		// if an internal frame is selected
		if (selectedInternalFrame != null) {
			// if the selected frame is an EventLogEditor
			if (selectedInternalFrame instanceof EventLogEditor) {	
				// get the editor
				EventLogEditor eventLogEditor = (EventLogEditor)selectedInternalFrame;				
		       	// set the available file extensions
		    	FileNameExtensionFilter csvFilter = new FileNameExtensionFilter("Comma-Separated Values (*.csv)", "csv");
		    	FileNameExtensionFilter txtFilter = new FileNameExtensionFilter("Comma-Separated Values (*.txt)", "txt");

		    	String selectedFile;
		    	// check if we have a specific selected file to show, the document name otherwise
		    	if (selectedFileName == null) {
		    		// set the selected file with the document name
		    		selectedFile = eventLogEditor.getTitle();
		    		// if the editor is set as modified, we need to eliminate the asterisk at the end of the title
		    		if (eventLogEditor.isContentModified()) {
		    			// get the selected file name without the asterisk
		    			selectedFile = selectedFile.substring(0, selectedFile.length()-1);
		    		}	    		
		    	} else {
		    		selectedFile = selectedFileName;
		    	}  	
		    	// open a dialog to make the user select a file where to export, using filters
		    	String selectedDestinationFile = elmView.showFileSelectionDialog("Export As", selectedFile, csvFilter, txtFilter);
		    	// check if the user have selected a file, otherwise do nothing
		    	if (selectedDestinationFile != null) {		    				    			
		    		// create a Path object using the String path
		    		Path targetFile = Paths.get(selectedDestinationFile);	    		
		    		// remember if we create a new file
		    		boolean newFileCreated = false;
		    		// if the file exists 
		    		if (Files.exists(targetFile)) {
		    			// check if the user wants to overwrite the file	    			
		    			String message = "The file already exists. Do you want to overwrite it?";
		    			String title = "Export As";
		    			Object[] options = {"Yes", "   No   "};
		    	    	int choice = elmView.showOptionDialog(message, title, JOptionPane.WARNING_MESSAGE, options, options[1]);
		    	    	if (choice != 0) { // if the user select No (choice=1) or close the dialog (choice=-1)
		    	    		// call again the export procedure to start back again
		    	    		return exportDocument(targetFile.getFileName().toString());		    	    		
		    			} // otherwise go on to the export procedure	    			
		    		} else if (Files.notExists(targetFile)) { // if the target file does not exist
		    			try {
		    				// create it
							Files.createFile(targetFile);
							// set the flag to remember that we create a new file
							newFileCreated = true;
						} catch (IOException e) {
							elmView.showErrorMessage("Unable to create the file. Please try again.");
							return fileExported; // false
						}
		    		} else { //if the process cannot determine if the file exists or not, because it does not have access to the file
		    			elmView.showErrorMessage("Unable to export the document. \n" 
		    									+ "The program does not have access to the selected file. \n"
		    									+ "Please select another file where to export.");
		    			// call again the export procedure to start back again
		    			return exportDocument(targetFile.getFileName().toString());
		    		}
		    		// now that we have a file where to export, get the content to export
					boolean xesLogGenerated = true;
		    		// if the content to export has not been generated yet (the XLog of the textual log)
					if (!eventLogEditor.isXesLogsGenerated()) {
						// ask the AnalysisManager to generate the XLog for this editor content
		    			xesLogGenerated = elmController.getAnalysisManager().generateEventLogFootprint(false, eventLogEditor);
					}
					// if we successfully generated the XLog
		    		if (xesLogGenerated) {
		    			// get the XLog
		    			Collection<XLog> xesLogs = eventLogEditor.getXesLogs();
		    			// get the serializer from XES to CSV
						XesCsvSerializer csvSerializer = new XesCsvSerializer();		
						// suppose we can export all the log successfully, if one fails this variable will be false
						boolean exportAll = true;
						boolean moreThanOne = false;
						// for each log in the collection
						for (XLog log : xesLogs) {	
							// if this is the second log or further, print first the log separator symbol (+)
							if (moreThanOne) {
								String newLine = System.getProperty("line.separator");
								String newCsvlog = newLine + "+" + newLine;
								try {
									// append to the file the symbol
									Files.write(targetFile, newCsvlog.getBytes(), StandardOpenOption.APPEND);
								} catch (IOException e) {
									// if an error occurred, stop the export process and show the error to the user
									exportAll = false;
			        			    elmView.showErrorMessage("Unable to write the XES document in the selected destination. \n"
			        			    						+ "Please try again.");
			        			    break;
								}												    			
			        		} else { // if this is the first log to export
			        			moreThanOne = true;
			        		}
							// open an output stream where to serialise
							// try-with-resources Statement (close the resource automatically at the end in any case)
							try (ByteArrayOutputStream outputStream = new ByteArrayOutputStream()) {					    				               			        	
								// set the application in a busy state in the meanwhile	   
			    	        	elmView.setBusyState(true);	
			    	        	// serialise the log in CSV format inside the output stream using the serializer
			    				csvSerializer.serialize(log, outputStream);
			    	        	elmView.setBusyState(false);	
				        		try {
					    			// set the busy state of the application
					    			elmView.setBusyState(true);
					    			// write the output stream with the CSV document to the file
					    			if (moreThanOne) { // if this was the second log or further
					    				// append to the file this content
					    				Files.write(targetFile, outputStream.toByteArray(), StandardOpenOption.APPEND);
					    			} else {
					    				// write to the file this content (truncating existing text and start the file with this content)
					    				Files.write(targetFile, outputStream.toByteArray());
					    			}					                
					                elmView.setBusyState(false);			                
					            } catch (IOException e) {
					            	// error while writing the file			            	
					            	elmView.setBusyState(false);
					            	// stop the export process and show the error to the user
					            	exportAll = false;
			        			    elmView.showErrorMessage("Unable to write the XES document in the selected destination. \n"
			        			    						+ "Please try again.");
			        			    break;
					            }				        					    			
			    			} catch (IOException e) {
			    				// error while creating the output stream
		    					elmView.setBusyState(false);
		    					// stop the export process and show the error to the user
		    					exportAll = false;
		        			    elmView.showErrorMessage("Unable to serialize the XES document in the selected destination. \n"
		        			    						+ "Please try again.");
		        			    break;
			    			}					    							    			
						}	
						// set the result of the export depending of the result of every single log export
						fileExported = exportAll;								
						// if we do not exported the file correctly and we have created a new output file, delete it
	    				if (!fileExported && newFileCreated) {		            	
							try {
								Files.delete(targetFile);
							} catch (IOException e) {
								/* we tried to delete the file that we created when we saw that the selected file did not exist,
								 * but we are not able to. No worry about that, if we cannot, we leave the file where the user
								 * decided to export it, because anyway we are going to inform him about the saving error.
								 * We just left the empty file there, as was going to happen without this try */
							}
		            	}

		    		} else { // if an error occurred while analysing the document
						elmView.showErrorMessage("Error while analysing the document. Please try again.");
					}					
		    	}
			} else if (selectedInternalFrame instanceof FootprintMatrixInfo) { // if the selected frame is a FootprintMatrixInfo
				// get the footprint matrix object
				FootprintMatrixInfo footprintMatrixInfo = (FootprintMatrixInfo)selectedInternalFrame;
				// set the possible export option to show as filters inside the file selection dialog
				FileNameExtensionFilter htmlFilter = new FileNameExtensionFilter("Hyper Text Markup Language (*.html)", "html");
		    	FileNameExtensionFilter latexFilter = new FileNameExtensionFilter("LaTeX (*.tex)", "tex");
		    	FileNameExtensionFilter pdfFilter = new FileNameExtensionFilter("Portable Document Format (*.pdf)", "pdf");
		    	FileNameExtensionFilter pngFilter = new FileNameExtensionFilter("Portable Network Graphics (*.png)", "png");    	
		    	
		    	String selectedFile;
		    	// check if we have a specific selected file to show, the document name otherwise
		    	if (selectedFileName == null) {
		    		// set the selected file with the document name
		    		selectedFile = footprintMatrixInfo.getTitle();	    		
		    	} else {
		    		selectedFile = selectedFileName;
		    	} 	
		    	// open a dialog to make the user select a file where to export, using filters
		    	String selectedDestinationFile = elmView.showFileSelectionDialog("Export As", selectedFile, htmlFilter, latexFilter, pdfFilter, pngFilter);
		    	// check if the user have selected a file, otherwise do nothing
		    	if (selectedDestinationFile != null) {		    				    			
		    		// create a Path object using the String path
		    		Path targetFile = Paths.get(selectedDestinationFile);	    		
		    		// remember if we create a new file
		    		boolean newFileCreated = false;
		    		// if the file exists 
		    		if (Files.exists(targetFile)) {
		    			// check if the user wants to overwrite the file	    			
		    			String message = "The file already exists. Do you want to overwrite it?";
		    			String title = "Export As";
		    			Object[] options = {"Yes", "   No   "};
		    	    	int choice = elmView.showOptionDialog(message, title, JOptionPane.WARNING_MESSAGE, options, options[1]);
		    	    	if (choice != 0) { // if the user select No (choice=1) or close the dialog (choice=-1)
		    	    		// call again the export procedure to start back again
		    	    		return exportDocument(targetFile.getFileName().toString());		    	    		
		    			} // otherwise go on to the export procedure	    			
		    		} else if (Files.notExists(targetFile)) { // if the target file does not exist
		    			try {
		    				// create it
							Files.createFile(targetFile);
							// set the flag to remember that we create a new file
							newFileCreated = true;
						} catch (IOException e) {
							elmView.showErrorMessage("Unable to create the file. Please try again.");
							return fileExported; // false
						}
		    		} else { //if the process cannot determine if the file exists or not, because it does not have access to the file
		    			elmView.showErrorMessage("Unable to export the document. \n" 
		    									+ "The program does not have access to the selected file. \n"
		    									+ "Please select another file where to export.");
		    			// call again the export procedure to start back again
		    			return exportDocument(targetFile.getFileName().toString());
		    		}
		    		// now that we have a file where to export, get the content to export
		    		String xhtml = footprintMatrixInfo.getTextAreaContent();		    		
		    		
		    		// export the file, selecting the right procedure depending on the type of file where to export
		    		if (selectedDestinationFile.endsWith(".html")) {		    			
			    		//if HTML, write directly in the target file
		    			boolean getError = false;
		    			try {
			    			// set the busy state of the application
			    			elmView.setBusyState(true); 
			                Files.write(targetFile, xhtml.getBytes());
			                elmView.setBusyState(false);			                
			            } catch (IOException e) {
			            	// error while writing the file			            	
			            	elmView.setBusyState(false);
			            	getError = true;		            				            	
			            }	
		    			// if we write the file correctly
		    			if(!getError) {
		    				// set the content as exported (this happens only with HTML export)
				    		footprintMatrixInfo.setContentExported(true);			    		
				    		// get the absolute path of the file associated with this footprint matrix, if exist 
							String absolutePath = footprintMatrixInfo.getAbsolutePath();
							// if the file path where we just exported is different from the last one, we update it
				    		if (absolutePath == null) {
				    			footprintMatrixInfo.setAbsolutePath(selectedDestinationFile);
				    		} else if (!absolutePath.equalsIgnoreCase(selectedDestinationFile)) {
				    			footprintMatrixInfo.setAbsolutePath(selectedDestinationFile);
				    		}
				    		// file exported correctly
				    		fileExported = true;
		    			} else { // otherwise, if an error occurred
		    				// show the error to the user
			            	elmView.showErrorMessage("Unable to export the document. \n" 
			            							+ "A problem with the selected file occurred. \n" 
			            							+ "Please try again selecting another output file.");
		    				// if we created the file, delete it
		    				if (newFileCreated) {		            	
								try {
									Files.delete(targetFile);
								} catch (IOException e) {
									/* we tried to delete the file that we created when we saw that the selected file did not exist,
									 * but we are not able to. No worry about that, if we cannot, we leave the file where the user
									 * decided to export it, because anyway we are going to inform him about the saving error.
									 * We just left the empty file there, as was going to happen without this try */
								}
			            	}
		    			}
			    		
		    		} else if (selectedDestinationFile.endsWith(".tex")) {
		    			// if LATEX, we first need to create a temp HTML file and then pass it to the htmltolatex library
		    			boolean getError = false;
		    			String errorMessage = "";
		    			Path htmlTempFile = null;
		    			try {
		    				// set the busy state
		    				elmView.setBusyState(true); 
		    				// create a temp file in the system default temp folder
		    				htmlTempFile = Files.createTempFile("EventLogManager_HTML-Export_", ".html");
		    				// delete the file when the virtual machine terminates
		    				htmlTempFile.toFile().deleteOnExit();
		    				// write in the temp file the XHTML content to export in LATEX
		    				Files.write(htmlTempFile, xhtml.getBytes());
		    				elmView.setBusyState(false); 
						} catch (IOException e) {
							// error while creating/writing the temp file
							elmView.setBusyState(false);							
							getError = true;
							errorMessage = "Unable to export the document. \n" 
											+ "A problem with the system temporary-file directory occurred. \n" 
											+ "Please try again.";
						}	
		    			// if the temp file has been created and written correctly
		    			if (!getError) {
		    				// get the parser
		    				Parser htmlLatexParser = new Parser();
				            try {
				            	// set the busy state
				            	elmView.setBusyState(true); 
				            	// parse the HTML temp file using the ParserHandler and get as result a LATEX file in the target path
				            	URL latexConfig = this.getClass().getResource("/latex/config.xml"); // get the image from the URL
				            	htmlLatexParser.parse(new File(htmlTempFile.toString()), new ParserHandler(targetFile.toFile(), latexConfig.toString()));
				            	elmView.setBusyState(false); 			            	
				            } catch (FatalErrorException e) {
				            	// error while parsing the HTML temp file
				            	elmView.setBusyState(false); 				            	
				            	getError = true;
				            	errorMessage = "Unable to parse the document and generate a LaTeX file.\n"
				            				  +"Please try again.";
				    		}
		    			}		
		    			// in any case, at this point we do not need anymore the temp file, so we delete it
			            if (htmlTempFile != null) {		            	
							try {
								Files.delete(htmlTempFile);
							} catch (IOException e) {
								/* we tried to delete the temp file that we created, but we are not able to.
								 * No worry about that, if we cannot, we leave the file and the virtual machine will
								 * delete it on exit. */
							}
		            	}
			            // if we did not get any error while creating the HTML temp file and while parsing it into a LATEX document
			            if (!getError) {
			            	// file exported correctly
				            fileExported = true;			            			            	
			            } else { // otherwise, if an error occurred
				            // show the error message to the user
				            elmView.showErrorMessage(errorMessage);
			            	// if we created the file, delete it
				            if (newFileCreated) {		            	
								try {
									Files.delete(targetFile);
								} catch (IOException e) {
									/* we tried to delete the file that we created when we saw that the selected file did not exist,
									 * but we are not able to. No worry about that, if we cannot, we leave the file where the user
									 * decided to export it, because anyway we are going to inform him about the saving error.
									 * We just left the empty file there, as was going to happen without this try */
								}
			            	}
			            }
			            
		    		} else if (selectedDestinationFile.endsWith(".pdf")) {	
		    			// if PDF, we use the flyingsaucer library to export the content, that in this case MUST be a well-formed XHTML code
		    			boolean getError = false;
		    			String errorMessage = "";
		    			ITextRenderer renderer;		    			
	    				try {   
	    				// set the busy state
		    			elmView.setBusyState(true);		    			
		    			// get the renderer
		    			renderer = new ITextRenderer();		    			
		    			/* we set the renderer using the XHTML code generated by the AnalysisManager (so well-formed for sure),
		    			 * or the XHTML code imported from an external file (but checked during import, so well-formed too) */
		    			renderer.setDocumentFromString(xhtml);
		    			renderer.layout();	
	    				} finally {
	    					elmView.setBusyState(false);
	    				}	    						    			
		    			// try-with-resources Statement (close the resource automatically at the end in any case)
		    			try (FileOutputStream fileStream = new FileOutputStream(selectedDestinationFile)) { 			    			
			    				// set the busy state
			    				elmView.setBusyState(true);
			    				// ask the renderer to create the PDF in the target destination
								renderer.createPDF(fileStream);	
								elmView.setBusyState(false);									    				    			
		    			} catch (IOException e) {   										
		    				// error while opening a stream to the destination file
		    				elmView.setBusyState(false);
		    				getError = true;
		    				errorMessage = "Unable to export the document. \n"
		    								+ "A problem with the selected file occurred. \n" 
		    								+ "Please try again selecting another output file.";
						} catch (DocumentException e) {
							// error while creating the PDF
							elmView.setBusyState(false);								
							getError = true;
							errorMessage = "Unable to export the document. Please try again.";
						}
		    			// if we did not get any error while creating the PDF file in the target destination
		    			if (!getError) {
		    				// complete the PDF file
		    				renderer.finishPDF();
		    				// file exported successfully
			    			fileExported = true;			            			            	
			            } else { // otherwise, if an error occurred
				            // show the error to the user
				            elmView.showErrorMessage(errorMessage);
			            	// if we created the file, delete it
				            if (newFileCreated) {		            	
								try {
									Files.delete(targetFile);
								} catch (IOException e) {
									/* we tried to delete the file that we created when we saw that the selected file did not exist,
									 * but we are not able to. No worry about that, if we cannot, we leave the file where the user
									 * decided to export it, because anyway we are going to inform him about the saving error.
									 * We just left the empty file there, as was going to happen without this try */
								}
			            	}
			            }		    			
		    			
		    		} else if (selectedDestinationFile.endsWith(".png")) {	   			
		    			// if PNG, get an image with the dimension of the frame and paint in it the frame content
		    			boolean getError = false;
		    			// get the footprintMatrix text area where the content is displayed
		    			JTextPane footprintMatrixTextArea = footprintMatrixInfo.getTextArea();
		    			//get the frame dimension
		    			Dimension textAreaDimension = footprintMatrixTextArea.getSize();
		    			// get an image big as the frame dimension
		    			BufferedImage img = new BufferedImage(textAreaDimension.width, textAreaDimension.height, BufferedImage.TYPE_INT_ARGB);
		    			// access to the graphic 
		    			Graphics imgGraphics = img.getGraphics();
		    			// paint the frame content into the image, disposing the content in that space 
		    			// (the frame shows to the user a preview of how the content will be displayed inside the image)
		    			footprintMatrixTextArea.paint(imgGraphics);
		    			// close the graphic
		    			imgGraphics.dispose();
		    			// now try to write the image object in the target file as a PNG
	    				try {
	    					// set the busy state
	    					elmView.setBusyState(true);
	    					// write the image object in the target file as a PNG 
							ImageIO.write(img, "png", new File(selectedDestinationFile));
							elmView.setBusyState(false);							
						} catch (IOException e) {
							// error while writing into the target file
							elmView.setBusyState(false);
							getError = true;
						} 	  
	    				// if we write the file correctly
	    				if(!getError) {	  
	    					// set the file as exported
	    					fileExported = true;
	    				} else { // otherwise, if an error occurred
							// show the error to the user
							elmView.showErrorMessage("Unable to export the document. \n" 
													+ "A problem with the selected file occurred. \n" 
													+ "Please try again selecting another output file.");
	    					// if we created the file, delete it
							if (newFileCreated) {		            	
								try {
									Files.delete(targetFile);
								} catch (IOException e) {
									/* we tried to delete the file that we created when we saw that the selected file did not exist,
									 * but we are not able to. No worry about that, if we cannot, we leave the file where the user
									 * decided to export it, because anyway we are going to inform him about the saving error.
									 * We just left the empty file there, as was going to happen without this try */
								}
			            	}
	    				}	    				
		    		}		   
		    	}
			} else { // for every other type of document
				elmView.showInfoMessage("The selected document cannot be exported by this application.");
			}
			// at the end of the export process (also the recursive ones), clear the file selection
	    	elmView.getFileChooser().setSelectedFile(new File(""));
		} else { // if no internal frames is selected 
			elmView.showInfoMessage("There is not a selected document to export.");
		}
		return fileExported; // true if we have successfully exported the file, false otherwise
    } 
    
    /**
     * This method, if an <code>EventLogEditor</code> frame is selected, cuts the
     * selected text, if present, into the system clipboard.
     * 
     * <p>The implementation of the this procedure is made directly by the 
     * <code>EventLogEditor</code> class, this method simply calls the 
     * {@link unipi.bpm.elm.santeramo.view.EventLogEditor#cut() cut()} method
     * of the selected frame to pass it the cut command to process. 
     */
    public void cutEventLog() {
    	// take the selected internal frame
		JInternalFrame selectedInternalFrame = elmView.getContentArea().getSelectedFrame();
		// if an internal frame is selected
		if (selectedInternalFrame != null) {
			// if the selected frame is an EventLogEditor
			if (selectedInternalFrame instanceof EventLogEditor) {
				// get the editor
				EventLogEditor eventLogEditor = (EventLogEditor)selectedInternalFrame;
				// cut the selected text, if present, into the system clipboard
				eventLogEditor.cut();				
			}// do nothing if other frame types are selected  
		}// do nothing if there are no selected frames  
    }
    
    /**
     * This method, if an <code>EventLogEditor</code> frame is selected, copies the
     * selected text, if present, into the system clipboard.
     * 
     * <p>The implementation of the this procedure is made directly by the 
     * <code>EventLogEditor</code> class, this method simply calls the 
     * {@link unipi.bpm.elm.santeramo.view.EventLogEditor#copy() copy()} method
     * of the selected frame to pass it the copy command to process. 
     */
    public void copyEventLog() {
    	// take the selected internal frame
		JInternalFrame selectedInternalFrame = elmView.getContentArea().getSelectedFrame();
		// if an internal frame is selected
		if (selectedInternalFrame != null) {
			// if the selected frame is an EventLogEditor
			if (selectedInternalFrame instanceof EventLogEditor) {
				// get the editor
				EventLogEditor eventLogEditor = (EventLogEditor)selectedInternalFrame;
				// copy the selected text, if present, into the system clipboard
				eventLogEditor.copy();				
			}// do nothing if other frame types are selected   
		}// do nothing if there are no selected frames  
    }
    
    /**
     * This method, if an <code>EventLogEditor</code> frame is selected, pastes the
     * content of the system clipboard in the <code>EventLogEditor</code> text area.
     * 
     * <p>The implementation of the this procedure is made directly by the 
     * <code>EventLogEditor</code> class, this method simply calls the 
     * {@link unipi.bpm.elm.santeramo.view.EventLogEditor#paste() paste()} method
     * of the selected frame to pass it the paste command to process. 
     */
    public void pasteEventLog() {
    	// take the selected internal frame
		JInternalFrame selectedInternalFrame = elmView.getContentArea().getSelectedFrame();
		// if an internal frame is selected
		if (selectedInternalFrame != null) {
			// if the selected frame is an EventLogEditor
			if (selectedInternalFrame instanceof EventLogEditor) {
				// get the editor
				EventLogEditor eventLogEditor = (EventLogEditor)selectedInternalFrame;
				// paste the content of the system clipboard in the editor text area
				eventLogEditor.paste();				
			}// do nothing if other frame types are selected  
		}// do nothing if there are no selected frames  
    }
    
    /**
     * This method, if an <code>EventLogEditor</code> frame is selected, undo the
     * last action done inside the <code>EventLogEditor</code> text area.
     * 
     * <p>The implementation of the this procedure is made directly by the 
     * <code>EventLogEditor</code> class, this method simply calls the 
     * {@link unipi.bpm.elm.santeramo.view.EventLogEditor#undo() undo()} method
     * of the selected frame to pass it the undo command to process. 
     */
    public void undoEventLog() {
    	// take the selected internal frame
		JInternalFrame selectedInternalFrame = elmView.getContentArea().getSelectedFrame();
		// if an internal frame is selected
		if (selectedInternalFrame != null) {
			// if the selected frame is an EventLogEditor
			if (selectedInternalFrame instanceof EventLogEditor) {
				// get the editor
				EventLogEditor eventLogEditor = (EventLogEditor)selectedInternalFrame;
				// undo the last edit
				try {
					eventLogEditor.undo();
			    } catch (CannotUndoException e) {
			    	elmView.showErrorMessage("Unable to undo the last action. Please try again.");
				}							
			}// do nothing if other frame types are selected  
		}// do nothing if there are no selected frames  
    }
    
    /**
     * This method, if an <code>EventLogEditor</code> frame is selected, redo the
     * last action undone inside the <code>EventLogEditor</code> text area.
     * 
     * <p>The implementation of the this procedure is made directly by the 
     * <code>EventLogEditor</code> class, this method simply calls the 
     * {@link unipi.bpm.elm.santeramo.view.EventLogEditor#redo() redo()} method
     * of the selected frame to pass it the redo command to process. 
     */
    public void redoEventLog() {
    	// take the selected internal frame
		JInternalFrame selectedInternalFrame = elmView.getContentArea().getSelectedFrame();
		// if an internal frame is selected
		if (selectedInternalFrame != null) {
			// if the selected frame is an EventLogEditor
			if (selectedInternalFrame instanceof EventLogEditor) {
				// get the editor
				EventLogEditor eventLogEditor = (EventLogEditor)selectedInternalFrame;
				// redo the last edit
				try {
					eventLogEditor.redo();
			    } catch (CannotRedoException e) {
			    	elmView.showErrorMessage("Unable to redo the last undone action. Please try again.");
				}							
			}// do nothing if other frame types are selected  
		}// do nothing if there are no selected frames  
    }     
    
    /**
     * Increases by 1pt the font size of the content of the selected 
     * <code>EventLogEditor</code> editor (that is inside the 
     * {@link unipi.bpm.elm.santeramo.view.EventLogManagerView#contentArea contentArea}).
     * 
     * @see EventLogEditor
     */
    public void zoomInEventLog() {
    	// take the selected internal frame
		JInternalFrame selectedInternalFrame = elmView.getContentArea().getSelectedFrame();
		// if an internal frame is selected
		if (selectedInternalFrame != null) {
			// if the selected frame is an EventLogEditor
			if (selectedInternalFrame instanceof EventLogEditor) {
				EventLogEditor eventLogEditor = (EventLogEditor)selectedInternalFrame;				
				// get the font size
				int actualFontSize = eventLogEditor.getFontSize();
				// increase the font size by 1
				int newFontSize = actualFontSize + 1;
				// do not increase over 1000
				if (newFontSize > 1000) {
					newFontSize = 1000;
				}
				// Zoom In the font size
				eventLogEditor.setFontSize(newFontSize);
			}// do nothing if other frame types are selected  
		}// do nothing if there are no selected frames   	
    }
    
    /**
     * Decreases by 1pt the font size of the content of the selected 
     * <code>EventLogEditor</code> editor (that is inside the 
     * {@link unipi.bpm.elm.santeramo.view.EventLogManagerView#contentArea contentArea}).
     * 
     * @see EventLogEditor
     */
    public void zoomOutEventLog() {
    	// take the selected internal frame
		JInternalFrame selectedInternalFrame = elmView.getContentArea().getSelectedFrame();
		// if an internal frame is selected
		if (selectedInternalFrame != null) {
			// if the selected frame is an EventLogEditor
			if (selectedInternalFrame instanceof EventLogEditor) {
				EventLogEditor eventLogEditor = (EventLogEditor)selectedInternalFrame;				
				// get the font size
				int actualFontSize = eventLogEditor.getFontSize();
				// decrease the font size by 1
				int newFontSize = actualFontSize - 1;
				// do not decrease under 1
				if (newFontSize < 1) {
					newFontSize = 1;
				}
				// Zoom Out the font size
				eventLogEditor.setFontSize(newFontSize);
			}// do nothing if other frame types are selected  
		}// do nothing if there are no selected frames   	
    }
    
    /**
     * Reset to the default value the font size of the content of the selected 
     * <code>EventLogEditor</code> editor (that is inside the 
     * {@link unipi.bpm.elm.santeramo.view.EventLogManagerView#contentArea contentArea}).
     * 
     * @see EventLogEditor
     */
    public void resetZoomEventLog() {
    	// take the selected internal frame
		JInternalFrame selectedInternalFrame = elmView.getContentArea().getSelectedFrame();
		// if an internal frame is selected
		if (selectedInternalFrame != null) {
			// if the selected frame is an EventLogEditor
			if (selectedInternalFrame instanceof EventLogEditor) {
				EventLogEditor eventLogEditor = (EventLogEditor)selectedInternalFrame;
				// reset the zoom by setting it to the default value
				eventLogEditor.setDefaultFontSize();
			}// do nothing if other frame types are selected  
		}// do nothing if there are no selected frames   	
    }
    
}
