/*
 * Event Log Manager
 *   
 *  Copyright (C) 2015 Francesco Santeramo
 *
 *  This file is part of Event Log Manager.
 *
 *  Event Log Manager is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Event Log Manager is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *	
 *  Author:  Francesco Santeramo 
 *  Website: www.francescosanteramo.com
 *  Contact: eventlogmanager@francescosanteramo.com    
 */

package unipi.bpm.elm.santeramo.controller;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.Stack;

/**
 * Utility class used to generate a set of strings representing all the possible
 * combination of a set of elements, passed using an infix expression that
 * defines the possible structure of the generated strings.
 * 
 * <p>The expression is composed by characters and strings for the definition of
 * elements, the operator + to define the sum of two set of elements,
 * the operator | to define the interleaving between two set of elements and
 * spaces between all of these to define elements composed by multiple characters.
 * 
 * <p>Here an example of a possible expression: a b c + ( ef g | 123 4 5). The
 * element "ef" and the element "123" are a single block, while for example "4" and "5" are
 * two different element of the same set, because they are divided by a space. Note that
 * operators define also the division between different sets, as they are operation between
 * them.
 * 
 * @author Francesco Santeramo (eventlogmanager@francescosanteramo.com)
 */
public class EventLogGenerator {
	
	/*
	 * VARIABLES
	 */
	
	/**
	 * This variable defines a constant that symbolises
	 * the left associativity of an operator.
	 */
	private int leftAssociative;
	
	/**
	 * This variable defines a constant that symbolises
	 * the right associativity of an operator.
	 */
	private int rightAssociative;
		
	/**
	 * This is the <code>Map</code> of the possible operators,
	 * together with their precedence and left/right associativity.
	 * 
	 * <p>The set is composed by: ("operator", []{precedence, associativity})
	 */
	private Map<String, int[]> operators; 
	
	
	
	/*
	 * CONSTRUCTOR
	 */
	
	/**
	 * Creates an <code>EventLogGenerator</code>, defining the value of the
	 * {@link leftAssociative} and {@link rightAssociative} constants and
	 * setting the possible operators and their precedence and left/right
	 * associativity.
	 */
	public EventLogGenerator() {	
		// set the constants value
		leftAssociative = 0;
		rightAssociative = 1;
		// set the possible operators and their precedence and left/right associativity
		operators = new HashMap<String, int[]>();
		operators.put("+", new int[] { 0, leftAssociative });  
		operators.put("|", new int[] { 5, leftAssociative });           
	}
	
	
	
	/*
	 * METHODS
	 */
	 
    /**
     * Returns <code>true</code> if the passed <code>String</code> is an operator
     * (defined in the {@link operators} map), <code>false</code> otherwise.
     * 
     * @param token the <code>String</code> to test
     * 
     * @return <code>true</code> if the passed <code>String</code> is an operator, <code>false</code> otherwise.
     */
	private boolean isOperator(String token) {  
        return operators.containsKey(token);  
    }
    
	/**
	 * Returns <code>true</code> if the passed <code>String</code> is an operator and
	 * its associativity value is the same as the one passed has parameter, <code>false</code>
	 * if it is an operator but its associativity is different from the one passed has parameter.
	 * 
	 * @param token the <code>String</code> representing the operator to check 
	 * @param associativity the constant value that defines left/right associativity
	 * 
	 * @return <code>true</code> if the operator has the specified associativity, <code>false</code> otherwise.
	 */
    private boolean isAssociative(String token, int associativity) {  
    	// if the token is not an operator
        if (!isOperator(token)) {  
        	// throw an unchecked exception 
            throw new IllegalArgumentException("Invalid token: " + token);  
        }  
        // if the token is an operator, check its associativity
        if (operators.get(token)[1] == associativity) {  
        	// if it is the one specified by the parameter
            return true;  
        }  
        // otherwise
        return false;  
    }
    
    /**
	 * Compares the precedence of the two passed operators (if one of them or both are not
	 * operators, an unchecked exception will be thrown), to understand which one has the 
	 * precedence during the calculation of the expression.
     * 
     * @param token1 the <code>String</code> representing the first operator to compare
     * @param token2 the <code>String</code> representing the second operator to compare
     * 
     * @return a positive value if the first operator has precedence over the second one; 
     * 			a negative value if the second operator has precedence over the first one;
     * 			0 if their precedence is the same and they can be processed either way.
     */
    private int getPrecedence(String token1, String token2) {  
    	// if one of the two parameters is not an operator
        if (!isOperator(token1) || !isOperator(token2)) { 
        	// throw an unchecked exception
            throw new IllegalArgumentException("Invalid tokens: " + token1 + " " + token2);  
        }  
        // return the difference from the precedence of the first passed operator and the precedence of the second one
        return operators.get(token1)[0] - operators.get(token2)[0];  
    }       
    
    /**
     * Converts the passed expression from infix notation to Reverse Polish Notation 
     * (also known as postfix notation) using the shunting-yard algorithm, invented
     * by Edsger Dijkstra.
     * 
     * @param infixExpression an array containing the elements of the expression in infix notation order (divided in symbols, operators and parentheses)
     * 
     * @return an array containing the elements of the expression in postfix notation order (divided in symbols and operators and without parentheses).
     */
    private String[] infixToPostfix(String[] infixExpression)   
    {  
    	// the output array where to add the elements in postfix notation
        ArrayList<String> output = new ArrayList<String>();  
        // the stack to use while calculating the postfix notation
        Stack<String> stack = new Stack<String>();            
        // for every token of the expression
        for (String token : infixExpression) {  
            // if the token is an operator  
            if (isOperator(token)) {    
                // while stack not empty AND stack top element is an operator  
                while (!stack.empty() && isOperator(stack.peek())) { 
                	// if the operator has minor (or equal, in case of left associative) precedence than the operator on top of the stack
                    if ((isAssociative(token, leftAssociative) && getPrecedence(token, stack.peek()) <= 0) ||   
                        (isAssociative(token, rightAssociative) && getPrecedence(token, stack.peek()) < 0)) {  
                    	// add to the output the top element of the stack (an operator)
                    	output.add(stack.pop());     
                    	// go to check next top element
                        continue;  
                    }  
                    // stop if the top element is not an operator or the stack is empty
                    break;  
                }  
                // push the new operator on the stack  
                stack.push(token);  
            } else if (token.equals("(")) { // else if the token is a left parenthesis                                
                // push the parenthesis on the stack
            	stack.push(token);   
            } else if (token.equals(")")) { // else if the token is a right parenthesis  
            	// while stack not empty AND stack top element is not a left parenthesis
                while (!stack.empty() && !stack.peek().equals("(")) {  
                	// add the elements from the stack to the postfix output
                	output.add(stack.pop());   
                }  
                // remove the left parenthesis from the stack
                stack.pop();   
            } else { // if the token is a symbol 
            	output.add(token);   
            }  
        }
        //at the end we move the last remaining elements from the stack to the postfix output
        while (!stack.empty()) {  
        	// add to the postfix output the remaining stack elements
        	output.add(stack.pop());   
        }  
        // get the output array and return it
        String[] result = new String[output.size()];  
        return output.toArray(result);  
    } 

    /**
     * This method implements the sum of two parts of the expression. 
     * 
     * The operation is expressed by the "+" operator.
     * 
     * <p>The two passed string are a representation of two parts of the expression, in which
     * there are sets of strings divided by a + sign and that contain single elements
     * divided by spaces.
     * 
     * <p>The method simply returns a <code>String</code> generated by the sum of the
     * two passed string with a "+" sign between them. This shows to the {@link #calculateExpression(String[])}
     * method that this is a set of elements composed by the two passed as parameter, and not just one
     * single element. The method will process this string accordingly to this information.
     * 
     * @param string1 the first <code>String</code> of the new set that represents the left part of the expression to sum
     * @param string2 the second <code>String</code> of the new set that represents the righ part of the expression to sum
     * 
     * @return a <code>String</code> composed by the two passed <code>String</code> merged together with a "+" sign between them
     */
    private String plus(String string1, String string2) {
    	// return the two passed string merged together with a "+" sign between them
    	return string1 + "+" + string2;
    }
	
    /**
     * This method implements the interleaving between two parts of the expression, that 
     * are composed by sets of strings connected together by the + sign and that contain
     * single elements separated by spaces.
     * 
     * The operation is expressed by the "|" operator.
     * 
     * <p>The method returns a <code>String</code> composed by the sum ("+" between strings)
     * of all the generated strings by the interleaving operation.
     * 
     * <p>The two passed strings can be simple strings, set of strings (with spaces between elements),
     * sum of strings ("+" between strings) or sum of string sets.
     * The method manage all the cases. In particular, when a + is present, the method splits the strings
     * in groups over the + sign, then it processes the interleaving operation between all the possible
     * combination of these groups.
     * 
     * @param string1 the first <code>String</code> for the interleaving operation. It can be a set of strings or a sum of strings or string sets ("+" between strings)
     * @param string2 the second <code>String</code> for the interleaving operation. It can be a set of strings or a sum of strings or string sets ("+" between strings)
     * 
     * @return a <code>String</code> composed by the sum ("+" between strings) of all the generated strings by the interleaving operation.
     */
    private String interleaving(String string1, String string2) {
    	// get here the result of the interleaving
    	StringBuffer result = new StringBuffer(); 
    	// split the left and right strings of the interleaving to get all the elements
    	String[] left = string1.split("[+]");
    	String[] right = string2.split("[+]");
    	// interleave all the possible combination of elements of the two strings
    	for (int i = 0; i < left.length; i++) {
    		for (int j = 0; j < right.length; j++) {
    			// add to the final result the generated set of strings using the interleaving operator
    			if (result.length()!=0) result.append("+");
    			result.append(interleaveSets(left[i], right[j]));
        	}
    	}
    	// return the final string composed by the sum ("+" between strings) of all the generated strings
    	return result.toString();
    }
    
    /**
     * This method implements the interleaving operation between two string sets 
     * where single elements are divided inside the string by spaces.
     * 
     * <p>The output is composed by all the possible interleaved strings between 
     * these two sets, separated by a "+" sign.
     *  
     * @param string1 the first <code>String</code> set to interleave
     * @param string2 the second <code>String</code> set to interleave
     * 
     * @return a <code>StringBuffer</code> representing a string of all possible interleaved solutions between the two passed sets.
     */
    private StringBuffer interleaveSets(String string1, String string2) {
    	// get the output
	    StringBuffer output = new StringBuffer();
	    // divide the two passed strings in arrays of elements
     	ArrayList<String> array1 = new ArrayList<String>(Arrays.asList(string1.split("\\s+")));
    	ArrayList<String> array2 = new ArrayList<String>(Arrays.asList(string2.split("\\s+")));
    	// interleave the two sets of strings
    	ArrayList<ArrayList<String>> result = interleaveElements(array1,array2);
    	// at the end, for every list in the main list
    	for (int i=0; i<result.size(); i++) {
    		// add a + to connect lists
    		if (i!=0) output.append("+");
    		// get the list
    		ArrayList<String> trace = result.get(i);
    		// transform the list in a string of elements, divided by space
        	for (int j=0; j<trace.size(); j++) {
	        	output.append(trace.get(j));
	        	if (j != trace.size()-1) output.append(" ");
	        }
	    }
    	// return the generated set of strings
    	return output;
	} 
    
    /**
     * This method implements the interleaving operation between two string sets 
     * where single elements are contained in two arrays.
     * 
     * <p>The output is composed by all the possible interleaved strings between 
     * these two sets, separated by a "+" sign.
	 *
     * @param array1 the <code>ArrayList</code> containing the elements of the first set
     * @param array2 the <code>ArrayList</code> containing the elements of the second set
     * 
     * @return an <code>ArrayList</code> of <code>ArrayList</code> representing the list of all the generated traces (i.e. list of interleaved elements).
     */
    private ArrayList<ArrayList<String>> interleaveElements(ArrayList<String> array1, ArrayList<String> array2) {
    	// get the output of this interleaving operation
    	ArrayList<ArrayList<String>> output =  new ArrayList<ArrayList<String>>();
    	// if the lists are not null, otherwise return an empty list
    	if (array1 != null && array2 != null) {
    		// if the first list is empty
    		if (array1.size()==0) {
    			// add to the output the second list if not empty
    			if (array2.size()!=0) output.add(new ArrayList<String>(array2));
    		} else if (array2.size()==0) { // else if the second list is empty
    			// add to the output the first list
    			output.add(new ArrayList<String>(array1));
    		} else { // if both lists are not empty
    			// get the first element of list one
    			String prefix = array1.get(0);
    			// the remaining part is the suffix
    			ArrayList<String> suffix = new ArrayList<String>(array1.subList(1, array1.size()));
    			// calculate the first half of the final output, calling the interleaving method recursively on this subset
    			ArrayList<ArrayList<String>> output1 = interleaveElements(suffix, array2);
    			// append the prefix on top of the generated lists
    			appendAll(prefix,output1);
    			// add the modified partial output to the final one
    			output.addAll(output1);
    			// get the first element of list two
    			prefix = array2.get(0);
    			// the remaining part is the suffix
    			suffix = new ArrayList<String>(array2.subList(1, array2.size()));
    			// calculate the second half of the final output, calling the interleaving method recursively on this subset
    			ArrayList<ArrayList<String>> output2 = interleaveElements(array1,suffix);
    			// append the prefix on top of the generated lists
    			appendAll(prefix,output2);
    			// add the modified partial output to the final one
    			output.addAll(output2);
    		}
    	}
    	// return the generated output
    	return output;
    }     
    
    /**
     * Adds on top of every list that is inside the passed <code>ArrayList</code>
     * the string elements passed as parameter.
     * 
     * @param s the <code>String</code> element to add on top of all the passed lists
     * @param output the <code>ArrayList</code> containing the <code>ArrayList</code> that represent list of string elements
     */
    private void appendAll(String s, ArrayList<ArrayList<String>> output) {
    	// for every list
    	for (ArrayList<String> list : output)
    		// add the passed element on top of the list
    		list.add(0,s);
    }
    
    /**
     * Given an expression in postfix notation that defines a set of event traces,
     * the method returns an array of all possible traces that can be generated 
     * from this expression.
     * 
     * @param postfixExpression a <code>String</code> array containing an expression in postfix notation, divided in single elements to be processed
     * 
     * @return a <code>String</code> array of all possible strings that can be generated from the passed expression.
     * 
     * @throws Exception if an error occurred during the process, an exception will be thrown to inform the main method and stop the process
     */
    private String[] calculateExpression(String[] postfixExpression) throws Exception {      
    	// a stack where to place elements while calculating the expression
        Stack<String> stack = new Stack<String>();           
        // for each token of the expression   
        for (String token : postfixExpression) {  
            // if the token is a value 
            if (!isOperator(token)) {
            	// push it onto the stack 
                stack.push(token);                  
            } else { // if the token is an operator
                // pop top two entries  
                String string2 = stack.pop();  
                String string1 = stack.pop();                    
                // calculate the result
                String result;
                if (token.compareTo("+") == 0) {
                	result = plus(string1, string2);
                } else if (token.compareTo("|") == 0) {
                	result = interleaving(string1, string2);
                } else {
                	// throw a specific exception
        			throw new Exception("Invalid expression: the operator \""+token+"\" is not permitted.");
                }                                  
                // push the result onto the stack  
                stack.push(result);                                                  
            }                          
        }          
         // pop from the stack the result of the expression (the only value remained inside the stack is the result)
        String result = stack.pop();  
        // return an array composed by all the generated strings (here we calculate the + sign as a definition of a set of strings)
        return result.split("[+]");
    }
    
    /**
     * Given an expression that defines a set of event traces, the method first checks if it is a valid
     * expression (not empty, equal number of left and right parentheses and the use of only these symbols
     * [ A-Z a-z 0-9 ( ) + | ]), then transforms it from infix notation to postfix notation, and at the end
     * generates from the postfix expression the result containing all the possible strings defined by the expression.
     * 
     * <p>In the expression sequences of events are expressed using a character for every event from the set [ A-Z a-z 0-9 ,
     * possible operators are "+" (plus) and "|" (interleaving) and the use of left and right parentheses for defining
     * operation precedence is admitted.
     * 
     * @param exp the <code>String</code> representing the expression that defines a set of event traces to generate
     * 
     * @return an array containing all the generated traces from the passed expression
     * 
     * @throws Exception when an error occurred during the expression check or the generation of the result, an exception is thrown, containing as message the description of the error
     */
    public String[] generateTraces(String exp) throws Exception {    	
    	// get the expression removing leadin and trailing spaces and making multiple spaces singles
    	String expression = exp.trim().replaceAll(" +", " ");
		// if the expression is empty
		if (expression.isEmpty()) {
			// throw a specific exception
			throw new Exception("Invalid expression: no character entered.");
		}		
		// get the number of left and right parentheses
		int leftParentheses = expression.length() - expression.replace("(", "").length();
		int rightParentheses = expression.length() - expression.replace(")", "").length();
		// if the number of left and right parentheses is not equal
		if (leftParentheses != rightParentheses) {
			// get the difference
			int difference = leftParentheses - rightParentheses;			
			StringBuilder error = new StringBuilder("there");	
			// if the difference is positive, the left parentheses are more than the right ones
			if (difference > 0) {
				// if there is one more left parenthesis than the right ones
				if (difference == 1){
					error.append(" is 1");
				} else { // if the left parentheses are more than one
					error.append(" are ").append(difference);
				}
				error.append(" more left parenthesis.");
			} else { // if negative, the right parentheses are more than the left ones
				// get the absolute value of the difference
				int posDiff = Math.abs(difference);
				// if there is one more right parenthesis than the left ones
				if (posDiff == 1){
					error.append(" is 1");
				} else { // if the right parentheses are more than one
					error.append(" are ").append(posDiff);
				}
				error.append(" more right parenthesis.");
			}
			// throw a specific exception
			throw new Exception("Invalid expression: " + error.toString());
		}
		// if the expression contains character not included in [ A-Z a-z 0-9 + | ( ) ]	
		if (!expression.matches("^[a-zA-Z0-9|+() ]*$")) {
			// throw a specific exception
			throw new Exception("Invalid expression: use only [ A-Z a-z 0-9 ] as symbols, [ | + ] as operators, \n"
								+ "spaces to divide elements of a set and parentheses to divide operations.");
		}
		/* now that we have a valid expression, we divide the string in elements inside an array.
		 * We split the string keeping all its elements: we put in different elements of the array
		 * operators, parentheses and adjacent symbols */
        String[] elements = expression.split("((?<=[()+|])|(?=[()+|]))");
        
        ArrayList<String> infixArray = new ArrayList<String>();
        // here we eliminate the possible spaces between symbols and operators, but we keep spaces between symbols (they define different elements)
        // we also remove elements that after the trim are empty, i.e. spaces between operators
        for (int i=0; i<elements.length; i++) {
        	String element = elements[i].trim();
        	if (!element.isEmpty()) {
        		infixArray.add(element);
        	}
        }
        // transform the ArrayList to an array to get the infix expression in it
        String[] infixExp = infixArray.toArray(new String[infixArray.size()]);

        // transform the infix expression in a postfix one, passing its elements to this method
        String[] postfixExp = infixToPostfix(infixExp);  

        // pass the postfix expression to this method for calculating all its possible solutions
        String[] result = calculateExpression(postfixExp);

        // return the array containing all possible solutions     
        return result;    	
    }
}
