/*
 * Event Log Manager
 *   
 *  Copyright (C) 2015 Francesco Santeramo
 *
 *  This file is part of Event Log Manager.
 *
 *  Event Log Manager is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Event Log Manager is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *	
 *  Author:  Francesco Santeramo 
 *  Website: www.francescosanteramo.com
 *  Contact: eventlogmanager@francescosanteramo.com    
 */

package unipi.bpm.elm.santeramo.model;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JInternalFrame;

/**
 * This class defines the Model  of the Application 
 * (This is the Model of the MVC pattern).
 * 
 * @author Francesco Santeramo (eventlogmanager@francescosanteramo.com)
 */

public class EventLogManagerModel {
	
	
	/*
	 * UTILITY OBJECTS AND VARIABLES
	 */
	
	/**
	 * This number counts how many <code>JInternalFrame</code> get opened during
	 * the current application session. 
	 * 
	 * <p>It also acts as an identifier for the new <code>JInternalFrame</code> frame 
	 * just created: the system increases by 1 this number when it creates a new
	 * <code>JInternalFrame</code> and then uses it as a unique name for the new frame.
	 * 
	 * <p>On this way, the internal frame has an unique number associated with it and can
	 * be connected to a <code>FooterAreaButton</code> using this identifier.
	 * 
	 * @see JInternalFrame
	 */
	private int internalFrameNumber;
	
	/**
	 * This variable, instance of the <code>MostRecentElementList</code> class, represents a 
	 * <code>LinkedList</code> object for managing a limited list of <code>String</code>
	 * representing absolute file paths (paths associated to the last opened files in the
	 * application). 
	 * 
	 * <p>The list will be composed by unique items, in order of insert: if we add a string, 
	 * this will be added to the top of the list; if that string was already in the list, 
	 * we delete the old one and add the new on top (we move that item on top of the list
	 * on this way).
	 */
	private MostRecentElementList<String> recentLogList;


	
	/*
	 * CONSTRUCTOR
	 */
	
	/**
	 * Constructs a new <code>EventLogMangerModel</code>, 
	 * initialising its instance variables.
	 */
	public EventLogManagerModel() {		
		// utility objects and variables
		internalFrameNumber = 0;
		recentLogList = new MostRecentElementList<>(0);
	}
	
	
	
	/*
	 * METHODS
	 */
	
	/**
	 * This method increases by 1 the numeric value of the {@link #internalFrameNumber} variable. 
	 */
	public void addNewInternalFrame() {		
		internalFrameNumber++;
	}
	
	/**
	 * Returns the numeric value of the {@link #internalFrameNumber} variable.
	 * 
	 * @return the numeric value of the <code>internalFrameNumber</code> variable.
	 */
	public int getInternalFrameNumber() {		
		return internalFrameNumber;
	}
	
	/**
	 * Adds a new <code>String</code>, representing an absolute file path to record, on top of the
	 * {@link recentLogList} list. 
	 * 
	 * <p>If the path was already in the list, the process deletes the old one and add the
	 * new on top of the list. 
	 * 
	 * <p>This method simply calls the {@link MostRecentElementList#addNewElement(Object) addNewElement(Object)}
	 * method of the {@link recentLogList}, that will implements the explained behaviour.
	 * 
	 * @param path the <code>String</code> absolute file path to record
	 */
	public void addRecentLog(String path) {
		recentLogList.addNewElement(path);
	}
	
	/**
	 * Returns the first <code>n</code> elements of the {@link #recentLogList} list 
	 * (instance of the <code>MostRecentElementList</code> class) in a <code>List</code>
	 * list object, or the entire {@link #recentLogList} if its elements are less than <code>n</code>. 
	 * 
	 * The number of how many elements will be included in the returned list is defined by the
	 * elm configuration file. The method uses the {@link #getRecentListNumber()} method to get it.
	 * 
	 * @return the first <code>n</code> elements of the {@link #recentLogList} list, that contains <code>String</code> values.
	 * 
	 * @see MostRecentElementList
	 * @see LinkedList
	 */
	public List<String> getRecentLogs() {
		// get the number of elements to return
		int endPoint = getRecentListNumber();
		// if the size of the list is less than the number to return
		if (recentLogList.size() < getRecentListNumber()) {
			// set the number to the list size
			endPoint = recentLogList.size();
		}
		// get a substring of the recent list with as many elements as defined
		return recentLogList.subList(0, endPoint);
	}
	
	/**
	 * Returns <code>true</code> if the {@link #recentLogList} has elements, <code>false</code> otherwise.
	 * 
	 * @return <code>true</code> if the {@link #recentLogList} has elements, <code>false</code> otherwise.
	 * 
	 * @see MostRecentElementList
	 */
	public boolean hasRecentLogs() {
		return (recentLogList.size() != 0);
	}
	
	/**
	 * Gets the value of the passed key that resides inside
	 * the configuration file defined by the id passed as parameter.
	 * 
	 * @param configID the id of the configuration file (i.e. the name of the file without extension)
	 * @param configKey the key to search inside the file
	 * 
	 * @return the value read from the given key inside the selected configuration file.
	 */
	private String getConfigValue(String configID, String configKey) {
		String filePath;
		// get the path of the selected config file
		switch (configID) {
	    	case "xes": 
	    				filePath = "./config/xes.txt";
						break;
	    	case "elm": 
	    				filePath = "./config/elm.txt";
						break;
	    	default:	
	    				filePath = "";
	    				break;
		}		
		// if we have a config file for that id
		if (!filePath.isEmpty()) {
			Path targetConfig = Paths.get(filePath);
			//check if we can read the file
			if (Files.exists(targetConfig)) {
		    	// try to read from the file
		    	// try-with-resources Statement (close the resource automatically at the end in any case)
				try (InputStream fileStream = new FileInputStream(filePath);
						BufferedReader fileReader = new BufferedReader(new InputStreamReader(fileStream))) {   					
					String line = null;
					// for every line of the file
					while ((line = fileReader.readLine()) != null) {
						// trim the line
						String trimLine = line.trim();
						// if the line starts with the key
						if (trimLine.startsWith(configKey)) {
							// we have found the line where is the value
					    	String[] config = trimLine.split("=", 2);
					    	// get the value associated to the passed key and check if it exists
					    	if (config.length > 1) {
					    		// if positive, return it
					    		return config[1];
					    	} else { 
					    		return null;
					    	}
					    }
	    	        }
					// if we ended reading the file without have found the given key
					return null;					
				} catch(IOException e) {
					return null;
				}
			} else { // if there is not a configuration file to read
				filePath = filePath.substring(1);
				// read from the default configuration file, that is inside the src
				try (InputStream fileStream = this.getClass().getResourceAsStream(filePath);
						BufferedReader fileReader = new BufferedReader(new InputStreamReader(fileStream))) {   					
					String line = null;
					// for every line of the file
					while ((line = fileReader.readLine()) != null) {
						// trim the line
						String trimLine = line.trim();
						// if the line starts with the key
						if (trimLine.startsWith(configKey)) {
							// we have found the line where is the value
					    	String[] config = trimLine.split("=", 2);
					    	// get the value associated to the passed key and check if it exists
					    	if (config.length > 1) {
					    		// if positive, return it
					    		return config[1];
					    	} else {
					    		return null;
					    	}
					    }
	    	        }
					return null;					
				} catch(IOException e) {
					return null;
				}
			}				
		} else {
			return null;
		}		
	}
	
	/**
	 * Sets the value of the passed key that resides inside
	 * the configuration file defined by the id passed as parameter,
	 * using the value string.
	 * 
	 * @param configID the id of the configuration file (i.e. the name of the file without extension)
	 * @param configKey the key to search inside the file
	 * @param value the <code>String</code> value to set
	 * 
	 * @return <code>true</code> if the method correctly set the value, <code>false</code> otherwise.
	 */
	private boolean setConfigValue(String configID, String configKey, String value) {
		String filePath;
		// get the path of the selected config file
		switch (configID) {
	    	case "xes": 
	    				filePath = "./config/xes.txt";
						break;
	    	case "elm": 
	    				filePath = "./config/elm.txt";
						break;
	    	default:	
	    				filePath = "";
	    				break;
		}		
		// if we have a config file for that id
		if (!filePath.isEmpty()) {
			Path targetConfig = Paths.get(filePath);
			//check if we can read the file
			if (Files.exists(targetConfig)) {
				// try to read from the file
		    	// try-with-resources Statement (close the resource automatically at the end in any case)
				try (InputStream inputStream = new FileInputStream(filePath);
						BufferedReader fileReader = new BufferedReader(new InputStreamReader(inputStream))) {   												
					// get the line separator specific of the machine system
					String newline = System.getProperty("line.separator");  			        
					StringBuffer fileContent = new StringBuffer();	
					String line = null;
					// for every line of the file
					while ((line = fileReader.readLine()) != null) {
						// trim the line
						String trimLine = line.trim();
						// if the line starts with the key
						if (trimLine.startsWith(configKey)) {
							// change the value of that key with the passed string
					    	trimLine = configKey + "=" + value.trim();
					    }		
						// append the line to the buffer
						fileContent.append(trimLine);
	    	        	fileContent.append(newline);
	    	        }
					// get the file content from th buffer and write it back to the file
	    	        String content = fileContent.toString();
		    		Files.write(targetConfig, content.getBytes());
		    		return true;		    		
				} catch(IOException e) {
					return false;
				}
			} else {// if we cannot read the file, we cannot store the value anywhere
				return false;
			}	    					
		} else {
			return false;
		}		
	}
	
	/**
	 * Resets all the values of the configuration file defined by
	 * the passed id to their respective default values.
	 * 
	 * @param configID the id of the configuration file (i.e. the name of the file without extension)
	 * 
	 * @return <code>true</code> if the method correctly reset the values, <code>false</code> otherwise.
	 */
	private boolean resetConfigFile(String configID) {
		String filePath;
		switch (configID) {
			// get the path of the selected config file
	    	case "xes": 
	    				filePath = "./config/xes.txt";
						break;
	    	case "elm": 
	    				filePath = "./config/elm.txt";
						break;
	    	default:	
	    				filePath = "";
	    				break;
		}	
		// if we have a config file for that id
		if (!filePath.isEmpty()) {
			Path targetConfig = Paths.get(filePath);
			//check if we can read the file
			if (Files.exists(targetConfig)) {
		    	// try to read from the default config file
		    	// try-with-resources Statement (close the resource automatically at the end in any case)
				String defaultConfigPath = filePath.substring(1);
				try (InputStream inputStream = this.getClass().getResourceAsStream(defaultConfigPath);
						BufferedReader fileReader = new BufferedReader(new InputStreamReader(inputStream))) {   												
					// get the line separator specific of the machine system
					String newline = System.getProperty("line.separator");
					StringBuffer fileContent = new StringBuffer();	
					String line = null;
					// for every line of the default congif file
					while ((line = fileReader.readLine()) != null) {
						// read the line to the buffer
						fileContent.append(line);
	    	        	fileContent.append(newline);
	    	        }
					// get the content of the default config file from the buffer and write it to the target config file
	    	        String content = fileContent.toString();
		    		Files.write(targetConfig, content.getBytes());
		    		return true;		    		
				} catch(IOException e) {
					return false;
				}
			} else { // if we cannot access to the default config file
				return true;
			}				
		} else {
			return false;
		}
	}
	
	/**
	 * Gets the XES Schema URL stored in the xes.txt config file.
	 * 
	 * @return the <code>String</code> representing XES Schema URL stored in the xes.txt config file.
	 */
	public String getXesSchemaUrl() {	
		// get the value
		String result = getConfigValue("xes", "xes-url");
		// if something went wrong, return an empty value
		if (result == null) {
			result = "";
		}
		return result;
	}
	
	/**
	 * Sets the XES Schema URL stored in the xes.txt config file
	 * using the value passed as parameter.
	 * 
	 * @param xesUrl the <code>String</code> value to set
	 * 
	 * @return <code>true</code> if the method correctly set the value, <code>false</code> otherwise.
	 */
	public boolean setXesSchemaUrl(String xesUrl) {
		String value = xesUrl;
		if (value == null) {
			value = "";
		}
		// set the value and return the result state
		return setConfigValue("xes", "xes-url", value);
	}
	
	/**
	 * Gets the XES Schema local path stored in the xes.txt config file.
	 * 
	 * @return the <code>String</code> representing XES Schema local path stored in the xes.txt config file.
	 */
	public String getXesSchemaLocal() {
		// get the value
		String result = getConfigValue("xes", "xes-local");
		// if something went wrong, return an empty value
		if (result == null) {
			result = "";
		}
		return result;
	}
	
	/**
	 * Sets the XES Schema local path stored in the xes.txt config file
	 * using the value passed as parameter.
	 * 
	 * @param xesLocal the <code>String</code> value to set
	 * 
	 * @return <code>true</code> if the method correctly set the value, <code>false</code> otherwise.
	 */
	public boolean setXesSchemaLocal(String xesLocal) {
		String value = xesLocal;
		if (value == null) {
			value = "";
		}
		// set the value and return the result state
		return setConfigValue("xes", "xes-local", value);
	}
	
	/**
	 * Gets the XESEXT Schema URL stored in the xes.txt config file.
	 * 
	 * @return the <code>String</code> representing XESEXT Schema URL stored in the xes.txt config file.
	 */
	public String getXesExtSchemaUrl() {
		// get the value
		String result = getConfigValue("xes", "xesext-url");
		// if something went wrong, return an empty value
		if (result == null) {
			result = "";
		}
		return result;
	}
	
	/**
	 * Sets the XESEXT Schema URL stored in the xes.txt config file
	 * using the value passed as parameter.
	 * 
	 * @param xesExtUrl the <code>String</code> value to set
	 * 
	 * @return <code>true</code> if the method correctly set the value, <code>false</code> otherwise.
	 */
	public boolean setXesExtSchemaUrl(String xesExtUrl) {
		String value = xesExtUrl;
		if (value == null) {
			value = "";
		}
		// set the value and return the result state
		return setConfigValue("xes", "xesext-url", value);
	}
	
	/**
	 * Gets the XESEXT Schema local path stored in the xes.txt config file.
	 * 
	 * @return the <code>String</code> representing XESEXT Schema local path stored in the xes.txt config file.
	 */
	public String getXesExtSchemaLocal() {
		//get the value
		String result = getConfigValue("xes", "xesext-local");
		// if something went wrong, return an empty value
		if (result == null) {
			result = "";
		}
		return result;
	}
	
	/**
	 * Sets the XESEXT Schema local path stored in the xes.txt config file
	 * using the value passed as parameter.
	 * 
	 * @param xesExtLocal the <code>String</code> value to set
	 * 
	 * @return <code>true</code> if the method correctly set the value, <code>false</code> otherwise.
	 */
	public boolean setXesExtSchemaLocal(String xesExtLocal) {
		String value = xesExtLocal;
		if (value == null) {
			value = "";
		}
		// set the value and return the result state
		return setConfigValue("xes", "xesext-local", value);
	}
	
	/**
	 * Resets all the values of the xes.txt configuration file.
	 *  
	 * @return <code>true</code> if the method correctly reset the values, <code>false</code> otherwise.
	 */
	public boolean resetXesConfig() {
		return resetConfigFile("xes");
	}
	
	/**
	 * Gets the recent list number of documents from the elm.txt configuration file.
	 * 
	 * @return the value of the "recent-list" key inside the eml.txt configuration file.
	 */
	public int getRecentListNumber() {
		// get the value
		String result = getConfigValue("elm", "recent-list");
		int resultIntValue;
		int defaultValue = 10;
		// if is not null or empty
		if ((result != null) && (!result.isEmpty())) {
			try {
				// try to parse the string as an integer
				resultIntValue = Integer.parseInt(result);
				// if we succeed, check then if the number is more than 0
				if (resultIntValue <= 0) {
					// otherwise set it to a default value;
					resultIntValue = defaultValue;
				}
			} catch (NumberFormatException e) {
				// if it is not a string representation of an integer
				resultIntValue = defaultValue;
			}
		} else {
			// if the result was null or empty
			resultIntValue = defaultValue;
		}
		// return the calculated result
		return resultIntValue;
	}
	
	/**
	 * Sets the recent list number of documents in the elm.txt configuration file
	 * using the value passed as parameter.
	 * 
	 * @param recentNumber the recent list number of documents to set
	 * 
	 * @return <code>true</code> if the method correctly set the value, <code>false</code> otherwise.
	 */
	public boolean setRecentListNumber(String recentNumber) {
		String value;
		String defaultValue = "10";
		// if the value is not null or empty
		if ((recentNumber != null) && (!recentNumber.isEmpty())) {
			try {
				// try to parse the string as an integer
				int intValue = Integer.parseInt(recentNumber);
				// if we succeed, check then if the number is more than 0
				if (intValue > 0) {
					// yes, so the value is valid to use
					value = recentNumber;
				} else {
					//no, then a default value is set
					value = defaultValue;
				}
			} catch (NumberFormatException e) {
				// if the value is not a string representation of an integer
				value = defaultValue;
			}
		} else {
			// if the result was null or empty
			value = defaultValue;
		}
		// set the calculated value and return the result state
		return setConfigValue("elm", "recent-list", value);
	}
	
	/**
	 * Gets the distance between generated event time-stamp (in seconds) 
	 * from the elm.txt configuration file.
	 * 
	 * @return the value of the "event-timestamp" key inside the eml.txt configuration file.
	 */
	public int getEventTimestampDistance() {
		// get the value
		String result = getConfigValue("elm", "event-timestamp");
		int resultIntValue;
		int defaultValue = 1;
		// if is not null or empty
		if ((result != null) && (!result.isEmpty())) {
			try {
				// try to parse the string as an integer
				resultIntValue = Integer.parseInt(result);
				// if we succeed, check then if the number is more than 0
				if (resultIntValue <= 0) {
					// otherwise set it to a default value;
					resultIntValue = defaultValue;
				}
			} catch (NumberFormatException e) {
				// if the value is not a string representation of an integer
				resultIntValue = defaultValue;
			}
		} else {
			// if the result was null or empty
			resultIntValue = defaultValue;
		}
		// return the calculated result
		return resultIntValue;
	}
	
	/**
	 * Sets the distance between generated event time-stamp (in seconds)
	 * in the elm.txt configuration file using the value passed as parameter.
	 * 
	 * @param eventTimeDist the distance between generated event time-stamp
	 * 
	 * @return <code>true</code> if the method correctly set the value, <code>false</code> otherwise.
	 */
	public boolean setEventTimestampDistance(String eventTimeDist) {
		String value;
		String defaultValue = "1";
		// if the value is not null or empty
		if ((eventTimeDist != null) && (!eventTimeDist.isEmpty())) {
			try {
				// try to parse the string as an integer
				int intValue = Integer.parseInt(eventTimeDist);
				// if we succeed, check then if the number is more than 0
				if (intValue > 0) {
					// yes, so the value is valid to use
					value = eventTimeDist;
				} else {
					//no, then a default value is set
					value = defaultValue;
				}
			} catch (NumberFormatException e) {
				// if the value is not a string representation of an integer
				value = defaultValue;
			}
		} else {
			// if the result was null or empty
			value = defaultValue;
		}
		// set the calculated value and return the result state
		return setConfigValue("elm", "event-timestamp", value);
	}
	
	/**
	 * Resets all the values of the elm.txt configuration file.
	 *  
	 * @return <code>true</code> if the method correctly reset the values, <code>false</code> otherwise.
	 */
	public boolean resetElmConfig() {
		return resetConfigFile("elm");
	}
	
	/**
	 * Returns a list of three lists, the first one containing the templates id from
	 * the templates.txt configuration file, the second one containing the relative
	 * begin part of the blocks, the third one containing the relative end part of the blocs.
	 * 
	 * @return a <code>List</code> of three <code>List</code> of <code>String</code> defined as described above.
	 */
	public List<List<String>> getEditorTemplates() {
		// get the three lists
		ArrayList<String> id = new ArrayList<String>();
		ArrayList<String> begin = new ArrayList<String>();
		ArrayList<String> end = new ArrayList<String>();
		// get the configuration file path
		String filePath = "/config/editor_templates.txt";
		// try to read from the file
    	// try-with-resources Statement (close the resource automatically at the end in any case)
		try (InputStream fileStream = this.getClass().getResourceAsStream(filePath);
				BufferedReader fileReader = new BufferedReader(new InputStreamReader(fileStream))) {
			// get the line separator
			String newLine = System.getProperty("line.separator");
			String line = null;
			// for every line of the config file
			while ((line = fileReader.readLine()) != null) {
				// remove leading and trailing spaces
				String trimLine = line.replaceAll("^\\s+", "");
				// if the line is a comment line or an empty line
				if (trimLine.startsWith("#") || trimLine.isEmpty()) {
					// go to the next line
			    	continue;
			    } else if (trimLine.startsWith("id")) { // if the line is an id
			    	// get the id value
			    	String[] values = trimLine.split("=", 2);
			    	// if the value exists
			    	if (values.length > 1) {
			    		// replace new line character \n with the correct character in the system
			    		String value = values[1].replace("\\n", newLine);
			    		// replace tab character \t with the correct one
			    		value = value.replace("\\t", "\t");
			    		// add the value to the id array
			    		id.add(value);
			    	} else {
			    		// error because an id is missing
			    		return null;
			    	}
			    } else if (trimLine.startsWith("begin")) { // if the line is a begin block
			    	// get the begin value
			    	String[] values = trimLine.split("=", 2);
			    	// if the value exists
			    	if (values.length > 1) {
			    		// replace new line character \n with the correct character in the system
			    		String value = values[1].replace("\\n", newLine);
			    		// replace tab character \t with the correct one
			    		value = value.replace("\\t", "\t");
			    		// add the value to the begin array
			    		begin.add(value);
			    	} else {
			    		// add an empty value to the begin array
			    		begin.add("");
			    	}
			    } else if (trimLine.startsWith("end")) { // if the line is an end block
			    	// get the end value
			    	String[] values = trimLine.split("=", 2);
			    	// if the value exists
			    	if (values.length > 1) {
			    		// replace new line character \n with the correct character in the system
			    		String value = values[1].replace("\\n", newLine);
			    		// replace tab character \t with the correct one
			    		value = value.replace("\\t", "\t");
			    		// add the value to the end array
			    		end.add(value);
			    	} else {
			    		// add an empty value to the begin array
			    		end.add("");
			    	}
			    } else {
			    	// error if the blocks are not well defined
			    	return null;
			    }
	        }
			// if all the arrays have same size, we read correctly the blocks
			if ((id.size() == begin.size()) && (id.size() == end.size())) {
				// create the result
				List<List<String>> result = new ArrayList<List<String>>(3);
				result.add(id);
				result.add(begin);
				result.add(end);
				// return it
				return result;
			} else {
				// else return an error
				return null;
			}		
		} catch(IOException e) {
			return null;
		}
	}
	
	/**
	 * Returns a <code>String</code> array of all the completions defined in the completions.txt 
	 * configuration file.
	 * 
	 * @return a <code>String</code> array of all the possible completions.
	 */
	public String[] getEditorCompletions() {
		// completion result
		ArrayList<String> completions = new ArrayList<String>();
		// get the config file path
		String filePath = "/config/editor_completions.txt";
		// try to read from the file
    	// try-with-resources Statement (close the resource automatically at the end in any case)
		try (InputStream fileStream = this.getClass().getResourceAsStream(filePath);
				BufferedReader fileReader = new BufferedReader(new InputStreamReader(fileStream))) {   					
			String line = null;
			// for every line of the config file
			while ((line = fileReader.readLine()) != null) {
				// trim the line
				String trimLine = line.trim();	
				// if the line is a comment or is empty, skip
				if (trimLine.startsWith("#") || trimLine.isEmpty()) {
			    	continue;
			    }
				// else add the read value to the completion list
				completions.add(trimLine);
			}
			// return the array representing the completion list
			return completions.toArray(new String[completions.size()]);
		} catch(IOException e) {
			return null;
		}	
	}
	
	
	/*
	 * INNER CLASSES
	 */
	
	/**
	 * This class, extension of the <code>LinkedList</code> class, implements a specific
	 * linked list that has a limited size, avoids duplicates and has a LIFO behaviour.
	 * 
	 * <p>The maximum size of the list is specified during the object creation, using the
	 * constructor parameter. Then, the new method {@link MostRecentElementList#addNewElement(Object)}
	 * added to the standard ones can be used to ensure the limit size of the list, avoid duplicates
	 * and implements a LIFO behaviour.

	 * <p>This method does tree basic steps: first of all, if the <code>E</code> element that
	 * we try to add is already inside the list, it deletes it (this avoids duplicates). Then it
	 * adds the element on top of the list (this implements the LIFO behaviour, because iterating
	 * the list will give as first element the last item inserted, and so on). At the end, if this
	 * add has made the list exceed the maximum defined size, the method deletes the oldest element of 
	 * the list (this ensure that the limit size capacity is respected and that only the n most recent
	 * elements are kept in the list).
	 *  
	 * @author Francesco Santeramo (eventlogmanager@francescosanteramo.com)
	 *
	 * @param <E> the element to record in this list
	 */
	class MostRecentElementList<E> extends LinkedList<E> {				
			
    	/*
    	 * VARIABLES
    	 */
		
		/**
		 * Auto-Generated Serial Version UID for this class.
		 */
		private static final long serialVersionUID = 8668152711058622566L;
		
		/**
		 * The maximum size of the list.
		 */
		int maxSize;
			
		
		/*
		 * CONSTRUCTOR
		 */
		
		/**
		 * Creates a new <code>MostRecentElementList</code> with the specified
		 * maximum capacity.
		 * 
		 * @param maxSize the maximum size of the list. 0 for unlimited size.
		 */
		public MostRecentElementList(int maxSize) {
			// set the variables
			this.maxSize = maxSize;
		}		
		
		
    	/*
    	 * METHODS
    	 */
		
		/**
		 * Returns the maximum size defined for this list.
		 * 
		 * @return the maximum size defined for this list. 0 for unlimited size.
		 */
		public int getMaxSize() {
			return maxSize;
		}
		
		/**
		 * Adds a new <code>E</code> element on top of this list, deleting any duplicates and
		 * ensuring the respect of the list maximum size (defined by the {@link maxSize} variable).
		 * 
		 * <p>This method does tree basic steps: first of all, if the <code>E</code> element that
		 * we try to add is already inside the list, it deletes it (this avoids duplicates). Then it
		 * adds the element on top of the list (this implements the LIFO behaviour, because iterating
		 * the list will give as first element the last item inserted, and so on). At the end, if this
		 * add has made the list exceed the maximum defined size, the method deletes the oldest element of 
		 * the list added by this method (this ensure that the limit size capacity is respected and that 
		 * only the n most recent elements are kept in the list).
		 * 
		 * @param element the <code>E</code> element to add to this list
		 */		
		public void addNewElement(E element) {
			// check if this element is already inside the list: if yes, remove it, otherwise do nothing
			remove(element);
			// add the element on top of the list 
			addFirst(element);	
			// if the list has a defined limited size
			if (maxSize > 0) {
				// if with this add we are out of the max size of the list
				if (this.size() > maxSize) {
					// delete the oldest element added by this method
					removeLast();
				}
			}						
		}		
	}
	
}
