/*
 * Event Log Manager
 *   
 *  Copyright (C) 2015 Francesco Santeramo
 *
 *  This file is part of Event Log Manager.
 *
 *  Event Log Manager is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Event Log Manager is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *	
 *  Author:  Francesco Santeramo 
 *  Website: www.francescosanteramo.com
 *  Contact: eventlogmanager@francescosanteramo.com    
 */

package unipi.bpm.elm.santeramo.view;

import java.awt.BorderLayout;
import java.awt.Font;
import java.awt.Toolkit;
import java.awt.event.KeyEvent;
import java.beans.PropertyVetoException;
import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import javax.swing.ImageIcon;
import javax.swing.InputMap;
import javax.swing.JDialog;
import javax.swing.JInternalFrame;
import javax.swing.KeyStroke;
import javax.swing.event.DocumentListener;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.model.XLog;
import org.fife.ui.autocomplete.AutoCompletion;
import org.fife.ui.autocomplete.BasicCompletion;
import org.fife.ui.autocomplete.DefaultCompletionProvider;
import org.fife.ui.rsyntaxtextarea.CodeTemplateManager;
import org.fife.ui.rsyntaxtextarea.RSyntaxTextArea;
import org.fife.ui.rsyntaxtextarea.SyntaxConstants;
import org.fife.ui.rsyntaxtextarea.Theme;
import org.fife.ui.rsyntaxtextarea.templates.StaticCodeTemplate;
import org.fife.ui.rtextarea.RTextScrollPane;
import org.fife.ui.rtextarea.RUndoManager;

/**
 * This class extends <code>JInternalFrame</code> to create a more detailed 
 * version of a generic internal frame, made to be an event log editor.
 * 
 * <p>To do this, this class contains a <code>RSyntaxTextArea</code> where to
 * manage the event log content (write, load, modify, cut, copy, paste, undo,
 * redo, etc). The text area also implements all the editor functions, like XML
 * syntax highlighting, code folding, line numbers, auto indentation, automatic
 * markup closing, code anti-aliasing and so on.
 * 
 * <p>This class also adds specific information to the editor like if the editor
 * content has been saved or not, if it has been validated and the absolute
 * path of the associated file in the file system.
 * 
 * @author Francesco Santeramo (eventlogmanager@francescosanteramo.com)
 * 
 * @see JInternalFrame
 * @see RSyntaxTextArea
 */
public class EventLogEditor extends JInternalFrame{

	/*
	 * VARIABLES
	 */
	
	/**
	 * Auto-Generated Serial Version UID for this class.
	 */
	private static final long serialVersionUID = -6944185540769665288L;
	
	/**
	 * Defines the text area where the user can write and modify the event log.
	 * 
	 * @see RSyntaxTextArea
	 */
	private RSyntaxTextArea textArea;
	
	/**
	 * Defines the {@link #textArea} manager for the undo and redo actions,
	 * that will remember all the edits and give the possibility to navigate
	 * through them.
	 * 
	 * @see RUndoManager
	 */
	private RUndoManager undoManager;
	
	/**
	 * Defines if the {@link #textArea} content has been modified or not.
	 * 
	 * <p>This variable is set to <code>true</code> when the user modifies the editor
	 * content. This variable is set to <code>false</code> otherwise to show that
	 * no modification has been made to the content after the last save or after the 
	 * opening of a document (a new empty one or an existing one from a file). 
	 */
	private boolean isContentModified;
	
	/**
	 * Defines if the actual {@link #textArea} content has been validated against
	 * the XSD schema of XES or not.
	 * 
	 * <p>This variable is set to <code>true</code> when the the editor content has
	 * been validated. This variable is set to <code>false</code> otherwise. 
	 */
	private boolean isContentValidated;
	
	/**
	 * This string contains the absolute path of the file associated with
	 * this editor. It has a <code>null</code> value instead if the content
	 * of the editor has never been saved yet.
	 */
	private String absolutePath;
	
	/**
	 * Convenience variable for defining in only one place (in the constructor) the 
	 * default value of the font size of the {@link #textArea} and to make all the
	 * processes that need this value, to retrieve it using the 
	 * {@link #getDefaultFontSize} method. 
	 * 
	 * <p>This is made because the value of the font size can change at runtime,
	 * so we keep the original one here and make an access to it for all the 
	 * processes that need the default value.
	 */
	private int defaultFontSize;
	
	/**
	 * This <code>Collection</code> of <code>XLog</code> records all the generated
	 * objects that are a Java representation of the XES textual logs that are inside
	 * the editor {@link #textArea}.
	 * 
	 * <p>These <code>XLog</code> are generated by the footprint algorithm and stored
	 * here to do not have to calculate them again when needed.
	 * 
	 * <p>In case of any modification of the editor content, this collection will be reset because
	 * the object inside of it are not valid anymore.   
	 * 
	 * @see Collection
	 * @see XLog
	 */
	private Collection<XLog> xesLogs;
	
	/**
	 * This variable contains <code>String</code> objects that represent the 
	 * footprint matrices associated to the logs that are inside this editor.
	 * 
	 * <p>These matrices will be created and stored in this variable by the footprint matrix
	 * algorithm, to retrieve them when needed instead of recreating a new one every time.
	 * 
	 * <p>In case of any modification of the editor content, this list will be reset because
	 * the matrices inside of it are not valid anymore.   
	 */
	private ArrayList<String> footprintMatrices;
	
	/**
	 * Defines the filter to use when generating the footprint matrix for this editor content.
	 * 
	 * <p>This filter will be applied to every log in this editor by the footprint algorithm to
	 * generate the respective footprint matrix.
	 * 
	 * <p>This variable remembers the user setting for this editor for the relative footprint generation,
	 * it can be modified and updated by the application, accordingly to the user decision. 
	 * 
	 * @see FootprintFilter
	 */
	private FootprintFilter footprintFilter;
	
	/**
	 * This variable takes trace of the associate filters panel opened for this editor.
	 * 
	 * <p>Because the panel is a non-modal dialog, it stays opened while the user does other
	 * operation inside the application. To close it automatically in certain situations,
	 * this reference is needed so that the editor can close automatically its associated
	 * filters panel.
	 * 
	 * @see JDialog
	 */
	private JDialog filterDialog;
	
	
	
	/*
	 * CONSTRUCTOR
	 */
	
	/**
	 * Creates a <code>EventLogEditor</code> with the specified title, resizability, 
	 * closability, maximizability, iconifiability and ID.
	 * 
	 * <p>Then defines all its settings by calling the {@link #frameSettings(int)}
	 * method.
	 * 
	 * @param title the String to display in the title bar
	 * @param resizable if true, the internal frame can be resized
	 * @param closable if true, the internal frame can be closed
	 * @param maximizable if true, the internal frame can be maximized
	 * @param iconifiable if true, the internal frame can be iconified
	 * @param internalFrameID the identifier of the <code>EventLogEditor</code> frame
	 */
	public EventLogEditor(String title, boolean resizable, boolean closable, boolean maximizable, boolean iconifiable, int internalFrameID) {
		super(title, resizable, closable, maximizable, iconifiable);
		// initialise the variables
		textArea = new RSyntaxTextArea();
		undoManager = new RUndoManager(textArea);
		isContentModified = false;
		isContentValidated = false;
		xesLogs = null;
		footprintMatrices = new ArrayList<String>();
		footprintFilter = new FootprintFilter();
		filterDialog = null;
		// set the absolute path of the file that contains the editor text as null
		absolutePath = null;
		//set the default font size value
		defaultFontSize = 14;		
		// define the frame settings
		frameSettings(internalFrameID);
	}

	
	
	/*
	 * METHODS
	 */
	
	/**
	 * Sets the graphic details and components of this internal frame.
	 * 
	 * @param internalFrameID the identifier of the <code>EventLogEditor</code> frame
	 */
	private void frameSettings(int internalFrameID) {
		// set the frame name with the internal frame ID and default position and dimension
    	this.setName(String.valueOf(internalFrameID));
        this.setBounds(0, 0, 600, 400);        
        // get the icon for the frame and set it if available
		String iconPath = "/images/icon16x16.png";
        URL iconURL = this.getClass().getResource(iconPath);               
        if (iconURL != null) {
        	this.setFrameIcon(new ImageIcon(Toolkit.getDefaultToolkit().getImage(iconURL)));
        }         
        // set the text area
        textAreaSettings();        
        
        // set a scroll pane with the text area and enable the fold indicator
        RTextScrollPane eventLogEditorScrollPane = new RTextScrollPane(textArea);
        eventLogEditorScrollPane.setFoldIndicatorEnabled(true);
        
		// add the scrolling editor area on the centre of the frame
		this.add(eventLogEditorScrollPane, BorderLayout.CENTER);
		// revalidate and repaint the internal frame
		this.revalidate();
		this.repaint();
	}
	
	/**
	 * Sets the graphic details and components of the {@link textArea}.
	 */
	private void textAreaSettings() {
		// set the text area properties
        textArea.getDocument().putProperty("owner", this);  
        // use XML syntax highlighting
        textArea.setSyntaxEditingStyle(SyntaxConstants.SYNTAX_STYLE_XML);
        // set code folding
        textArea.setCodeFoldingEnabled(true);
        // set antialiasing
        textArea.setAntiAliasingEnabled(true);
        // set auto indentation
        textArea.setAutoIndentEnabled(true);
        // set auto close mark-up tags
        textArea.setCloseMarkupTags(true); 
        // set no line wrap
        textArea.setLineWrap(false);
        // set tab pixel dimension
        textArea.setTabSize(4);
                
        // get the selected theme file and apply it to the text area
		try {
			String themePath = "/themes/eclipse.xml";
			Theme theme = Theme.load(getClass().getResourceAsStream(themePath));
			theme.apply(textArea);
		} catch (IOException e) {
			// no need to do anything, the editor will automatically have the default theme
		} 
		// change the theme font and font size
		textArea.setFont(new Font("Courier", Font.PLAIN, defaultFontSize));
		
        // remove the default pop-up menu
        textArea.setPopupMenu(null);     
        
        // remove default keystrokes from the text area input map (we want to manage these shortcuts in the controller)
        int modifier = textArea.getToolkit().getMenuShortcutKeyMask();
        InputMap textAreaInputMap = textArea.getInputMap();
        textAreaInputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_Z, modifier), "nothing");
        textAreaInputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_Y, modifier), "nothing");
        //textAreaInputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_X, modifier), "nothing");
        //textAreaInputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_C, modifier), "nothing");
        //textAreaInputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_V, modifier), "nothing"); 
        
        // associate the undoManager to the text area
        textArea.getDocument().addUndoableEditListener(undoManager);
	}

	/**
	 * This method returns the <code>RSyntaxTextArea</code> {@link #textArea}.
	 * 
	 * @return the <code>RSyntaxTextArea</code> text area.
	 * 
	 * @see RSyntaxTextArea
	 */
	public RSyntaxTextArea getTextArea() {
		return textArea;
	}
	
	/**
	 * This method writes the content of the <code>RSyntaxTextArea</code> {@link #textArea} into the
	 * passed <code>Writer</code>.
	 * 
	 * <p>This get method uses the <code>write(Writer)</code> method of <code>JTextComponent</code> to directly
	 * write the content into an output, avoiding the known errors that come with the use of the
	 * <code>getText()</code> method of <code>JTextComponent</code> objects.
	 * 
	 * <p>Instead of using the <code>getText()</code> method and then write the returned string into an
	 * output stream, the developer should use this method to directly write the {@link #textArea} into the selected
	 * output.
	 * 
	 * @param out the <code>Writer</code> where to write the {@link #textArea} content
	 * 
	 * @throws IOException if an exception occurred while calling the <code>write(Writer)</code> method of <code>JTextComponent</code>
	 * 
	 * @see RSyntaxTextArea
	 */
	public void getTextAreaContent(Writer out) throws IOException {
		textArea.write(out);
	}
	
	/**
	 * This method sets the content of the <code>RSyntaxTextArea</code> {@link #textArea} using the
	 * passed <code>Reader</code> as source for the content.
	 * 
	 * <p>This set method uses the <code>read(Writer, Object)</code> method of <code>JTextComponent</code>
	 * to directly insert content into the <code>RSyntaxTextArea</code>, avoiding the known errors that come
	 * with the use of the <code>setText(String)</code> method of <code>JTextComponent</code> objects.
	 * 
	 * <p>Instead of reading the content from an input stream into a string and then use the <code>setText(String)</code>
	 * method to set the {@link #textArea} content, the developer should use this method to directly set the 
	 * {@link #textArea} content using the passed input source.
	 * 
	 * @param in the <code>Reader</code> from where to read the {@link #textArea} content to set
	 * 
	 * @throws IOException if an exception occurred while calling the <code>read(Writer, Object)</code> method of <code>JTextComponent</code>
	 * 
	 * @see RSyntaxTextArea
	 */
	public void setTextAreaContent(Reader in) throws IOException {
		// set the text area content
		textArea.read(in, null);
		// set the caret at the beginning of the document, to show first the top of it
		textArea.setCaretPosition(0);
	}	
	
	/**
	 * This method returns the <code>boolean</code> value of the {@link #isContentModified}
	 * variable, that specifies if the {@link #textArea} content has been modified or not.
	 * 
	 * @return the <code>boolean</code> value of the {@link #isContentModified} variable.
	 */
	public boolean isContentModified() {
		return isContentModified;
	}
	
	/**
	 * This method sets the value of the {@link #isContentModified} variable, that specifies 
	 * if the {@link #textArea} content has been modified or not.
	 * 
	 * @param contentModified <code>true</code> if the content has been modified, <code>false</code> if the content has been saved
	 */
	public void setContentModified(boolean contentModified) {
		isContentModified = contentModified;
	}	
	
	/**
	 * This method returns the <code>boolean</code> value of the {@link #isContentValidated} variable, 
	 * that specifies if the {@link #textArea} content has been validated or not against the XSD schema of XES.
	 * 
	 * @return the <code>boolean</code> value of the {@link #isContentValidated} variable.
	 */
	public boolean isContentValidated() {
		return isContentValidated;
	}
	
	/**
	 * This method sets the value of the {@link #isContentValidated} variable, that specifies 
	 * if the {@link #textArea} content has been validated or not against the XSD schema of XES.
	 * 
	 * @param contentValidated <code>true</code> if the content has been validated, <code>false</code> otherwise
	 */
	public void setContentValidated(boolean contentValidated) {
		isContentValidated = contentValidated;
	}
	
	/**
	 * This method returns the {@link #absolutePath} of the file associated with this editor.
	 * 
	 * @return the <code>String</code> representing the absolute path of the associated file, <code>null</code> if there is not an associated file yet.
	 */
	public String getAbsolutePath() {
		return absolutePath;
	}	
	
	/**
	 * This method sets the value of the {@link #absolutePath} variable, that specifies 
	 * the absolute path of the file associated with this editor.
	 * 
	 * @param path the <code>String</code> representing the absolute path of the file associated with this editor
	 */
	public void setAbsolutePath(String path) {
		absolutePath = path;
	}	
	
	/**
	 * This method returns the {@link #defaultFontSize} value of the {@link #textArea}.
	 * 
	 * @return the default font size (expressed in pt) of the <code>textArea</code>.
	 */
	public int getDefaultFontSize() {
		return defaultFontSize;
	}
	
	/**
	 * This method sets the font size of the {@link #textArea} to its default value, 
	 * defined in the {@link #defaultFontSize} variable.
	 */
	public void setDefaultFontSize() {
		Font actualFont = textArea.getFont();
		textArea.setFont(new Font(actualFont.getName(), actualFont.getStyle(), defaultFontSize));
	}
	
	/**
	 * This method returns the font size of the {@link #textArea}.
	 * 
	 * @return the font size (expressed in pt) of the {@link #textArea}.
	 */
	public int getFontSize() {		
		Font actualFont = textArea.getFont();		
		return actualFont.getSize();
	}	
	
	/**
	 * This method sets the font size of the {@link #textArea}.
	 * 
	 * @param size the font size (expressed in pt) to set to the {@link #textArea}
	 */
	public void setFontSize(int size) {
		Font actualFont = textArea.getFont();
		textArea.setFont(new Font(actualFont.getName(), actualFont.getStyle(), size));
	}	
	
	/**
	 * Returns the <code>Collection</code> of <code>XLog</code> associated with this editor
	 * content and represented by the {@link #xesLogs} variable.
	 * 
	 * @return the <code>Collection</code> of <code>XLog</code> generated for this editor content.
	 */
	public Collection<XLog> getXesLogs() {
		return xesLogs;
	}
	
	/**
	 * Sets the <code>Collection</code> of <code>XLog</code> associated with this editor
	 * content and represented by the {@link #xesLogs} variable.
	 * 
	 * @param logs the <code>Collection</code> of <code>XLog</code> to set as generated log for this editor content
	 */
	public void setXesLogs(Collection<XLog> logs) {
		xesLogs = logs;
	}
	
	/**
	 * This method reset the <code>Collection</code> of <code>XLog</code> associated with this editor
	 * content and represented by the {@link #xesLogs} variable.
	 */
	public void resetXesLogs() {
		xesLogs = null;
	}
	
	/**
	 * This method returns <code>true</code> if there is a <code>Collection</code> of <code>XLog</code>
	 * generated for this editor content, <code>false</code> otherwise.
	 * 
	 * @return <code>true</code> if there is a <code>Collection</code> of <code>XLog</code> generated for this editor content, <code>false</code> otherwise.
	 */
	public boolean isXesLogsGenerated() {
		return (xesLogs != null);
	}
	
	/**
	 * Adds a new footprint matrix to the {@link #footprintMatrices} list.
	 * 
	 * @param footprintMatrix the <code>String</code> representing the footprint matrix to add
	 */
	public void addFootprintMatrix(String footprintMatrix) {
		footprintMatrices.add(footprintMatrix);
	}
	
	/**
	 * This method returns an <code>Iterator</code> over the {@link #footprintMatrices} list to
	 * retrieve all the list elements.
	 * 
	 * @return the <code>Iterator</code> of the {@link #footprintMatrices} list.
	 */
	public Iterator<String> getFootprintMatrices() {		
		return footprintMatrices.iterator();
	}
	
	/**
	 * Clears the {@link #footprintMatrices} list, removing all the footprint matrices in it.
	 */
	public void resetFootprintMatrices() {
		footprintMatrices.clear();
	}
	
	/**
	 * This method returns <code>true</code> if there is at least one footprint matrix inside
	 * the {@link #footprintMatrices} list, <code>false</code> otherwise.
	 * 
	 * <p>If <code>true</code>, it means that the footprint matrix algorithm has been processed at 
	 * least one time over the content of this editor and it has generated one or more footprint 
	 * matrices for the log (or logs, if more than one matrix are present) contained inside this editor.
	 * 
	 * @return <code>true</code> if there is at least one footprint matrix inside the {@link #footprintMatrices} list, <code>false</code> otherwise.
	 */	
	public boolean isFootprintMatrixGenerated() {
		return !(footprintMatrices.isEmpty());
	}
	
	/** 
	 * This method returns the <code>FootprintFilter</code> {@link footprintFilter} to use when generating the footprint matrix of this editor content.
	 * 
	 * @return the <code>FootprintFilter</code> to use for the editor content when generating the respective footprint matrix.
	 */
	public FootprintFilter getFootprintFilter() {
		return footprintFilter;
	}
	
	/**
	 * This method updates the <code>FootprintFilter</code> {@link footprintFilter} associate to this editor content, using the passed parameters.
	 * 
	 * <p>The method also deletes the generated {@link footprintMatrices} associated with this editor, because they are not valid anymore due to the
	 * change just made to the filter used to generate them and they need to be regenerated again with the new filter settings.
	 * 
	 * @param minTraceOccurrences the minimum number of occurrence per trace to be considered
	 * @param classifiers the <code>Collection</code> of all <code>XEventClassifier</code> registered in the editor logs
	 * @param selectedClassifier the <code>XEventClassifier</code> selected to use for the footprint generation
	 * @param startEvents the <code>HashMap</code> of all the start <code>XEvent</code> events in the editor logs, with a <code>true</code> value if they are selected in the filter, <code>false</code> otherwise
	 * @param endEvents the <code>HashMap</code> of all the end <code>XEvent</code> events in the editor logs, with a <code>true</code> value if they are selected in the filter, <code>false</code> otherwise
	 */
	public void updateFootprintFilter(int minTraceOccurrences, Collection<XEventClassifier> classifiers, XEventClassifier selectedClassifier, 
										HashMap<String, Boolean> startEvents, HashMap<String, Boolean> endEvents) {
		// call the update method of the filter, it will manage the update
		footprintFilter.update(minTraceOccurrences, classifiers, selectedClassifier, startEvents, endEvents);
		// if there is at least one footprint matrix associated with this editor
 		if (isFootprintMatrixGenerated()) {
 			// remove all the footprint matrices saved, because they are not valid anymore due to the change just made to the filter used to generate them
 			resetFootprintMatrices();
 		}
	}
	
	/**
	 * This method resets the values of the <code>FootprintFilter</code> {@link footprintFilter}, setting them
	 * as not defined (except the minimum number of occurrences per trace, that is always defined as a number value). 
	 */
	public void resetFootprintFilter() {
		// call the reset method of the filter, it will manage the reset
		footprintFilter.reset();
	}
	
	/**
	 * This method returns the opened <code>JDialog</code> {@link filterDialog} that implements the filters panel associated with this editor.
	 * 
	 * @return the <code>JDialog</code> {@link filterDialog} opened in the View, associated with this editor.
	 */
	public JDialog getFilterDialog() {
		return filterDialog;
	}
	
	/**
	 * This method sets the {@link filterDialog} value, using the passed <code>JDialog</code>.
	 * 
	 * @param dialog the <code>JDialog</code> that implements the filters panel associated with this editor opened by the View, <code>null</code> to remove the association
	 */
	public void setFilterDialog(JDialog dialog) {
		filterDialog = dialog;
	}
	
	/**
	 * This method returns <code>true</code> if there is a {@link filterDialog} associated with this editor, <code>false</code> otherwise (if <code>null</code>).
	 * 
	 * @return <code>true</code> if there is a {@link filterDialog} associated with this editor, <code>false</code> otherwise (if <code>null</code>).
	 */
	public boolean hasFilterDialog() {
		return (filterDialog != null);
	}
	
	/**
	 * This method first checks if this editor has an associated filter dialog (using {@link #hasFilterDialog()})
	 * and only in that case it closes that dialog and dissociate it from this editor (using {@link #setFilterDialog(JDialog)})
	 * with <code>null</code> value.
	 */
	public void closeFilterDialog() {
		// if it had an associated filter dialog
		if (hasFilterDialog()) {
			// close it
			getFilterDialog().dispose();
			setFilterDialog(null);
    	}
	}
	
	/**
	 * Reset all the saved data associated with this editor.
	 * 
	 * The method deletes the saved {@link xesLogs} and {@link footprintMatrices} if present,
	 * resets the {@link footprintFilter} if complete and close the {@link filterDialog} if opened.
	 */
	public void resetAll() {
 		// if there are some xes logs generated and associated with this editor
 		if (isXesLogsGenerated()) {
 			// remove all the xes logs saved
 			resetXesLogs();
 		}
 		// if there is at least one footprint matrix associated with this editor
 		if (isFootprintMatrixGenerated()) {
 			// remove all the footprint matrices saved
 			resetFootprintMatrices();
 		}
 		// if the footprint filter of the editor is complete
 		if (footprintFilter.isComplete()) {
 		// reset the footprint filter of the editor
 	 		resetFootprintFilter();
 		} 		
 		// if the editor has a filters panel opened associated to it, close that panel
    	if (hasFilterDialog()) {
    		getFilterDialog().dispose();
    		setFilterDialog(null);
    	}
	}	 
	
	/**
	 * Undo the last edit made in the {@link #textArea}, using the {@link #undoManager},
	 * if there is any edit to be undone.
	 * 
	 * @throws CannotUndoException if the {@link #undoManager} cannot undo the last edit
	 */
	public void undo() throws CannotUndoException {	
		// if there is something to be undone
		if (undoManager.canUndo()) {
		     undoManager.undo();
		}	
	}
	
	/**
	 * Redo the last undo made in the {@link #textArea}, using the {@link #undoManager},
	 * if there is any edit to be redone.
	 * 
	 * @throws CannotRedoException if the {@link #undoManager} cannot redo the last undo
	 */
	public void redo() throws CannotRedoException {
		// if there is something to be redone
		if (undoManager.canRedo()) {
		     undoManager.redo();
		}
	}
	
	/**
	 * Empties the {@link #undoManager} sending each edit a die message in the process.
	 */
	public void discardAllEdits() {
		// discard and remove the undo manager (this is the process suggested by the RTextArea library)
		undoManager.discardAllEdits();
		textArea.getDocument().removeUndoableEditListener(undoManager);
		// create a new empty undo manager
		undoManager = new RUndoManager(textArea);
		// add the new undo manager to the text area
		textArea.getDocument().addUndoableEditListener(undoManager);
		undoManager.updateActions();
	}
	
	/** 
	 * Transfers the currently selected range in the {@link #textArea} to the system 
	 * clipboard, removing the contents from the {@link #textArea}. The current
	 * selection is reset. The command does nothing for null selections. 
	 */
	public void cut() {	
		// call the cut method of the text area
		textArea.cut();
	}
	
	/**
	 * Transfers the currently selected range in the {@link #textArea} to the system 
	 * clipboard, leaving the contents in the {@link #textArea}. The current selection
	 * remains intact. The command does nothing for null selections. 
	 */
	public void copy() {
		// call the copy method of the text area
		textArea.copy();	
	}
	
	/** 
	 * Transfers the contents of the system clipboard into the {@link #textArea}. 
	 * If there is a selection in the {@link #textArea}, it is replaced with the contents
	 * of the clipboard. If there is no selection, the clipboard contents are inserted in
	 * front of the current insert position in the {@link #textArea}. If the clipboard is 
	 * empty, the command does nothing. 
	 */
	public void paste() {
		// call the paste method of the text area
		textArea.paste();	
	}
	
	/**
	 * Sets the specified <code>DocumentListener</code> to the {@link #textArea}.
	 * 
	 * @param listener the <code>DocumentListener</code> to add to the {@link #textArea}
	 */
	public void addDocumentListener(DocumentListener listener) {		
		textArea.getDocument().addDocumentListener(listener);
	}
	
	/**
	 * This method allows to add to the editor <code>RSyntaxTextArea</code> a set of
	 * predefined templates, i.e. a block of code that can be added inside the editor
	 * using a specific keyword. 
	 * 
	 * <p>It receives a list composed by three list: the first one is the list of
	 * IDs of the templates, the ones that the user can write in the editor for
	 * recalling the template using CTRL+SHIFT+SPACE; the second list is the list
	 * of begin parts of the templates, i.e. the part that will be added before the
	 * position where the cursor will be positioned; the third list is the list of
	 * end parts of the templates, i.e. the part that will be added after the position
	 * where the cursor will be positioned.
	 * 
	 * <p>The method will take these three lists and add to the editor the templates,
	 * reading the same element of each list as a template (that is composed by an id,
	 * a begin part and an end part) until the end of the lists.
	 *  
	 * @param templates a <code>List</code> of three <code>List</code> of <code>String</code>, defined as above.
	 */
	public void addTemplates(List<List<String>> templates) {
		// if the main list contain 3 sub-lists
		if ((templates != null) && (templates.size()==3)) {
	        // enable the template function
	        RSyntaxTextArea.setTemplatesEnabled(true);
	        // get the template manager
	        CodeTemplateManager codeTemplateManager = RSyntaxTextArea.getCodeTemplateManager();
			// create the three list for template components
			ArrayList<String> id = new ArrayList<String>(templates.get(0));
			ArrayList<String> begin = new ArrayList<String>(templates.get(1));
			ArrayList<String> end = new ArrayList<String>(templates.get(2));
			// for every template, identified by an id
			for (int i=0; i<id.size(); i++) {
				// add the template to the editor, using id, begin part and end part reading from the same index in the three lists
		        codeTemplateManager.addTemplate(new StaticCodeTemplate(id.get(i), begin.get(i), end.get(i)));
			}
		}      
	}
	
	/**
	 * This method allows to add to the editor <code>RSyntaxTextArea</code> a set of
	 * predefined completion for keywords and templates id.
	 * 
	 * <p>It receives an array of keywords and an array with all the templates id.
	 * 
	 * <p>The method sets this two set of <code>String</code> elements to the completion
	 * provider to make them available in the editor.
	 * 
	 * <p>While writing a word in the editor, pressing the CTRL +SPACE macro will open the
	 * menu of all the possible keywords that start with that text substring.
	 * 
	 * @param completions an array of <code>String</code> elements that represents the keywords to set as completions
	 * @param templatesId an array of <code>String</code> elements that represents the templates id to set as completions
	 */
	public void addCompletions(String[] completions, String[] templatesId) {
		// create the completion provider that will contain the list of possible XES keyword completions
	    DefaultCompletionProvider provider = new DefaultCompletionProvider();
	    // add completions reading them from the first String array
	    // a BasicCompletion is a straightforward word completion  
	    for (String completion : completions) {
	    	provider.addCompletion(new BasicCompletion(provider, completion));
	    }
	    // add the keywords just added to the completion list in a check list
	    List<String> alreadyIn = Arrays.asList(completions);
	    // add the id of all the templates reading them from the second String array
	    for (String templateId : templatesId) {
	    	// check if the keyword is already in the completion list
	    	if (!alreadyIn.contains(templateId)) {
	    		provider.addCompletion(new BasicCompletion(provider, templateId));
	    	}	    	
	    }	    
        // create the auto completion, that will manage the completion in the installed text area, using the completion list from the given provider
        AutoCompletion autoCompletion = new AutoCompletion(provider);
        // associate the auto completion with the text area
        autoCompletion.install(textArea); 
	}
	
	/**
	 * Selects or not the internal frame if it's showing.
	 * 
	 * <p>This method first calls the superclass method to implement the correct
	 * operation, then adds to this process the status changing of the {@link #filterDialog}
	 * associated with this editor, if present, modifying its visualisation on screen
	 * depending on the value of the passed variable.
	 * 
	 * @param selected a boolean, where <code>true</code> means this internal frame should become selected (currently active) and <code>false</code> means it should become deselected
	 * 
	 * @throws PropertyVetoException when the attempt to set the property is vetoed by the JInternalFrame
	 * 
	 * @see JInternalFrame
	 */
	@Override
	public void setSelected(boolean selected) throws PropertyVetoException {
		super.setSelected(selected);
		// if the editor has a filter dialog associated to it
		if (hasFilterDialog()) {
			// set visible or not depending on the passed value
			getFilterDialog().setVisible(selected);
		}		
	}
	
	/**
	 * Makes the component visible or invisible.
	 * 
	 * <p>This method first calls the superclass method to implement the correct
	 * operation, then adds to this process the status changing of the {@link #filterDialog}
	 * associated with this editor, if present, modifying its visualisation on screen
	 * depending on the value of the passed variable.
	 * 
	 * @param aFlag <code>true</code> to make the component visible; <code>false</code> to make it invisible
	 * 
	 * @see JInternalFrame
	 */
	@Override
	public void setVisible(boolean aFlag) {
		super.setVisible(aFlag);
		// if the editor has a filter dialog associated to it
		if (hasFilterDialog()) {
			// set visible or not depending on the passed value
			getFilterDialog().setVisible(aFlag);
		}		
	}
	
}
