/*
 * Event Log Manager
 *   
 *  Copyright (C) 2015 Francesco Santeramo
 *
 *  This file is part of Event Log Manager.
 *
 *  Event Log Manager is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Event Log Manager is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *	
 *  Author:  Francesco Santeramo 
 *  Website: www.francescosanteramo.com
 *  Contact: eventlogmanager@francescosanteramo.com    
 */

package unipi.bpm.elm.santeramo.view;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Image;
import java.awt.KeyEventDispatcher;
import java.awt.KeyboardFocusManager;
import java.awt.Toolkit;
import java.awt.event.ActionListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.beans.PropertyVetoException;
import java.util.LinkedList;
import java.util.List;
import java.io.File;
import java.net.URL;

import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JDesktopPane;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JInternalFrame;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JSeparator;
import javax.swing.JTabbedPane;
import javax.swing.SwingConstants;
import javax.swing.UIManager;
import javax.swing.WindowConstants;
import javax.swing.filechooser.FileFilter;
import javax.swing.filechooser.FileNameExtensionFilter;

/**
 * This class defines the GUI Application (This is the View of the MVC pattern).
 * 
 * @author Francesco Santeramo (eventlogmanager@francescosanteramo.com)
 * 
 * @see JFrame
 */
public class EventLogManagerView extends JFrame{
	
	/*
	 * VARIABLES - GUI COMPONENTS
	 */
	
	/**
	 * Auto-Generated Serial Version UID for this class.
	 */	
	private static final long serialVersionUID = 1815082237123334767L;	
	
	/**
	 * The <code>JTabbedPane</code> that contains all the
	 * tabs of the menu and their components.
	 * 
	 * @see JTabbedPane
	 */
	private JTabbedPane menuArea;
	
	/**
	 * The <code>JPanel</code> that defines the first tab of
	 * the menu and contains all the buttons of the File tab.
	 * 
	 * @see JPanel
	 */
	private JPanel fileTab;
	
	/**
	 * The <code>JPanel</code> that defines the second tab of
	 * the menu and contains all the buttons of the Discovery tab.
	 *  
	 * @see JPanel 
	 */
	private JPanel discoveryTab;
	
	/**
	 * The <code>JPanel</code> that defines the third tab of the
	 * menu and contains all the buttons of the Info tab.
	 *  
	 * @see JPanel 
	 */
	private JPanel infoTab;
	
	/**
	 * The <code>JDesktopPane</code> that defines the main
	 * central area of the application. Inside this area the 
	 * application opens all the <code>JInternalFrame</code> frames.
	 *  
	 * @see JDesktopPane
	 * @see JInternalFrame
	 */
	private JDesktopPane contentArea;
	
	/**
	 * The <code>JPanel</code> that defines the bar at the 
	 * bottom of the <code>JFrame</code> frame. This bar contains all the 
	 * {@link FooterAreaButton FooterAreaButton} 
	 * buttons associated to the opened <code>JInternalFrame</code> frames.
	 * 
	 * @see JPanel
	 * @see JFrame
	 * @see JInternalFrame
	 */
	private JPanel footerArea;	
	
	/**
	 * This <code>ButtonGroup</code> contains all the 
	 * {@link FooterAreaButton FooterAreaButton}
	 * buttons that are inside the {@link #footerArea}.
	 * 
	 * @see ButtonGroup
	 */
	private ButtonGroup footerAreaButtonsGroup;

	/**
	 * This is the object that manage the user's files selection
	 * in many situation, such as opening a file or saving a content
	 * in a destination file.
	 * 
	 * <p>This object remembers user's directory choice and shows the
	 * last directory visited at the next file selection. 
	 * 
	 * @see JFileChooser
	 */
	private JFileChooser fileChooser;


	
	/*
	 * CONSTRUCTOR
	 */
	
	/**
	 * Constructs a new <code>EventLogManagerView</code>, initialising its
	 * instance variables and calling the method {@link #initGUI()}.
	 */
	public EventLogManagerView() {		
		// GUI components
		menuArea = new JTabbedPane(JTabbedPane.TOP, JTabbedPane.SCROLL_TAB_LAYOUT);
		fileTab = new JPanel();
		discoveryTab = new JPanel();
		infoTab = new JPanel();
		contentArea = new JDesktopPane();
		footerArea = new JPanel();
		footerAreaButtonsGroup = new ButtonGroup();
		fileChooser = new JFileChooser();
		// create the GUI
		initGUI();
	}
	

	
	/*
	 * METHODS
	 */
	
	/**
	 * This method returns the <code>JTabbedPane</code> {@link #menuArea} GUI component.
	 * 
	 * @return the <code>menuArea</code> GUI component.
	 */
	public JTabbedPane getMenuArea() {
		return menuArea;
	}
	
	/**
	 * This method returns the <code>JPanel</code> {@link #fileTab} GUI component.
	 * 
	 * @return the <code>fileTab</code> GUI component.
	 */
	public JPanel getFileTab() {
		return fileTab;
	}
	
	/**
	 * This method returns the <code>JPanel</code> {@link #discoveryTab} GUI component.
	 * 
	 * @return the <code>discoveryTab</code> GUI component.
	 */
	public JPanel getDiscoveryTab() {
		return discoveryTab;
	}
	
	/**
	 * This method returns the <code>JPanel</code> {@link #infoTab} GUI component.
	 * 
	 * @return the <code>infoTab</code> GUI component.
	 */
	public JPanel getInfoTab() {
		return infoTab;
	}
	
	/**
	 * This method returns the <code>JDesktopPane</code> {@link #contentArea} GUI component.
	 * 
	 * @return the <code>contentArea</code> GUI component.
	 */
	public JDesktopPane getContentArea() {
		return contentArea;
	}
	
	/**
	 * This method returns the <code>JPanel</code> {@link #footerArea} GUI component.
	 * 
	 * @return the <code>footerArea</code> GUI component.
	 */
	public JPanel getFooterArea() {
		return footerArea;
	}
	
	/**
	 * This method returns the <code>ButtonGroup</code> {@link #footerAreaButtonsGroup}
	 * GUI utility component.
	 * 
	 * @return the <code>footerAreaButtonsGroup</code> GUI utility component.
	 */
	public ButtonGroup getFooterAreaButtonsGroup() {
		return footerAreaButtonsGroup;
	}	
	
	/**
	 * This method returns the <code>JFileChooser</code> {@link #fileChooser} component.
	 * 
	 * @return the <code>JFileChooser</code> component.
	 */
	public JFileChooser getFileChooser() {
		return fileChooser;
	}
	
    /**
     * Starts the GUI by setting its main <code>JFrame</code> frame visible.
     * 
     * @see JFrame
     */
    public void start() {    	
        setVisible(true);
    }
    
	/**
	 * Initialises the GUI by setting all its components.
	 */
    private void initGUI() {
		/* 
		 * create the buttons for the File tab, setting for each
		 * of them the text, icon, tool tip and  action command 
		 */		
		// NEW		
        MenuAreaButton newButton = newMenuAreaButton("New", "/images/new.png", "Create a new Event Log (CTRL+N)", "new");
		// OPEN	
        MenuAreaButton openButton = newMenuAreaButton("Open", "/images/open.png", "Open existing Event Log (CTRL+O)", "open");
  		// RECENT	
        MenuAreaButton recentButton = newMenuAreaButton("Recent", "/images/recent.png", "Select from recently opened Event Log", "recent");
		// CLOSE
        MenuAreaButton closeButton = newMenuAreaButton("Close", "/images/close.png", "Close the selected Event Log (CTRL+W)", "close");
		// SAVE	
        MenuAreaButton saveButton = newMenuAreaButton("Save", "/images/save.png", "Save the selected Event Log (CTRL+S)", "save");
		// SAVE AS	
        MenuAreaButton saveAsButton = newMenuAreaButton("Save As", "/images/save-as.png", "Save the selected Event Log under different file name (CTRL+ALT+S)", "save-as");
		// PRINT
        MenuAreaButton printButton = newMenuAreaButton("Print", "/images/print.png", "Print the selected Event Log (CTRL+P)", "print");
        // IMPORT
        MenuAreaButton importButton = newMenuAreaButton("Import", "/images/import.png", "Import Events Logs and Footprint Matrices in alternative formats (CTRL+I)", "import");
        // EXPORT AS
        MenuAreaButton exportButton = newMenuAreaButton("Export As", "/images/export.png", "Export the selected document in alternative formats (CTRL+E)", "export");
		// CUT
        MenuAreaButton cutButton = newMenuAreaButton("Cut", "/images/cut.png", "Cut text in the selected Event Log (CTRL+X)", "cut");
		// COPY
        MenuAreaButton copyButton = newMenuAreaButton("Copy", "/images/copy.png", "Copy text in the selected Event Log (CTRL+C)", "copy");
		// PASTE
        MenuAreaButton pasteButton = newMenuAreaButton("Paste", "/images/paste.png", "Paste text in the selected Event Log (CTRL+V)", "paste");
		// UNDO
        MenuAreaButton undoButton = newMenuAreaButton("Undo", "/images/undo.png", "Undo typing in the selected Event Log (CTRL+Z)", "undo");
		// REDO
        MenuAreaButton redoButton = newMenuAreaButton("Redo", "/images/redo.png", "Redo typing in the selected Event Log (CTRL+Y)", "redo");
        // ZOOM IN
        MenuAreaButton zoomInButton = newMenuAreaButton("Zoom In", "/images/zoom-in.png", "Zoom in the selected Event Log (CTRL+Num+)", "zoom-in");
        // ZOOM OUT
        MenuAreaButton zoomOutButton = newMenuAreaButton("Zoom Out", "/images/zoom-out.png", "Zoom out the selected Event Log (CTRL+Num-)", "zoom-out");
		// ZOOM RESET
        MenuAreaButton resetZoomButton = newMenuAreaButton("Reset Zoom", "/images/reset-zoom.png", "Reset the zoom to the default value (CTRL+0)", "reset-zoom");

		/* 
		 * create the buttons for the Discovery tab, setting for 
		 * each of them the text, icon, tool tip and  action command  
		 */
        //VALIDATE 
        MenuAreaButton validateButton = newMenuAreaButton("Validate", "/images/validate.png", "Validate the selected Event Log against the XES XSD schema (CTRL+Q)", "validate");
        //FOOTPRINT
        MenuAreaButton footprintButton = newMenuAreaButton("Footprint", "/images/footprint.png", "Create a Footprint Matrix of the selected Event Log (CTRL+R)", "footprint");
		//FILTERS
        MenuAreaButton filtersButton = newMenuAreaButton("Filters", "/images/filters.png", "Set filters for the Footprint Matrix generation (CTRL+T)", "filters");       
        //GENERATE
        MenuAreaButton generateButton = newMenuAreaButton("Generate", "/images/generate.png", "Generate a XES log from the selected CSV document (CTRL+G)", "generate");
	
		/* 
		 * create the buttons for the Help tab, setting for
		 * each of them the text, icon, tool tip and  action command 
		 */
        //SETTINGS
        MenuAreaButton settingsButton = newMenuAreaButton("Settings", "/images/settings.png", "Application Settings (CTRL+H)", "settings");
		//ABOUT ELM	
        MenuAreaButton elmAboutButton = newMenuAreaButton("About ELM", "/images/about-elm.png", "Show information about ELM and its creator (F1)", "about");
        
		/*
		 * add the buttons to the respective tabs 
		 */		
        // FILE TAB
		fileTab.add(newButton);
		fileTab.add(openButton);
		fileTab.add(recentButton);
		fileTab.add(closeButton);		
		JSeparator menuSeparator1 = new JSeparator(SwingConstants.VERTICAL);
		menuSeparator1.setPreferredSize(new Dimension(1,80));		
		fileTab.add(menuSeparator1);
		fileTab.add(saveButton);
		fileTab.add(saveAsButton);
		JSeparator menuSeparator2 = new JSeparator(SwingConstants.VERTICAL);
		menuSeparator2.setPreferredSize(new Dimension(1,80));		
		fileTab.add(menuSeparator2);
		fileTab.add(printButton);
		fileTab.add(importButton);		
		fileTab.add(exportButton);
		JSeparator menuSeparator3 = new JSeparator(SwingConstants.VERTICAL);
		menuSeparator3.setPreferredSize(new Dimension(1,80));		
		fileTab.add(menuSeparator3);
		fileTab.add(cutButton);
		fileTab.add(copyButton);
		fileTab.add(pasteButton);
		JSeparator menuSeparator4 = new JSeparator(SwingConstants.VERTICAL);
		menuSeparator4.setPreferredSize(new Dimension(1,80));		
		fileTab.add(menuSeparator4);
		fileTab.add(undoButton);
		fileTab.add(redoButton);
		JSeparator menuSeparator5 = new JSeparator(SwingConstants.VERTICAL);
		menuSeparator5.setPreferredSize(new Dimension(1,80));		
		fileTab.add(menuSeparator5);
		fileTab.add(zoomInButton);
		fileTab.add(zoomOutButton);	
		fileTab.add(resetZoomButton);		
		// DISCOVERY TAB
		discoveryTab.add(validateButton);
		discoveryTab.add(footprintButton);
		discoveryTab.add(filtersButton);	
		JSeparator menuSeparator7 = new JSeparator(SwingConstants.VERTICAL);
		menuSeparator7.setPreferredSize(new Dimension(1,80));		
		discoveryTab.add(menuSeparator7);
		discoveryTab.add(generateButton);
		// INFO TAB
		infoTab.add(settingsButton);
		JSeparator menuSeparator8 = new JSeparator(SwingConstants.VERTICAL);
		menuSeparator8.setPreferredSize(new Dimension(1,80));
		infoTab.add(menuSeparator8);
		infoTab.add(elmAboutButton);		
		
		/*
		 * set the tabs
		 */
		// FILE TAB
		fileTab.setName("file-tab");
		fileTab.setOpaque(false);
		fileTab.setBorder(new EmptyBorder(0, 0, 4, 0) );
		fileTab.setLayout(new MenuAreaTabLayout(MenuAreaTabLayout.LEFT, 5, 5, fileTab));
		// DISCOVERY TAB
		discoveryTab.setName("discovery-tab");
		discoveryTab.setOpaque(false);
		discoveryTab.setBorder(new EmptyBorder(0, 0, 4, 0) );
		discoveryTab.setLayout(new MenuAreaTabLayout(MenuAreaTabLayout.LEFT, 5, 5, discoveryTab));
		// INFO TAB
		infoTab.setName("info-tab");
		infoTab.setOpaque(false);
		infoTab.setBorder(new EmptyBorder(0, 0, 4, 0) );
		infoTab.setLayout(new MenuAreaTabLayout(MenuAreaTabLayout.LEFT, 5, 5, infoTab));		
		/* add a component listener to this frame to set, when it gets resized, to 
		 * re-validate and repaint these three tab panels, calling on this way their 
		 * respective layout managers. This listener is in the View because this 
		 * setting depends only from the View itself, is a graphic setting. */
		this.addComponentListener(new ComponentAdapter() {
		    public void componentResized(ComponentEvent e) {
		    	// File Tab
		    	fileTab.revalidate();
		    	fileTab.repaint();
		    	// Discovery Tab
		    	discoveryTab.revalidate();
		    	discoveryTab.repaint();
		    	// Info Tab
		    	infoTab.revalidate();
		    	infoTab.repaint();
		    }
		});
		
		/*
		 * add the tabs to the menuArea
		 */		
		menuArea.addTab("File", null, fileTab, "File");
		menuArea.addTab("Discovery", null, discoveryTab, "Discovery");
		menuArea.addTab("Info", null, infoTab, "Info");
    	
		/*
		 * set the name of the menuArea and add it at the top of the JFrame frame
		 */		
		menuArea.setName("menu-area");
    	this.add(menuArea, BorderLayout.NORTH);	
    			
		/*
		 * set the name of the contentArea, its background and Desktop Manager. 
		 * Then add it to the main frame setting its position on it
		 */
    	contentArea.setName("content-area");
		contentArea.setBackground(Color.GRAY);
		contentArea.setDesktopManager(new CustomDesktopManager());
		this.add(contentArea, BorderLayout.CENTER);
		
			
		/*
		 * set the name of the footerArea, its size, borders and Layout Manager. 
		 * Then add it to the main frame, setting its position on it
		 */
		footerArea.setName("opened-internal-frames-area");
		footerArea.setPreferredSize(new Dimension(-1, 26));
		footerArea.setBorder(new CompoundBorder(BorderFactory.createEmptyBorder(1, 0, 0, 0), 
												new CompoundBorder(BorderFactory.createMatteBorder(1, 0, 0, 0, Color.gray), 
																	BorderFactory.createEmptyBorder(1, 0, 0, 0))));
		footerArea.setLayout(new OneRowFlowLayout());
		this.add(footerArea, BorderLayout.SOUTH);
						
    	// set the title of the main frame: Event Logs Manager
        setTitle("Event Log Manager");
        
        // set the action of the close button of the main frame: close the application
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		        
        // set the icons of the main frame
        List<Image> imagesList = new LinkedList<Image>();
        String[] imagesURLList = {
        		"/images/icon16x16.png",
        		"/images/icon32x32.png",
        		"/images/icon64x64.png",
        		"/images/icon128x128.png"};
        
        for (String imageURL: imagesURLList) {        	
            URL iconURL = this.getClass().getResource(imageURL);            
            if (iconURL != null) {
            	imagesList.add(Toolkit.getDefaultToolkit().getImage(iconURL));
            }
        }       
        setIconImages(imagesList);    
		
		// set the size of the main frame and its position on the screen
        int inset = 50;
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        setBounds(inset, inset, screenSize.width  - inset*2, screenSize.height - inset*2);	
        setLocationRelativeTo(null); // null = centred on the screen        
    }
    
	/**
	 * Creates a new {@link MenuAreaButton} button with text,
	 * icon (optional), tool tip and action command.
	 * 
	 * @param buttonText the text of the button
	 * @param buttonIconURL the URL of the image to display on the button
	 * @param buttonToolTip the tool tip to display
	 * @param buttonActionCommand the action command of the button
	 * 
	 * @return a {@link MenuAreaButton} button.
	 */
	private MenuAreaButton newMenuAreaButton(String buttonText, String buttonIconURL, String buttonToolTip, String buttonActionCommand) {		
        Icon buttonIcon = null;
		URL buttonIconURLResource = this.getClass().getResource(buttonIconURL); // get the image from the URL
		if (buttonIconURLResource != null) { // if the image exists
        	buttonIcon = new ImageIcon(Toolkit.getDefaultToolkit().getImage(buttonIconURLResource)); // get the icon from the image
        }
		return new MenuAreaButton(buttonText, buttonIcon, buttonToolTip, buttonActionCommand);
	}
    
    /**
     * Adds a new <code>EventLogEditor</code> object, extension of the 
     * <code>JInternalFrame</code> class, inside the {@link #contentArea}.
     * 
     * <p>This method also sets the internal frame dimension, calculating it 
     * based on the available space in the {@link #contentArea}, and position,
     * calling the DesktopManager to position the frame. 
     * 
     * <p>One of the passed parameter is the internal frame identifier to set, this is
     * needed for creating a connection between this frame and its associated <code>FooterAreaButton</code>.
     * 
     * @param title the title of the internal frame
     * @param internalFrameID the identifier of the <code>EventLogEditor</code>
     * 
     * @return the <code>EventLogEditor</code> object created.
     * 
     * @see EventLogEditor
     * @see FooterAreaButton
     * @see JInternalFrame
     */
    public EventLogEditor newEventLogEditor(String title, int internalFrameID) {    	    	
    	/* this method, when creates a new EventLogEditor for the first time, makes the system loads the 
    	 * bounce libraries to get the XMLEditorKit. This can take few seconds, so a busy state is set */
    	setBusyState(true);
    	// create a new internal frame, setting true its values for resizable, closable, maximizable, iconifiable 
    	EventLogEditor newEventLogEditor = new EventLogEditor(title, true, true, true, true, internalFrameID);              
        // add the internal frame to the content area of the application
        contentArea.add(newEventLogEditor);
        // get the dimension of the content Area
        Dimension mainFrameSize = contentArea.getSize();
        // set the default position and the right dimension for the internal frame
        int w = (int)(mainFrameSize.width * 0.5);
        int h = (int)(mainFrameSize.height * 0.8);
        newEventLogEditor.setBounds(0, 0, w, h);
		// set the internal frame as visible on the main frame
    	newEventLogEditor.setVisible(true);
        // set the right position for the internal frame on the contentArea, using the DesktopManager
        ((CustomDesktopManager)contentArea.getDesktopManager()).positionFrame(newEventLogEditor);  
        try {
        	// set the internal frame as the selected one inside the contentArea
        	newEventLogEditor.setSelected(true);
		} catch (PropertyVetoException exception) {
        	showErrorMessage("The application is not able to select the window. Try again.");
        } 
        contentArea.revalidate();
        contentArea.repaint();
        
        // process done, busy state = off
        setBusyState(false);
        
        return newEventLogEditor;
    }
    
    /**
     * Adds a new <code>FootprintMatrixInfo</code> object, extension of the 
     * <code>JInternalFrame</code> class, inside the {@link #contentArea}.
     * 
     * <p>This method also sets the internal frame dimension, calculating it 
     * based on the available space in the {@link #contentArea}, and position,
     * calling the DesktopManager to position the frame. 
     * 
     * <p>One of the passed parameter is the internal frame identifier to set, this is
     * needed for creating a connection between this frame and its associated <code>FooterAreaButton</code>.
     * 
     * @param title the title of the internal frame
     * @param internalFrameID the identifier of the <code>FootprintMatrixInfo</code>
     * 
     * @return the <code>FootprintMatrixInfo</code> object created.
     * 
     * @see FootprintMatrixInfo
     * @see FooterAreaButton
     * @see JInternalFrame
     */    
    public FootprintMatrixInfo newFootprintMatrixInfo(String title, int internalFrameID) {    	    	
    	// create a new internal frame, setting true its values for resizable, closable, maximizable, iconifiable 
    	FootprintMatrixInfo newFootprintMatrixTable = new FootprintMatrixInfo(title, true, true, true, true, internalFrameID);              
        // add the internal frame to the content area of the application
        contentArea.add(newFootprintMatrixTable);
        // get the dimension of the content Area
        Dimension mainFrameSize = contentArea.getSize();
        // set the default position and the right dimension for the internal frame
        int w = (int)(mainFrameSize.width * 0.5);
        int h = (int)(mainFrameSize.height * 0.8);
        newFootprintMatrixTable.setBounds(0, 0, w, h);
		// set the internal frame as visible on the main frame
        newFootprintMatrixTable.setVisible(true);
        // set the right position for the internal frame on the contentArea, using the DesktopManager
        ((CustomDesktopManager)contentArea.getDesktopManager()).positionFrame(newFootprintMatrixTable);  
        try {
        	// set the internal frame as the selected one inside the contentArea
        	newFootprintMatrixTable.setSelected(true);
		} catch (PropertyVetoException exception) {
        	showErrorMessage("The application is not able to select the window. Try again.");
        } 
        contentArea.revalidate();
        contentArea.repaint();
        
        return newFootprintMatrixTable;
    }
    
    /**
     * Adds a new generic <code>JInternalFrame</code> object inside the {@link #contentArea}.
     * 
     * <p>This method also sets the internal frame dimension, calculating it 
     * based on the available space in the {@link #contentArea}, and position,
     * calling the DesktopManager to position the frame. 
     * 
     * <p>One of the passed parameter is the internal frame identifier to set, this is
     * needed for creating a connection between this frame and its associated <code>FooterAreaButton</code>. 
     * 
     * @param title the title of the internal frame
     * @param internalFrameID the identifier of the <code>JInternalFrame</code>
     * 
     * @return the <code>JInternalFrame</code> object created.
     * 
     * @see JInternalFrame
     */
    public JInternalFrame newInternalFrame(String title, int internalFrameID) {    	    	
    	// create a new internal frame, setting true its values for resizable, closable, maximizable, iconifiable 
    	JInternalFrame newInternalFrame = new JInternalFrame(title, true, true, true, true);
    	// set name and default position and dimension
    	newInternalFrame.setName(String.valueOf(internalFrameID));
    	newInternalFrame.setBounds(0, 0, 600, 400);        
        // get the icon for the frame and set it if available
		String resourcePath = "/images/icon16x16.png";
        URL icon16x16URL = this.getClass().getResource(resourcePath);               
        if (icon16x16URL != null) {
        	newInternalFrame.setFrameIcon(new ImageIcon(Toolkit.getDefaultToolkit().getImage(icon16x16URL)));
        }		    	
        // add the internal frame to the content area of the application
        contentArea.add(newInternalFrame);
        // get the dimension of the content Area
        Dimension mainFrameSize = contentArea.getSize();
        // set the default position and the right dimension for the internal frame
        int w = (int)(mainFrameSize.width * 0.5);
        int h = (int)(mainFrameSize.height * 0.8);
        newInternalFrame.setBounds(0, 0, w, h);
		// set the internal frame as visible on the main frame
        newInternalFrame.setVisible(true);
        // set the right position for the internal frame on the contentArea, using the DesktopManager
        ((CustomDesktopManager)contentArea.getDesktopManager()).positionFrame(newInternalFrame);  
        try {
        	// set the internal frame as the selected one inside the contentArea
        	newInternalFrame.setSelected(true);
		} catch (PropertyVetoException exception) {
        	showErrorMessage("The application is not able to select the window. Try again.");
        } 
        contentArea.revalidate();
        contentArea.repaint();
        
        return newInternalFrame;
    }
    
    /**
     * Adds a new <code>FooterAreaButton</code> button inside the {@link #footerArea} 
     * with the specified settings, included the identifier of the <code>JInternalFrame</code>
     * to manage.
     * 
     * @param text the text of the button
     * @param internalFrameID the identifier of the <code>JInternalFrame</code> to manage
     *
     * @return the <code>FooterAreaButton</code> button created.
     */
    public FooterAreaButton newFooterAreaButton(String text, int internalFrameID) {
    	FooterAreaButton newFooterAreaButton;   
    	// get the icon for the button
    	String resourcePath = "/images/icon13x13.png";
        URL iconURL = this.getClass().getResource(resourcePath); 
        //if the icon is available
        if (iconURL != null) {
        	// create a button with the passed parameters and the specified icon
        	newFooterAreaButton = new FooterAreaButton(text, new ImageIcon(Toolkit.getDefaultToolkit().getImage(iconURL)), true, internalFrameID);
        } else {
        	// create a button with the passed parameters and no icon
        	newFooterAreaButton = new FooterAreaButton(text, true, internalFrameID);
        }        
        // add the button to the footerArea component of the GUI
        footerArea.add(newFooterAreaButton);
        /* add the button to the openedInternalFramesButtonsGroup component of the GUI that contains all the buttons of the 
         * footerArea area. This component makes only one button of the group to be graphically selected, automatically
         * removing the selection to the others when another one of them get pressed */
        footerAreaButtonsGroup.add(newFooterAreaButton);
		// re-validate and repaint the footerArea
        footerArea.revalidate();
        footerArea.repaint();
        // set the button selected
        newFooterAreaButton.setSelected(true);
        
        return newFooterAreaButton;
    }
    
    /**
     * Opens a dialog that displays the passed message and then makes the user
     * decides between an array of options. All this options are passed as parameter,
     * together with the title and the type of the message. 
     * 
     * @param message the message to display
     * @param title the title for the dialog
     * @param messageType an integer designating the kind of message this is, primarily used to determine
     *  the icon from the pluggable Look and Feel: ERROR_MESSAGE, INFORMATION_MESSAGE, WARNING_MESSAGE,
     *  QUESTION_MESSAGE, or PLAIN_MESSAGE
     * @param options an array of objects indicating the possible choices the user can make; 
     *  if the objects are components, they are rendered properly; non-String objects are rendered using their
     *  toString methods; if this parameter is <code>null</code>, the options YES/NO/CANCEL are used
     * @param initialValue the object that represents the default selection for the dialog; only meaningful
     *  if options is used; can be <code>null</code> 
     * 
     * @return an integer indicating the option chosen by the user, or CLOSED_OPTION if the user closed the dialog
     */    
    public int showOptionDialog(Object message, String title, int messageType, Object[] options, Object initialValue) {
		int choice = JOptionPane.showOptionDialog(this, // the parent of this dialog
		message, // message of the dialog
		title, // title of the dialog
		JOptionPane.YES_NO_CANCEL_OPTION, // option type
		messageType, // message type
		null, // do not use a custom Icon
		options, // titles of the buttons
		initialValue); // default button title		
		return choice;
    }
    
    /**
     * This method shows a dialog for selecting a file.
     * 
     * @param dialogTitle the title of the dialog and the text of the approve button
     * @param selectedFile the name of the file to set as selected on the dialog, <code>null</code> for no selection
     * @param filesFilter the filters (zero, one or more, passing zero, one or more <code>FileNameExtensionFilter</code>
     * objects, or an array containing them, to this method) to use for displaying the list of files in the dialog
     * 
     * @return the absolute path of the selected file if the user have selected one, <code>null</code> otherwise.
     */
    public String showFileSelectionDialog(String dialogTitle, String selectedFile, FileFilter... filesFilter) {  
    	// set a selected file
    	if (selectedFile != null) {    	
    		fileChooser.setSelectedFile(new File(selectedFile));
    	}    	
    	// set the files filter
    	if (filesFilter != null) {
			if (filesFilter.length > 0) {  
				// set the passed filters on the list of possible options for filtering
				for (int i = 0; i < filesFilter.length; i++) {
					fileChooser.addChoosableFileFilter(filesFilter[i]);   			
				}    
				// set the first passed filter as the selected one
				fileChooser.setFileFilter(filesFilter[0]);
				// remove the "All files" choice because we just want a selection of extensions and not all
				fileChooser.setAcceptAllFileFilterUsed(false);
			}    	
    	}
    	// show the dialog and get user action
    	int userAction = fileChooser.showDialog(EventLogManagerView.this, dialogTitle);
    	String absolutePath;
    	// check if the user selects a file or not
    	if (userAction == JFileChooser.APPROVE_OPTION) {    		
    		// if the user have selected a file 
    		absolutePath = fileChooser.getSelectedFile().getAbsolutePath();
    		// if the selected filter is not "All Files"
    		FileFilter selectedFilter = fileChooser.getFileFilter();
    		if (!selectedFilter.equals(fileChooser.getAcceptAllFileFilter())) {
    			if (selectedFilter instanceof FileNameExtensionFilter) {
    				// retrieve all the extensions for the selected filter
    		        String[] extensions = ((FileNameExtensionFilter)selectedFilter).getExtensions();
    		        boolean extensionCheck = false;
    		        // check if the file name has already at its end one of the possible extension for the selected filter
    		        for (String extension : extensions) { 
    		            if (absolutePath.endsWith('.' + extension)) {
    		            	extensionCheck = true;
    		            }
    		        }
    		        // if it does not have one of them, add to the selected name the first extension for this filter
    		        if (extensionCheck == false) {
    		        	absolutePath = absolutePath + '.' + extensions[0];
    		        }
    		    }
    		}
    	} else {
    		absolutePath = null; 	
    	}		
    	// remove the choosable filters from the file chooser to reset it for next time
    	if (filesFilter != null) {
			if (filesFilter.length > 0) {  
				// set the passed filters on the list of possible options for filtering
				for (int i = 0; i < filesFilter.length; i++) {
					fileChooser.removeChoosableFileFilter(filesFilter[i]);   			
				}    
				// set the first passed filter as the selected one
				fileChooser.setAcceptAllFileFilterUsed(true);
				fileChooser.setFileFilter(fileChooser.getAcceptAllFileFilter());
			}    	
    	}
    	return absolutePath;
    }
    
    /**
     * Shows a non-modal dialog with the passed title, associated with the passed <code>EventLogEditor</code>
     * (the dialog will be closed automatically by the editor when needed and only one dialog associated to
     * a specific editor can be opened, if a new one gets opened, the old one gets closed).
     * 
     * <p>The dialog will contain a <code>FilterPanel</code> object, that is calculated based on the editor
     * <code>FootprintFilter</code>, to show to the user the possible selections.
     * 
     * To the <code>FilterPanel</code>, the passed <code>ActionListener</code> will be set.
     * 
     * @param title the <code>String</code> representing the title of the dialog box
     * @param editor the <code>EventLogEditor</code> that will be the owner of the dialog
     * @param listener the <code>ActionListener</code> to set for the <code>FilterPanel</code> buttons
     * 
     * @see EventLogEditor
     * @see FilterPanel
     * @see FootprintFilter
     */
    public void newFiltersDialog(String title, EventLogEditor editor, ActionListener listener) {
    	// if the editor has already a filter dialog associated to it, close it (this method does all the checks)    	
    	editor.closeFilterDialog();
		// create a new filters panel
    	FilterPanel filterPanel = new FilterPanel(editor);
    	filterPanel.addButtonActionListener(listener);
    	
    	// set a scroll pane with the content pane inside
		JScrollPane scrollPane = new JScrollPane(filterPanel);
		// the scroll bars will appear only as needed
        scrollPane.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        scrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED); 
    	// get the content pane dimension
    	Dimension paneDimension = filterPanel.getPreferredSize();
    	// get the content area dimension
        Dimension areaDimension = contentArea.getSize();
        // set the dimension of the scroll pane, using a max dimension available for the content pane
        int paneWidth = paneDimension.width;  
        int maxWidth = (int)(areaDimension.width * 0.35);
        int paneHeight = paneDimension.height; 
        int maxHeight = (int)(areaDimension.height * 0.85);
        // if the pane width is bigger than the maximum available
        if(paneWidth > maxWidth) {
        	// set the pane width as the maximum available
        	paneWidth = maxWidth;
    	}
        // if the pane height is bigger than the maximum available
    	if(paneHeight > maxHeight) {
    		// set the pane height as the maximum available
    		paneHeight = maxHeight;
    	}
        // get the dimension of the horizontal and vertical scroll bar, plus/minus insets
        int hScrollBarWidth = ((Integer)UIManager.get("ScrollBar.width")) + 2; 
        int vScrollBarWidth = ((Integer)UIManager.get("ScrollBar.width")) - 12; 
        // set the dimension of the scroll pane
        int scrollPaneWidth = paneWidth + hScrollBarWidth;
        int scrollPaneHeight = paneHeight + vScrollBarWidth;
        scrollPane.setPreferredSize(new Dimension(scrollPaneWidth, scrollPaneHeight));       
        
    	// create a non-modal dialog with the passed title
    	JDialog dialog = new JDialog(this, title, false);
    	// set the dialog close operation to do not close
    	dialog.setDefaultCloseOperation(JDialog.DO_NOTHING_ON_CLOSE); 
    	// use instead a window listener to ask the editor to close its dialog properly
    	dialog.addWindowListener(new WindowAdapter() {
            @Override
            public void windowClosing(WindowEvent e) {
            	editor.closeFilterDialog();
            }
        });
    	// set the dialog content (the scroll pane)
    	dialog.setContentPane(scrollPane);
    	dialog.pack();
    	// set the position where the dialog will appear
    	dialog.setLocationRelativeTo(this);
    	// show the dialog
    	dialog.setVisible(true);
    	// associate this dialog with its owner
    	editor.setFilterDialog(dialog);
    }
    
    /**
     * Shows a non-modal dialog with the passed title and that contains the passed content pane.
     * 
     * @param title the <code>String</code> representing the title of the dialog box
     * @param contentPane the <code>Container</code> representing the content pane to show inside the dialog box
     * @param modal specifies whether dialog blocks user input to other top-level windows when shown. If <code>true</code>, the modality type
     * 				 property is set to <code>DEFAULT_MODALITY_TYPE</code> otherwise the dialog is modeless
     */
    public void showDialog(String title, Container contentPane, boolean modal) {
   	
    	// create a non-modal dialog with the passed title
    	JDialog dialog = new JDialog(this, title, modal);
    	// set the close operation
    	dialog.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);   
    	// set the content pane to the dialog      
    	dialog.setContentPane(contentPane);
    	dialog.pack();
    	// set the position where the dialog will appear
    	dialog.setLocationRelativeTo(this);
    	// show the dialog
    	dialog.setVisible(true);
    }
    
    /**
     * Shows a pop-up menu connected with the passed <code>JButton</code>, using as names for the items
     * the ones that are inside the <code>List</code> and adding to them the passed <code>ActionListener</code>.
     * 
     * @param targetButton the <code>JButton</code> that will be the graphical owner of the pop-up menu
     * @param itemList the <code>List</code> containing the <code>String</code> representing the <code>JMenuItem</code> names
     * @param itemListener the <code>ActionListener</code> to associate to each <code>JMenuItem</code> of the pop-up menu
     * 
     * @see JButton
     * @see LinkedList
     * @see ActionListener
     * @see JPopupMenu
     * @see JMenuItem
     */
    public void showPopup(JButton targetButton, List<String> itemList, ActionListener itemListener) {   	   	
    	// create the pop-up menu
    	JPopupMenu popupMenu = new JPopupMenu();    	
    	int itemNumber = 0;
    	// for every names in the list, create an item
		for (String itemName : itemList) {	
			itemNumber++;
			// create a JMenuItem item with text and an icon if the icon is available, with only text otherwise
			JMenuItem menuItem;			
			String resourcePath = "/images/recent-little.png";
	        URL iconURL = this.getClass().getResource(resourcePath);                        
	        if (iconURL != null) {            	
	        	menuItem = new JMenuItem(itemNumber+": "+itemName, new ImageIcon(Toolkit.getDefaultToolkit().getImage(iconURL)));       
	        } else {     	
	        	menuItem = new JMenuItem(itemNumber+": "+itemName);
	        }			
	        // set the action command of the item
	        menuItem.setActionCommand(itemName);
	        // add the item actionlistener
	        menuItem.addActionListener(itemListener);
	        // add the item to the pop-up menu
	        popupMenu.add(menuItem);
		}
    	// show the pop-up menu
		popupMenu.show(targetButton, 0, targetButton.getHeight());
    }
	
    /**
     * This method displays the passed info message string in a dialog on the application frame.
     * 
     * <p>It defines already the title of the dialog ("Event Logs Manager Info System") and its
     * type (<code>INFORMATION_MESSAGE</code>).
     * 
     * @param infoMessage the info message to display
     */
	public void showInfoMessage(String infoMessage) {
		// show the dialog with message, title and type as specified
		JOptionPane.showMessageDialog(this, infoMessage, "Event Logs Manager Info System", JOptionPane.INFORMATION_MESSAGE);
	}
	
    /**
     * This method displays the passed warning message string in a dialog on the application frame.
     * 
     * <p>It defines already the title of the dialog ("Event Logs Manager Warning System") and its
     * type (<code>WARNING_MESSAGE</code>).
     * 
     * @param warningMessage the warning message to display
     */
	public void showWarningMessage(String warningMessage) {
		// show the dialog with message, title and type as specified
		JOptionPane.showMessageDialog(this, warningMessage, "Event Logs Manager Warning System", JOptionPane.WARNING_MESSAGE);
	}
	
    /**
     * This method displays the passed error message string in a dialog on the application frame.
     * 
     * <p>It defines already the title of the dialog ("Event Logs Manager Error System") and its
     * type (<code>ERROR_MESSAGE</code>).
     * 
     * @param errorMessage the error message to display
     */
	public void showErrorMessage(String errorMessage) {
		// show the dialog with message, title and type as specified
		JOptionPane.showMessageDialog(this, errorMessage, "Event Logs Manager Error System", JOptionPane.ERROR_MESSAGE);
	}
	
	/**
	 * Switches the Operative System cursor icon between <code>DEFAULT</code>
	 * and <code>WAIT</code>, depending on the value passed as parameter 
	 * (<code>true</code> to express the busy state of the application, 
	 * <code>false</code> to switch back to the default state).
	 * 
	 * <p>This method helps to show to the user a busy state of the application 
	 * using the cursor icon. Every time an operation could take more than few 
	 * seconds, this method should be called before (with a <code>true</code>
	 * value) and after it (with a <code>false</code> one).
	 * 
	 * @param isBusy the state of the application, <code>true</code> if busy to do some long operation, <code>false</code> otherwise
	 */
	public void setBusyState(boolean isBusy) {
		// create a glass pane
		Component glassPane = this.getGlassPane();
		// get the wait or default cursor depending on the isBusy value
		Cursor theCursor = isBusy ? Cursor.getPredefinedCursor( Cursor.WAIT_CURSOR ) : Cursor.getPredefinedCursor( Cursor.DEFAULT_CURSOR );
		// set the cursor
		this.setCursor(theCursor);
		glassPane.setCursor(theCursor);
		// show or hide the glass pane
		glassPane.setVisible(isBusy); 	
	}
	
    /**
     * Adds a <code>KeyEventDispatcher</code> to the <code>KeyboardFocusManager</code> of this frame.
     * 
     * @param dispatcher the <code>KeyEventDispatcher</code> to add to the <code>KeyboardFocusManager</code> of the frame
     * 
     * @see KeyEventDispatcher
     * @see KeyboardFocusManager
     */
	public void addKeyEventDispatcher(KeyEventDispatcher dispatcher) {
		// get the current keyboard focus manager
		KeyboardFocusManager manager = KeyboardFocusManager.getCurrentKeyboardFocusManager();
        // add the key event dispatcher to the manager
		manager.addKeyEventDispatcher(dispatcher);
	}
}
