/*
 * Event Log Manager
 *   
 *  Copyright (C) 2015 Francesco Santeramo
 *
 *  This file is part of Event Log Manager.
 *
 *  Event Log Manager is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Event Log Manager is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *	
 *  Author:  Francesco Santeramo 
 *  Website: www.francescosanteramo.com
 *  Contact: eventlogmanager@francescosanteramo.com    
 */

package unipi.bpm.elm.santeramo.view;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSeparator;
import javax.swing.JTextField;
import javax.swing.SwingConstants;

import org.deckfour.xes.classification.XEventClassifier;

/**
 * This class extends <code>JPanel</code> to define a particular Swing panel,
 * that decides its elements dynamically by looking at the values of the
 * <code>FootprintFilter</code> object associated with the <code>EventLogEditor</code>
 * passed to the constructor.
 * 
 * <p>The class reads from this <code>FootprintFilter</code> all the information needed
 * and creates a panel that shows them, making the user able to modify this information
 * and the Controller to install a listener to wait for user'actions on the panel buttons.
 * 
 * @author Francesco Santeramo (eventlogmanager@francescosanteramo.com)
 * 
 * @see FootprintFilter
 * @see EventLogEditor
 */
public class FilterPanel extends JPanel implements ItemListener {

	/*
	 * VARIABLES
	 */
	
	/**
	 * Auto-Generated Serial Version UID for this class.
	 */
	private static final long serialVersionUID = -8580074480578821489L;
	
	/**
	 * The <code>EventLogEditor</code> selected for showing this panel.
	 * 
	 * <p>The event log editor will be the owner of this panel and the
	 * panel will be created based on the information inside the editor
	 * <code>FootprintFilter</code>.
	 * 
	 * @see EventLogEditor
	 * @see FootprintFilter
	 */
	EventLogEditor eventLogEditor;	
	
	/**
	 * The <code>FootprintFilter</code> associated with the {@link eventLogEditor}
	 * that contains all the information for creating this panel.
	 * 
	 * @see FootprintFilter
	 */
	FootprintFilter footprintFilter;
	
	/**
	 * The <code>ArrayList</code> of all the <code>JButton</code> used inside this
	 * panel.
	 * 
	 * This list is useful to associate dynamically to all the buttons inside of it
	 * an <code>ActionListener</code> to manage them. A button that needs to be associated
	 * with an action listener just needs to be inside this list to be automatically connected
	 * with the specified <code>ActionListener</code> by the 
	 * {@link #addButtonActionListener(ActionListener) addButtonActionListener(ActionListener)}.
	 * 
	 * @see ArrayList
	 * @see JButton
	 * @see ActionListener 
	 */
	ArrayList<JButton> buttonList;
	
	/**
	 * The <code>JTextField</code> that contains the value of the minimum number of trace occurrences.
	 * 
	 * @see JTextField
	 */
	JTextField minTraceOccurrenceField;	
	
	/**
	 * The <code>JComboBox</code> that contains all the registered classifiers.
	 * 
	 * @see JComboBox
	 */
	JComboBox<String> classifierMenu;
	
	/**
	 * The <code>HashMap</code> that contains all the possible start events
	 * and their selection.
	 * 
	 * @see HashMap
	 */
	HashMap<String, Boolean> startEvents;
	
	/**
	 * The <code>HashMap</code> that contains all the possible end events
	 * and their selection.
	 * 
	 * @see HashMap
	 */
    HashMap<String, Boolean> endEvents;
	
    
    
	/*
	 * CONSTRUCTOR
	 */
	
    /**
     * Creates a <code>FilterPanel</code> associated with the passed <code>EventLogEditor</code>.
     * 
     * @param editor the <code>EventLogEditor</code> that will be the owner of this panel and from where to read the information to build the panel
     * 
     * @see EventLogEditor
     */
	public FilterPanel(EventLogEditor editor) {
		// set the owner of the panel
		eventLogEditor = editor;	
		// get the filters to use to build the panel
		footprintFilter = editor.getFootprintFilter();
		// set the button list
		buttonList = new ArrayList<JButton>(3);
		// call the method for creating the panel view depending on the filter values
		panelSettings();
	}
	

	
	/*
	 * METHODS
	 */
	
	/**
	 * Creates the panel view using the information inside the {@link #footprintFilter}.
	 * 
	 * <p>The panel will be incomplete if the filter is incomplete (and therefore showing 
	 * only the {@link #minTraceOccurrenceField}), complete otherwise (showing in
	 * this case not only the {@link #minTraceOccurrenceField}, but also the
	 * {@link #classifierMenu} and a list of check buttons based on the {@link #startEvents}
	 * and {@link #endEvents} maps).
	 */
	private void panelSettings() {
		// set an empty border around the content
		setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
		// set the panel layout
		setLayout(new GridBagLayout());
		// create the constraints to use for positioning the elements
		GridBagConstraints constraints = new GridBagConstraints();
		// set the general constraints
		constraints.fill = GridBagConstraints.HORIZONTAL;
		constraints.weightx = 0.5;			
		constraints.insets = new Insets(5,5,5,5);

		// put the first label in the panel, setting its position in the layout
		constraints.gridx = 0;
		constraints.gridy = 0;				
    	add(new JLabel("Min Trace Occurrences:"), constraints);
    	
    	// put the relative element in the panel, setting its position in the layout
    	constraints.gridx = 1;
    	constraints.gridy = 0;
	    minTraceOccurrenceField = new JTextField(6);
	    // set a value inside the text filed, that is the value saved in the filter
	    minTraceOccurrenceField.setText(Integer.toString(footprintFilter.getMinTraceOccurrences()));
    	add(minTraceOccurrenceField, constraints);	
		
    	// now, if the filter is complete, add other options to the panel
		if (footprintFilter.isComplete()) {
			// put the second label in the panel, setting its position in the layout
			constraints.gridx = 0;
			constraints.gridy = 1;    	
	    	add(new JLabel("Select Classifier:"), constraints);
	    	
	    	// put the relative element in the panel, setting its position in the layout
	    	constraints.gridx = 1;
	    	constraints.gridy = 1;
		    classifierMenu = new JComboBox<String>();	   
	        // set the menu with the registered classifiers from the footprint filter
	        for (XEventClassifier classifier : footprintFilter.getClassifiers()) {
	        	classifierMenu.addItem(classifier.name());
	        }
	        // set as selected in the menu the classifier selected for the analysis
	    	classifierMenu.setSelectedItem(footprintFilter.getSelectedClassifier().name());  
	        add(classifierMenu, constraints); 
	        
	        // add a separator, setting its position in the layout and a full width dimension
	        constraints.gridwidth = 2;
	        constraints.gridx = 0;
	        constraints.gridy = 2;	
			add(new JSeparator(SwingConstants.HORIZONTAL), constraints);
	        
			// restore the column width dimension
			constraints.gridwidth = 1;
			// put the third label in the panel, setting its position in the layout
			constraints.gridx = 0;
			constraints.gridy = 3;	
	        add(new JLabel("Start Events:"), constraints);
	        
	        // put the relative element in the panel, setting its position in the layout
	        constraints.gridx = 1;
	        constraints.gridy = 3;
	        // get the HashMap from the filter
	        startEvents = new HashMap<String, Boolean>(footprintFilter.getStartEvents());
	        // create the check box panel
	        JPanel checkPanel = new JPanel(new GridLayout(0, 1));
	        // iterate over the map elements
	        for (Map.Entry<String, Boolean> startEvent : startEvents.entrySet()) {
	            // create a check box with the event name
	            JCheckBox checkBox = new JCheckBox(startEvent.getKey());
	            // set the action command with the event name
	            checkBox.setActionCommand("start-"+startEvent.getKey());
	            // set or not as selected the check box, depending on the boolean value in the HashMap
	            checkBox.setSelected(startEvent.getValue());
	            // add the item listener to the check box
	            checkBox.addItemListener(this);
	            // add the check box to the check box panel
	            checkPanel.add(checkBox);
	        }
	        add(checkPanel, constraints); 
	        
	        // add a separator, setting its position in the layout and a full width dimension
	        constraints.gridwidth = 2;
	        constraints.gridx = 0;
	        constraints.gridy = 4;	
			add(new JSeparator(SwingConstants.HORIZONTAL), constraints);
	        
			// restore the column width dimension
			constraints.gridwidth = 1;
			// put the fourth label in the panel, setting its position in the layout
			constraints.gridx = 0;
			constraints.gridy = 5;	
	        add(new JLabel("End Events:"), constraints);	    	
	    	
	     	// put the relative element in the panel, setting its position in the layout
	        constraints.gridx = 1;
	        constraints.gridy = 5;
	        // get the HashMap from the filter
	        endEvents = new HashMap<String, Boolean>(footprintFilter.getEndEvents());	
	        // create the check box panel
	        JPanel checkPanel2 = new JPanel(new GridLayout(0, 1));
	        // iterate over the map elements
	        for (Map.Entry<String, Boolean> endEvent : endEvents.entrySet()) {
	            // create a check box with the event name
	            JCheckBox checkBox = new JCheckBox(endEvent.getKey());
	            // set the action command with the event name
	            checkBox.setActionCommand("end-"+endEvent.getKey());
	            // set or not as selected the check box, depending on the boolean value in the HashMap
	            checkBox.setSelected(endEvent.getValue());
	            // add the item listener to the check box
	            checkBox.addItemListener(this);	
	            // add the check box to the check box panel
	            checkPanel2.add(checkBox);	
	        }
	        add(checkPanel2, constraints);	        
		} else { // otherwise, if the filter is incomplete
	        // set the other unnecessary elements as null
	        classifierMenu = null;
	        startEvents = null;
	        endEvents = null;
		}		
		// add a separator, setting its position in the layout and a full width dimension
		constraints.gridwidth = 2;
		constraints.gridx = 0;
		constraints.gridy = 6;	
		add(new JSeparator(SwingConstants.HORIZONTAL), constraints);
		
		// now for every situation described above, we add the same buttons to read and manage the values
		
		// add the first button to the panel, setting its position in the layout
		constraints.gridx = 0;
		constraints.gridy = 7;
	    JButton setFiltersAndStartAlgorithm = new JButton("Save Filters AND Start Algorithm");
	    // set the button action command
        setFiltersAndStartAlgorithm.setActionCommand("set-start");
        // add the button to the list for the action listener management
        buttonList.add(setFiltersAndStartAlgorithm);
        add(setFiltersAndStartAlgorithm, constraints);
        
        // add the second button to the panel, setting its position in the layout
        constraints.gridx = 0;
        constraints.gridy = 8;
        JButton setFilters = new JButton("Save Filters AND NOT Start Algorithm");
        // set the button action command
        setFilters.setActionCommand("set");
        // add the button to the list for the action listener management
        buttonList.add(setFilters);
        add(setFilters, constraints);

        // add the third button to the panel, setting its position in the layout
        constraints.gridx = 0;
        constraints.gridy = 9;
        JButton startAlgorithm = new JButton("Start Algorithm AND NOT Save Filters");
        // set the button action command
        startAlgorithm.setActionCommand("start");
        // add the button to the list for the action listener management
        buttonList.add(startAlgorithm);
        add(startAlgorithm, constraints);	        

        // set the panel as opaque
        setOpaque(true);	
	}
	
	/**
	 * Returns the <code>EventLogEditor</code> owner of this panel.
	 * 
	 * @return the <code>EventLogEditor</code> owner of this panel.
	 */
	public EventLogEditor getEventLogEditor() {
		return eventLogEditor;
	}	
	
	/**
	 * Returns the minimum trace occurrences set inside the {@link #minTraceOccurrenceField}.
	 * 
	 * @return the minimum trace occurrences set inside the {@link #minTraceOccurrenceField}.
	 */
	public String getMinTraceOccurrenceValue() {
		return minTraceOccurrenceField.getText();
	}
	
	/**
	 * Returns the selected <code>XEventClassifier</code> inside the {@link #classifierMenu}.
	 * 
	 * @return the selected <code>XEventClassifier</code> inside the {@link #classifierMenu}.
	 */
	public XEventClassifier getSelectedClassifier() {
		// get the selected classifier in the filter, then check if a new one has been selected
		XEventClassifier selectedClassifier = footprintFilter.getSelectedClassifier();
		if(classifierMenu != null && footprintFilter.isComplete()) {
			// get the classifier selected in the menu
			String selectedClassifierName = (String)classifierMenu.getSelectedItem();
			// iterate over the registered classifiers in the filter
			Iterator<XEventClassifier> classifierIterator = footprintFilter.getClassifiers().iterator();
	        while (classifierIterator.hasNext()) {
	        	XEventClassifier classifier = classifierIterator.next();
	        	// if we find the selected classifier in the menu, inside the registered ones in the filter, return that element
	        	if (selectedClassifierName.equals(classifier.name())) {
	        		selectedClassifier = classifier;
	        		break;
	        	}	        	
	        }
		}
		return selectedClassifier;
	}
	
	/**
	 * Returns the <code>Collection</code> of <code>XEventClassifier</code> registered for the log inside the <code>EventLogEditor</code>.
	 *
	 * @return the <code>Collection</code> of <code>XEventClassifier</code> registered for the log inside the <code>EventLogEditor</code>.
	 */
	public Collection<XEventClassifier> getClassifiers() {
		return footprintFilter.getClassifiers();
	}
	
	/**
	 * Returns the <code>HashMap</code> that contains all the possible start events and their selection in a boolean value.
	 * 
	 * @return the <code>HashMap</code> that contains all the possible start events and their selection in a boolean value.
	 */
	public HashMap<String, Boolean> getStartEvents() {
		return startEvents;
	}
	
	/**
	 * Returns the <code>HashMap</code> that contains all the possible end events and their selection in a boolean value.
	 * 
	 * @return the <code>HashMap</code> that contains all the possible end events and their selection in a boolean value.
	 */
	public HashMap<String, Boolean> getEndEvents() {
		return endEvents;
	}

	/**
	 * Sets an <code>ActionListener</code> to all the <code>JButton</code> of the panel, that are
	 * inside the {@link #buttonList}.
	 * 
	 * <p>The method also sets a "owner" propriety to every button, associated with this class, to 
	 * define the button owner (that is this panel).
	 * 
	 * @param listener the <code>ActionListener</code> for all the <code>JButton</code> of the panel
	 */
	public void addButtonActionListener(ActionListener listener) {
		// for every button inside the button list
		for (JButton button : buttonList) {			
			// set the button owner
			button.putClientProperty("owner", this);
			// set the action listener
			button.addActionListener(listener);
		}		
	}

	/**
	 * Invoked when an item has been selected or deselected by the user.
	 * 
	 * <p>The method gets the check box that generates the event, then depending
	 * on its action command (that defines the event name associated with this
	 * check box), it implements the relative operation.
	 * 
	 * <p>If the check box has been selected, the method updates its value to <code>true</code>
	 * in the start events map or in the end events map, depending where the event resides.
	 * 
	 * <p>if the check box has been deselected instead, the method updates its value to <code>false</code>
	 * in the start events map or in the end events map, depending where the event resides.
	 */
	@Override
	public void itemStateChanged(ItemEvent e) {
		// get the check box action command
		String actionCommand = ((JCheckBox)e.getItemSelectable()).getActionCommand();
		// divide the command, the first part is the map, the second one is the event name
		String[] eventName = actionCommand.split("-");
		if(eventName != null) {
			// if the event resides in the start events map
			if(eventName[0].equals("start")) {
				// check if the map contains this key
				if (startEvents.containsKey(eventName[1])) {
					// if the event has been selected
					if (e.getStateChange() == ItemEvent.SELECTED) {
			            // update the value to true
			            startEvents.put(eventName[1], true);				
			        } else if (e.getStateChange() == ItemEvent.DESELECTED) { // if the event has been deselected	  
			        	 // update the value to false
			        	startEvents.put(eventName[1], false);
			        }
				}
			} else if (eventName[0].equals("end")) {
				// check if the map contains this key
				if (endEvents.containsKey(eventName[1])) {
					if (e.getStateChange() == ItemEvent.SELECTED) {
						// update the value to true
						endEvents.put(eventName[1], true);
					} else if (e.getStateChange() == ItemEvent.DESELECTED) {// if the event has been deselected
						// update the value to false
						endEvents.put(eventName[1], false);
					}
				}
			}
		}			
	}
	
}
