/*
 * Event Log Manager
 *   
 *  Copyright (C) 2015 Francesco Santeramo
 *
 *  This file is part of Event Log Manager.
 *
 *  Event Log Manager is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Event Log Manager is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *	
 *  Author:  Francesco Santeramo 
 *  Website: www.francescosanteramo.com
 *  Contact: eventlogmanager@francescosanteramo.com    
 */

package unipi.bpm.elm.santeramo.view;

import java.awt.BorderLayout;
import java.awt.Insets;
import java.awt.Toolkit;
import java.net.URL;

import javax.swing.ImageIcon;
import javax.swing.JInternalFrame;
import javax.swing.JScrollPane;
import javax.swing.JTextPane;

/**
 * This class extends <code>JInternalFrame</code> to create a more detailed 
 * version of a generic internal frame.
 * 
 * <p>This specific internal frame is set to be an info page that displays the result of the 
 * {@link unipi.bpm.elm.santeramo.controller.AnalysisManager#generateEventLogFootprint(boolean, EventLogEditor, FootprintFilter...) generateEventLogFootprint(boolean, EventLogEditor, FootprintFilter...)}
 * method.
 * 
 * <p>The <code>JTextPane</code> of this class will contain the XHTML code produced by this method
 * and it will display it as the content of the frame.
 * 
 * <p>This class also implements specific functions to remind if this page has already been 
 * exported or not and in which file.
 * 
 * @author Francesco Santeramo (eventlogmanager@francescosanteramo.com)
 * 
 * @see JInternalFrame
 * @see JTextPane
 */
public class FootprintMatrixInfo extends JInternalFrame{
	
	/*
	 * VARIABLES
	 */
	
	/**
	 * Auto-Generated Serial Version UID for this class.
	 */
	private static final long serialVersionUID = 5901481353976541071L;
	
	/**
	 * Defines the area where to display HTML content.
	 * 
	 * @see JTextPane
	 */
	JTextPane textArea;
	
	/**
	 * Defines the <code>String</code> that will contain
	 * the XHTML code displayed in the {@link #textArea}.
	 * 
	 * <p>This variable keeps the original XHTML code that has
	 * been set inside the {@link #textArea}: in fact when we set
	 * the content of this area, this will be elaborated and then
	 * showed. So, if we get it back from the container, it won't
	 * be exactly the same one as the beginning. Using this variable,
	 * we are sure to have the original XHTML code ready to use when
	 * needed.
	 * 
	 * <p>The method {@link #getTextAreaContent()} returns this value.
	 */
	private String xhtml;	
	
	/**
	 * Defines if the {@link #textArea} content has been exported or not.
	 * 
	 * <p>This variable is set to <code>true</code> when the user exports the content
	 * for the first time. It is set to <code>false</code> otherwise to show that
	 * no exporting has been made to the content. 
	 */
	private boolean isContentExported;
	
	/**
	 * This string contains the absolute path of the file associated with
	 * this page. It has a <code>null</code> value instead if the content
	 * has never been exported yet.
	 */
	private String absolutePath;
	
	
	
	/*
	 * CONSTRUCTOR
	 */
	
	/**
	 * Creates a <code>FootprintMatrixInfo</code> with the specified title, resizability, 
	 * closability, maximizability, iconifiability and ID.
	 * 
	 * <p>Then defines all its settings by calling the {@link #frameSettings(int)}
	 * method.
	 * 
	 * @param title the String to display in the title bar
	 * @param resizable if true, the internal frame can be resized
	 * @param closable if true, the internal frame can be closed
	 * @param maximizable if true, the internal frame can be maximized
	 * @param iconifiable if true, the internal frame can be iconified
	 * @param internalFrameID the identifier of the <code>FootprintMatrixInfo</code> frame
	 */
	public FootprintMatrixInfo(String title, boolean resizable, boolean closable, boolean maximizable, boolean iconifiable, int internalFrameID) {
		super(title, resizable, closable, maximizable, iconifiable);
		// initialise the text area and the xhtml container
		textArea = new JTextPane();	
		xhtml = "";
		// set the content as NOT exported
		isContentExported = false;
		// set the absolute path of the associated file as null (no associated file)
		absolutePath = null;
		// define the frame settings
		frameSettings(internalFrameID);
	}
	
	
	
	/*
	 * METHODS
	 */
	
	/**
	 * Sets the graphic details and components of the frame.
	 * 
	 * @param internalFrameID the identifier of the <code>FootprintMatrixInfo</code> frame
	 */
	private void frameSettings(int internalFrameID) {
		// set name with the internal frame ID and default position and dimension
    	this.setName(String.valueOf(internalFrameID));
        this.setBounds(0, 0, 600, 400);        
        // get the icon for the frame and set it if available
		String resourcePath = "/images/icon16x16.png";
        URL icon16x16URL = this.getClass().getResource(resourcePath);               
        if (icon16x16URL != null) {
        	this.setFrameIcon(new ImageIcon(Toolkit.getDefaultToolkit().getImage(icon16x16URL)));
        } 
		// set the text area
		textArea.setContentType("text/html");														
		textArea.setEditable(false);
		textArea.setMargin(new Insets(2,4,2,4));
		JScrollPane eventLogEditorScrollPane = new JScrollPane(textArea);	   
		// add the scrolling area on the centre of the frame
		this.add(eventLogEditorScrollPane, BorderLayout.CENTER);
		this.revalidate();
		this.repaint();
	}	
	
	/**
	 * This method returns the <code>JTextPane</code> {@link #textArea}.
	 * 
	 * @return the <code>JTextPane</code> text area.
	 */	
	public JTextPane getTextArea() {
		return textArea;
	}
	
	/**
	 * This method returns the content of the <code>JTextPane</code> {@link #textArea}.
	 * 
	 * <p>More precisely, it returns the value of the {@link #xhtml} variable which keeps
	 * the original XHTML code used to set the <code>JTextPane</code> content.
	 * 
	 * <p>On this way, the method returns the (right) original XHTML code and not the 
	 * (wrong) elaborated one by the <code>JTextPane</code>.
	 * 
	 * @return the <code>String</code> representing the XHTML content of the <code>JTextPane</code>.
	 */
	public String getTextAreaContent() {
		return getXHTML();
	}
	
	/**
	 * This method sets the content of the <code>JTextPane</code> {@link #textArea}.
	 * 
	 * <p>This method also sets the value of the {@link #xhtml} variable using the passed
	 * parameter, to save the original XHTML code used to set the <code>JTextPane</code> content.
	 * 
	 * @param content the content to set inside the <code>JTextPane</code>
	 */
	public void setTextAreaContent(String content) {
		// set the text area content
		setXHTML(content);
		// record the original XHTML code used to generate the text area content
		textArea.setText(content);
		// set the caret at the beginning of the document, to show first the top of it
		textArea.setCaretPosition(0);
	}
	
	/**
	 * This method returns the value of the {@link #xhtml} variable.
	 * 
	 * @return the value of the {@link #xhtml} variable.
	 */
	private String getXHTML() {
		return xhtml;
	}
	
	/**
	 * This method sets the value of the {@link #xhtml} variable.
	 * 
	 * @param xhtmlCode the original xhtml code used to set the {@link #textArea} content
	 */
	private void setXHTML(String xhtmlCode) {
		xhtml = xhtmlCode;
	}
	
	/**
	 * This method returns the <code>boolean</code> {@link #isContentExported} variable value.
	 * 
	 * @return the <code>boolean</code> value of the {@link #isContentExported} variable.
	 */
	public boolean isContentExported() {
		return isContentExported;
	}
	
	/**
	 * This method sets the value of the {@link #isContentExported} variable, that specifies 
	 * if the {@link #textArea} content has been exported or not.
	 * 
	 * @param contentExported <code>true</code> if the content has been exported, <code>false</code> otherwise
	 */
	public void setContentExported(boolean contentExported) {
		isContentExported = contentExported;
	}	
	
	/**
	 * This method returns the {@link #absolutePath} of the file associated with this document.
	 * 
	 * @return the string representing the absolute path of the associated file, <code>null</code> if there is not an associated file yet.
	 */
	public String getAbsolutePath() {
		return absolutePath;
	}	
	
	/**
	 * This method sets the value of the {@link #absolutePath} variable, that specifies 
	 * the absolute path of the file associated with this document.
	 * 
	 * @param path the <code>String</code> representing the absolute path of the file associated with this editor
	 */
	public void setAbsolutePath(String path) {
		absolutePath = path;
	}	
}
