/*
 * Event Log Manager
 *   
 *  Copyright (C) 2015 Francesco Santeramo
 *
 *  This file is part of Event Log Manager.
 *
 *  Event Log Manager is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Event Log Manager is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *	
 *  Author:  Francesco Santeramo 
 *  Website: www.francescosanteramo.com
 *  Contact: eventlogmanager@francescosanteramo.com    
 */

package unipi.bpm.elm.santeramo.view;

import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Insets;
import java.awt.Toolkit;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.net.URL;

import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JSeparator;

/**
 * This class extends <code>FlowLayout</code> to create a specific version of this layout
 * for the {@link unipi.bpm.elm.santeramo.view.EventLogManagerView#menuArea menuArea} tabs.
 * 
 * <p>This class implements a layout that acts as the <code>FlowLayout</code>
 * layout until there is enough space in the container to position all the components
 * inside one single row.
 * 
 * <p>When there is not enough space for positioning all the components in one single line
 * inside the container, differently from the <code>FlowLayout</code> layout that starts putting
 * the remaining components on a new line and continues on this way until all of them are
 * positioned, the <code>MenuAreaTabLayout</code> layout positions all the components that the
 * container can contain in one single line and then add, as last component of the line, a special
 * component with a pop-up linked to it that contains all the remaining components that the
 * container can't contain in one single line.
 * 
 * <p>In our specific case, the components inside the container that we want to manage are <code>JButton</code>
 * objects that need to be repositioned when it is not possible anymore to display them all in the same
 * line after a resizing.
 * 
 * <p>When a resize of the container occurs, this layout checks if there is the need to reposition
 * the components, calculating how many buttons now can be displayed in one single line. 
 * If not all the buttons can fit in one line, it puts as many buttons as it can inside the container
 * and the remaining in a pop-up menu linked to a last special button added at the end of the line
 * that shows the remaining ones as a list. *
 *  
 * @author Francesco Santeramo (eventlogmanager@francescosanteramo.com)
 * 
 * @see FlowLayout
 */
public class MenuAreaTabLayout extends FlowLayout {
	
	/*
	 * VARIABLES
	 */
	
	/**
	 * Auto-Generated Serial Version UID for this class.
	 */
	private static final long serialVersionUID = -7960282433267216319L;
	
	/**
	 * The <code>JPanel</code> component that this layout has to manage 
	 * (we need it here for setting the layout for this specific container).
	 * 
	 * @see JPanel
	 */
	private JPanel targetPanel;
	
	/**
	 * This array contains the <code>Component</code> elements of the container
	 * that have been added to it at the start of the application.
	 * 
	 * @see Component
	 */
	private Component[] targetPanelComponents;
	
	/**
	 * The "More" button to add to the container when not all of its <code>Component</code>
	 * elements fit in one single row.
	 * 
	 * @see Component
	 */
	private MenuAreaButton more;
	
	/**
	 * Pop-up used when the user clicks on the More button.
	 * 
	 * @see JPopupMenu
	 */
	private JPopupMenu morePopup;
	
	/**
	 * The <code>JPanel</code> panel inside {@link #morePopup} that contains the <code>MenuAreaButton</code> 
	 * buttons that didn't fit in the first row of {@link #targetPanel}.
	 * 
	 * @see JPanel
	 * @see MenuAreaButton
	 */
	private JPanel morePopupPanel;
	
	/**
	 * Registers how many components are in the panel after the last modification,
	 * so we can check if, after a resize, the space is still enough for this number
	 * of components or not and modify the panel depending on this.
	 */
	private int howManyComponentsNow;
	
	
	
	/*
	 * CONSTRUCTORS
	 */
	
	/**
	 * Constructor with the target <code>JPanel</code> container as parameter that
	 * calls the {@link #MenuAreaTabLayout(int, int, int, JPanel)} constructor to
	 * create a layout with default parameters (align=LEFT, hgap=0, vgap=0) associated
	 * with the target container.
	 * 
	 * @param target the target container to manage
	 * 
	 * @see JPanel
	 */
	public MenuAreaTabLayout(JPanel target) {
      this(FlowLayout.LEFT, 0, 0, target);
	}
	
	/**
	 * Constructor with alignment and the target <code>JPanel</code> container as 
	 * parameters that calls the {@link #MenuAreaTabLayout(int, int, int, JPanel)}
	 * constructor to create a layout with the specified alignment and default hgap and vgap 
	 * (hgap=0, vgap=0) associated with the target container.
	 * 
	 * @param align alignment of the positioned components
	 * @param target the target container to manage
	 * 
	 * @see JPanel
	 */
	public MenuAreaTabLayout(int align, JPanel target) {
	    this(align, 0, 0, target);
	}
		
	/**
	 * Constructor with alignment, horizontal gap, vertical gap and the target
	 * <code>JPanel</code> container as parameters that calls the super class 
	 * constructor to create a layout with the specified information associated
	 * with the target container.
	 *  
	 * @param align alignment of the positioned components
	 * @param hgap horizontal gap between components
	 * @param vgap vertical gap between components
	 * @param target the target container to manage
	 *
	 * @see JPanel 
	 */
	public MenuAreaTabLayout(int align, int hgap, int vgap, JPanel target) {
		// call the constructor of the super class
	    super(align, hgap, vgap);
	    // initialise all the instance variables
        targetPanel = target;   
		synchronized (targetPanel.getTreeLock()) {    			
    		targetPanelComponents = targetPanel.getComponents();
		}  		
		// create a "More" button with these settings
		String resourcePath = "/images/more.png";
		this.more = newMenuAreaButton("More..", resourcePath, "Show more options", "more");
		this.morePopup = new JPopupMenu();
		this.morePopupPanel = new JPanel();
		this.howManyComponentsNow = targetPanelComponents.length;
		// set graphic details of the pop-up menu panel and add a listener to the More button
		setMoreButton();	
	}
	
	
	
	/*
	 * METHODS
	 */
	
	/**
	 * Creates a new {@link MenuAreaButton} button with text,
	 * icon (optional), tool tip and action command.
	 * 
	 * @param buttonText the text of the button
	 * @param buttonIconURL the URL of the image to display on the button
	 * @param buttonToolTip the tool tip to display
	 * @param buttonActionCommand the action command of the button
	 * 
	 * @return a {@link MenuAreaButton} button.
	 */
	private MenuAreaButton newMenuAreaButton(String buttonText, String buttonIconURL, String buttonToolTip, String buttonActionCommand) {		
        Icon buttonIcon = null;
		URL buttonIconURLResource = this.getClass().getResource(buttonIconURL); // get the image from the URL
		if (buttonIconURLResource != null) { // if the image exists
        	buttonIcon = new ImageIcon(Toolkit.getDefaultToolkit().getImage(buttonIconURLResource)); // get the icon from the image
        }
		return new MenuAreaButton(buttonText, buttonIcon, buttonToolTip, buttonActionCommand);
	}
	
	/**
	* Sets the graphic details of the pop-up menu panel and add it to the pop-up menu
	* of the More button. Then adds a listener to the More button that show the pop-up menu
	* every time the user clicks on this button.
	*/
	private void setMoreButton() { 
		// set pop-up menu panel graphics
		morePopupPanel.setOpaque(true);
		morePopupPanel.setBackground(Color.WHITE);
		// add the panel to the pop-up menu (it will be the content of the pop-up)
		morePopup.insert(morePopupPanel, 0);    		
		// add a listener to the More button
		more.addMouseListener(new MouseAdapter() { 
			@Override
			public void mousePressed(MouseEvent e) {
				// when the button get pressed, show the related pop-up menu
				morePopup.show(e.getComponent(), 0, e.getComponent().getHeight());
			}
		});
	}

	/**
     * Lays out the container. This method lets each visible component take its
     * preferred size until there is enough space for displaying all the components
     * in one single row inside the container.
     * 
     * <p>When there is not enough space for positioning all the components with their
     * preferred size in one single row inside the container, the method first finds the
     * number of components that it can display in one row and it positions them on the 
     * container, then it adds a More button at the end of the line with a pop-up menu
     * linked to it, containing the remaining components that didn't fit in the first row
     * of the container.
     *
     * @param target the specified container being laid out
     */
	@Override
	public void layoutContainer(Container target) {	 
		
		/*
		 * call the method of the super class to set components size (based on their
		 * preferred size) and position in the container, so they can now be displayed 
		 */
		super.layoutContainer(target);
		/*
		 * now take all the components and, based on the container dimension, display 
		 * some of them (or all) in the container panel and some of them (or none) in the
		 * pop-up menu of the "More" button 
		 */
		synchronized (target.getTreeLock()) {  
			// if the container is a JPanel
			if (target instanceof JPanel) {   
	    		// check if the JPanel is exactly the one that we are managing with this layout
	    		if (targetPanel.equals((JPanel)target)) { 
	    			// calculate how many elements can now fit in one row
	    			int componentsOneRow = howManyInOneRow(); 
	    			// if not all the elements can fit in one single row
	            	if (componentsOneRow < targetPanelComponents.length) {   
	            		// clear the panel from all elements
		    			targetPanel.removeAll();	            		  
	            		/* considering that the last position in the row is for the More button (and so the last element
	            		 * that we could have displayed is actually the first that goes inside the pop-up menu of the More button), if the
	            		 * last but one element to display is a separator, ignore it and show only until the last but two element */                		
		    			int moreButtonIndex; 
		    			if (targetPanelComponents[componentsOneRow-1] instanceof JSeparator) {
	            			moreButtonIndex = (componentsOneRow-2); // until the last but two element
	            		} else {
	            			moreButtonIndex = (componentsOneRow-1); // until the last but one element
	            		}      
	            		// display on the main panel all the components that we can, leaving space for the More button
	            		for (int i = 0; i < moreButtonIndex; i++) {
	            			targetPanel.add(targetPanelComponents[i]);
	            		}       
	            		// clear the pop-up menu panel from all elements
	            		morePopupPanel.removeAll();
	            		/* add to the pop-up menu panel all the remaining components 
	            		 * (this panel is the content of the pop-up menu, displayed when the user clicks on the More button) */
	            		for (int i = moreButtonIndex; i < targetPanelComponents.length; i++) {
	            			morePopupPanel.add(targetPanelComponents[i]);
	            		} 
	            		// re-validate and repaint the pop-up menu panel
	            		morePopupPanel.revalidate();
	            		morePopupPanel.repaint();
	            		// add the More button to the main panel as last element in the row
	            		targetPanel.add(more);  
	            		howManyComponentsNow = componentsOneRow;	            		
	            	} else { // if all the elements can be displayed
	            		if (howManyComponentsNow < targetPanelComponents.length) {
	            			// clear the panel from all elements
	    	    			targetPanel.removeAll();
	    	    			// add all the elements to the main panel
		            		for (int i = 0; i < targetPanelComponents.length; i++) {
		            			targetPanel.add(targetPanelComponents[i]);
		            		}
		            		howManyComponentsNow = targetPanelComponents.length;
		    			}	            		
	            	}	            	
	    		}
			}
		}
	}

	/**
	 * Returns the number of components that fits in one single row
	 * in the <code>target</code> container in the specific moment
	 * when the method is called (the container can have different 
	 * sizes in different moments).
	 * 
	 * @return the number of components that fits in one single row.
	 */
	private int howManyInOneRow() {
		// get the insets of the container and the hgap
		Insets containerInsets = targetPanel.getInsets();
		int horizGap = getHgap();
		// get the width of the container without the insets and hgap			
		int containerWidth = targetPanel.getWidth() - (containerInsets.left + containerInsets.right + horizGap*2);
		
		int rowWidth = 0;
		// for every component of the container
		for (int i = 0; i < targetPanelComponents.length; i++) {
			// get the component
			Component component = targetPanelComponents[i];
			// if the component has to be displayed
			if (component.isVisible()) {
				// get the size of the component
				Dimension componentSize = component.getPreferredSize();
				/* if there is no space on the first row for displaying this component, return the index of it
				 * (that is also the number of component before it that can be displayed) */
				if (rowWidth + componentSize.width > containerWidth) {
					return i;
				}
				// otherwise add the component size and the hgap to the total amount and go on checking
				rowWidth = rowWidth + componentSize.width + horizGap;
			}
		}
		// if all the components can fit the first row, return the total number
		return targetPanelComponents.length;
	}
}
		